/*-
 * Copyright (c) 2010 The FreeBSD Foundation
 * All rights reserved.
 *
 * This software was developed by Max Khon under sponsorship from
 * the FreeBSD Foundation and Ethon Technologies GmbH.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bsd-compat.c 9253 2010-09-02 10:12:09Z fjoe $
 */

#include <sys/types.h>
#include <sys/module.h>

#include <asm/atomic.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/workqueue.h>

MALLOC_DEFINE(M_FARSYNC, "farsync", "FarSync data structures");

/*
 * Tasklet API <linux/interrupt.h>
 */
static void
tasklet_run(void *context, int pending)
{
	struct tasklet_struct *t = (struct tasklet_struct *) context;

	if (atomic_read(&t->disable_count))
		return;
	t->func(t->data);
}

void
_tasklet_init(struct tasklet_struct *t)
{
	TASK_INIT(&t->task, PI_REALTIME, tasklet_run, t);
	t->disable_count = 0;
}

void
tasklet_init(struct tasklet_struct *t, void (*func)(unsigned long), unsigned long data)
{
	_tasklet_init(t);
	t->func = func;
	t->data = data;
}

void
tasklet_hi_schedule(struct tasklet_struct *t)
{
	taskqueue_enqueue(taskqueue_fast, &t->task);
}

void
tasklet_disable(struct tasklet_struct *t)
{
	atomic_inc(&t->disable_count);
}

void
tasklet_enable(struct tasklet_struct *t)
{
	atomic_dec(&t->disable_count);
}

void
tasklet_kill(struct tasklet_struct *t)
{
	taskqueue_drain(taskqueue_fast, &t->task);
}

/*
 * Kernel module API <linux/module.h>
 */
int
try_module_get(struct module *m)
{
	atomic_inc(&m->refcount);
	return (1);
}

void
module_put(struct module *m)
{
	atomic_dec(&m->refcount);
}

int
module_refcount(struct module *m)
{
	return atomic_read(&m->refcount);
}

int
_linux_module_modevent(module_t mod, int type, void *data)
{
	int res = 0;
	struct module *m = (struct module *) data;

	switch (type) {
	case MOD_LOAD:
		if (m->init)
			res = m->init();
		return (-res);
	case MOD_UNLOAD:
		if (m->exit)
			m->exit();
		return (0);
	case MOD_QUIESCE:
		if (module_refcount(m) > 0)
			return (EBUSY);
		return (0);
	default:
		return (EOPNOTSUPP);
	}
}

void
_linux_module_ptr_sysinit(void *arg)
{
	struct module_ptr_args *args = (struct module_ptr_args *) arg;
	*args->pfield = args->value;
}

/*
 * sk_buff API <linux/skbuff.h>
 */
struct sk_buff *
dev_alloc_skb(unsigned int length)
{
	struct sk_buff *m;
	unsigned int len;

	len = length + MLEN - MHLEN;
	if (len <= MHLEN)
		len = MSIZE;
	else if (len <= MCLBYTES)
		len = MCLBYTES;
	else if (len <= MJUM9BYTES)
		len = MJUM9BYTES;
	else if (len <= MJUM16BYTES)
		len = MJUM16BYTES;
	else {
		printf("%s: cannot allocate mbuf of %u bytes\n", __func__, len);
		return NULL;
	}

	m = m_getjcl(M_DONTWAIT, MT_DATA, M_PKTHDR, len);
	if (m == NULL)
		return NULL;
	m->m_pkthdr.len = m->m_len = length;
	return m;
}

void
kfree_skb(struct sk_buff *skb)
{
	m_freem(skb);
}

void
dev_kfree_skb(struct sk_buff *skb)
{
	m_freem(skb);
}

/*
 * Workqueue API
 */
void
_work_run(void *context, int pending)
{
	struct work_struct *work = (struct work_struct *) context;
	work->func(work);
}

void
schedule_work(struct work_struct *work)
{
	work->tq = taskqueue_fast;
	taskqueue_enqueue(taskqueue_fast, &work->task);
}

void
cancel_work_sync(struct work_struct *work)
{
	taskqueue_drain(work->tq, &work->task);
}

void
flush_work(struct work_struct *work)
{
	taskqueue_drain(work->tq, &work->task);
}

struct workqueue_struct *
create_singlethread_workqueue(const char *name)
{
	int res;
	struct workqueue_struct *wq;

	wq = malloc(sizeof(*wq), M_LINUX, M_NOWAIT);
	if (wq == NULL)
		return NULL;

	wq->tq = taskqueue_create_fast(name, M_NOWAIT, taskqueue_thread_enqueue, &wq->tq);
	if (wq->tq == NULL) {
		free(wq, M_LINUX);
		return NULL;
	}

	res = taskqueue_start_threads(&wq->tq, 1, PI_REALTIME, "%s taskq", name);
	if (res) {
		destroy_workqueue(wq);
		return NULL;
	}

	return wq;
}

void
destroy_workqueue(struct workqueue_struct *wq)
{
	taskqueue_free(wq->tq);
	free(wq, M_LINUX);
}

void
flush_workqueue(struct workqueue_struct *wq)
{
	/* TODO */
}

void
queue_work(struct workqueue_struct *wq, struct work_struct *work)
{
	work->tq = wq->tq;
	taskqueue_enqueue(wq->tq, &work->task);
}
