/*-
 * Copyright (c) 2011 The FreeBSD Foundation
 * All rights reserved.
 *
 * This software was developed by Max Khon.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: ng_fst_netdev.c 10322 2011-11-04 11:55:05Z fjoe $
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/queue.h>
#include <sys/linker.h>
#include <sys/kernel.h>
#include <sys/syscallsubr.h>

#include <linux/netdevice.h>
#include <netgraph/ng_message.h>
#include <netgraph/netgraph.h>
#include <netgraph/ng_iface.h>

#include "ng_farsync_netdev.h"

MALLOC_DECLARE(M_FARSYNC);

int _fst_start_xmit(struct mbuf *m, struct fst_port_info *port);

#if __FreeBSD_version < 800000
struct ng_node *ng_name2noderef(struct ng_node *node, const char *name);
#endif

#define FARSYNC_NETDEV_HOOK_UPPER "upper"

static ng_rcvmsg_t ng_fst_netdev_rcvmsg;
static ng_shutdown_t ng_fst_netdev_shutdown;
static ng_newhook_t ng_fst_netdev_newhook;
static ng_disconnect_t ng_fst_netdev_disconnect;
static ng_rcvdata_t ng_fst_netdev_rcvdata;

static struct ng_type ng_fst_netdev_typestruct = {
	.version =	NG_ABI_VERSION,
	.name =		"farsync_netdev",
	.rcvmsg =	ng_fst_netdev_rcvmsg,
	.shutdown =	ng_fst_netdev_shutdown,
	.newhook =	ng_fst_netdev_newhook,
	.rcvdata =	ng_fst_netdev_rcvdata,
	.disconnect =	ng_fst_netdev_disconnect,
};
NETGRAPH_INIT_ORDERED(farsync_netdev, &ng_fst_netdev_typestruct,
    SI_SUB_DRIVERS, SI_ORDER_FIRST);

struct fst_netdev {
	struct fst_port_info *port;
	struct ng_node *node;		/**< our netgraph node */
	struct ng_hook *upper;		/**< our upper hook */
};

/**
 * Ensure that specified netgraph type is available
 */
static int
ng_ensure_type(const char *type)
{
	int error;
	int fileid;
	char filename[NG_TYPESIZ + 3];

	if (ng_findtype(type) != NULL)
		return (0);

	/* Not found, try to load it as a loadable module. */
	snprintf(filename, sizeof(filename), "ng_%s", type);
	error = kern_kldload(curthread, filename, &fileid);
	if (error != 0)
		return (-1);

	/* See if type has been loaded successfully. */
	if (ng_findtype(type) == NULL) {
		(void)kern_kldunload(curthread, fileid, LINKER_UNLOAD_NORMAL);
		return (-1);
	}

	return (0);
}

/**
 * Create a netgraph node and connect it to ng_iface instance
 *
 * @return 0 on success, -1 on error
 */
struct net_device *
fst_netdev_create(struct fst_port_info *port, const char *name)
{
	struct net_device *dev = NULL;
	struct fst_netdev *fdev;
	struct ng_node *node;
	struct ng_mesg *msg;
	char node_name[64];
	int error;
	struct ngm_mkpeer *ngm_mkpeer;

	/* check if a netgraph node for that device already exists */
	snprintf(node_name, sizeof(node_name), "fst@%s", name);
	node = ng_name2noderef(NULL, node_name);
	if (node != NULL) {
		printf("fst_netdev(%s): Error: netgraph node `%s' already exists\n",
		    NG_NODE_NAME(node), NG_NODE_NAME(node));
		NG_NODE_UNREF(node);
		return (NULL);
	}

	/* create new network device */
	dev = malloc(sizeof(*dev) + sizeof(*fdev), M_FARSYNC, M_WAITOK | M_ZERO);
	fdev = netdev_priv(dev);
	fdev->port = port;

	/* create new netgraph node */
	error = ng_make_node_common(&ng_fst_netdev_typestruct, &node);
	if (error != 0) {
		printf("fst_netdev(%s): Error: can not create netgraph node (error %d)\n",
		    node_name, error);
		goto error;
	}
	fdev->node = node;
	NG_NODE_SET_PRIVATE(node, dev);
	error = ng_name_node(node, node_name);
	if (error != 0) {
		printf("fst_netdev(%s): Error: can not set netgraph node name (error %d)\n",
		    node_name, error);
		goto error;
	}

	/* create HDLC encapsulation layer peer node */
	error = ng_ensure_type(NG_IFACE_NODE_TYPE);
	if (error != 0) {
		printf("fst_netdev(%s): Error: can not load %s netgraph type error %d)\n",
		    NG_NODE_NAME(node), NG_IFACE_NODE_TYPE, error);
		goto error;
	}

	NG_MKMESSAGE(msg, NGM_GENERIC_COOKIE, NGM_MKPEER, sizeof(*ngm_mkpeer), M_WAITOK);
	ngm_mkpeer = (struct ngm_mkpeer *) msg->data;
	strlcpy(ngm_mkpeer->type, NG_IFACE_NODE_TYPE, sizeof(ngm_mkpeer->type));
	strlcpy(ngm_mkpeer->ourhook, FARSYNC_NETDEV_HOOK_UPPER, sizeof(ngm_mkpeer->ourhook));
	strlcpy(ngm_mkpeer->peerhook, NG_IFACE_HOOK_INET, sizeof(ngm_mkpeer->peerhook));
	NG_SEND_MSG_ID(error, node, msg, NG_NODE_ID(node), NG_NODE_ID(node));
	if (error) {
		printf("fst_netdev(%s): Error: NGM_MKPEER: error %d (%s)\n",
		    NG_NODE_NAME(node), error, NG_IFACE_NODE_TYPE);
		goto error;
	}

	/* get iface name */
	NG_MKMESSAGE(msg, NGM_IFACE_COOKIE, NGM_IFACE_GET_IFNAME, 0, M_WAITOK);
	NG_SEND_MSG_HOOK(error, node, msg, fdev->upper, NG_NODE_ID(node));
	if (error) {
		printf("fst_netdev(%s): Error: NGM_MKPEER: error %d (%s)\n",
		    NG_NODE_NAME(node), error, NG_IFACE_NODE_TYPE);
		goto error;
	}
	printf("fst_netdev(%s): new netgraph node\n", NG_NODE_NAME(node));

	return (dev);

error:
	if (fdev->node != NULL)
		NG_NODE_UNREF(fdev->node);
	free(dev, M_FARSYNC);
	return (NULL);
}

/**
 * Destroy a netgraph node and ng_iface instance associated with it
 */
void
fst_netdev_destroy(struct net_device *dev)
{
	struct fst_netdev *fdev = netdev_priv(dev);

	if (fdev->upper != NULL) {
		int error;
		struct ng_mesg *msg;

		NG_MKMESSAGE(msg, NGM_GENERIC_COOKIE,
		    NGM_SHUTDOWN, 0, M_WAITOK);
		NG_SEND_MSG_HOOK(error, fdev->node, msg,
		    fdev->upper, NG_NODE_ID(fdev->node));
	}
	NG_NODE_REALLY_DIE(fdev->node);	/* Force real removal of node */
	ng_rmnode_self(fdev->node);
}

/**
 * Receive data frame from the synchronous line
 *
 * Receives data frame from the synchronous line and sends it up to the upstream.
 */
void
fst_netdev_rx(struct net_device *dev, struct mbuf *m)
{
	int error;
	struct fst_netdev *fdev = netdev_priv(dev);

	NG_SEND_DATA_FLAGS(error, fdev->upper, m, NG_QUEUE);
}

/**
 * Abort receiving a data frame
 */
void
fst_netdev_abort(struct net_device *dev, int event)
{
	/* nothing to do */
}

/**
 * Wake up transmitter
 */
void
fst_netdev_wakeup_tx(struct net_device *dev)
{
	/* TODO */
}

/**
 * Receive an incoming control message
 */
static int
ng_fst_netdev_rcvmsg(struct ng_node *node, struct ng_item *item, struct ng_hook *lasthook)
{
	struct net_device *dev = NG_NODE_PRIVATE(node);
	struct ng_mesg *msg, *resp = NULL;
	int error = 0;

	NGI_GET_MSG(item, msg);
	switch (msg->header.typecookie) {
	case NGM_IFACE_COOKIE:
		switch (msg->header.cmd) {
		case NGM_IFACE_GET_IFNAME:
			strlcpy(dev->name, msg->data, sizeof(dev->name));
			printf("fst_netdev(%s): interface %s\n",
			    NG_NODE_NAME(node), dev->name);
			break;
		default:
			error = EINVAL;
			break;
		}
		break;
	default:
		error = EINVAL;
		break;
	}
	NG_RESPOND_MSG(error, node, item, resp);
	NG_FREE_MSG(msg);
	return (error);
}

/**
 * Shutdown node
 *
 * Reset the node but does not remove it unless the REALLY_DIE flag is set.
 */
static int
ng_fst_netdev_shutdown(struct ng_node *node)
{
	struct net_device *dev = NG_NODE_PRIVATE(node);

	if (node->nd_flags & NGF_REALLY_DIE) {
		/* destroy the node itself */
		printf("fst_netdev(%s): destroying netgraph node\n",
		    NG_NODE_NAME(node));
		NG_NODE_SET_PRIVATE(node, NULL);
		NG_NODE_UNREF(node);

		/* destroy the iface */
		free(dev, M_FARSYNC);
		return (0);
	}

	NG_NODE_REVIVE(node);		/* Tell ng_rmnode we are persistent */
	return (0);
}

/**
 * Check for attaching a new hook
 */
static int
ng_fst_netdev_newhook(struct ng_node *node, struct ng_hook *hook, const char *name)
{
	struct net_device *dev = NG_NODE_PRIVATE(node);
	struct fst_netdev *fdev = netdev_priv(dev);
	struct ng_hook **hookptr;

	if (strcmp(name, FARSYNC_NETDEV_HOOK_UPPER) == 0) {
		hookptr = &fdev->upper;
	} else {
		printf("fst_netdev(%s): unsupported hook %s\n",
		    NG_NODE_NAME(fdev->node), name);
		return (EINVAL);
	}

	if (*hookptr != NULL) {
		printf("fst_netdev(%s): %s hook is already connected\n",
		    NG_NODE_NAME(fdev->node), name);
		return (EISCONN);
	}

	*hookptr = hook;
	return (0);
}

/**
 * Hook disconnection
 */
static int
ng_fst_netdev_disconnect(struct ng_hook *hook)
{
	struct ng_node *node = NG_HOOK_NODE(hook);
	struct net_device *dev = NG_NODE_PRIVATE(node);
	struct fst_netdev *fdev = netdev_priv(dev);

	if (hook == fdev->upper) {
		fdev->upper = NULL;
	} else {
		panic("fst_netdev(%s): %s: weird hook", NG_NODE_NAME(fdev->node), __func__);
	}

	return (0);
}

/**
 * Receive data
 *
 * Receives data frame from the upstream and sends it down to the synchronous line.
 */
static int
ng_fst_netdev_rcvdata(struct ng_hook *hook, struct ng_item *item)
{
	struct ng_node *node = NG_HOOK_NODE(hook);
	struct net_device *dev = NG_NODE_PRIVATE(node);
	struct fst_netdev *fdev = netdev_priv(dev);
	struct mbuf *m;

	/* get mbuf */
	NGI_GET_M(item, m);
	NG_FREE_ITEM(item);

	if (_fst_start_xmit(m, fdev->port) == 0)
		return (ENOBUFS);

	return (0);
}
