/*
 *	FarSync WAN driver for Linux
 *
 *	DSL-S1 Loopback test program
 *
 *	Copyright (C) 2004 FarSite Communications Ltd.
 *	www.farsite.co.uk
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	File:	fardsltest.c
 *	Author:	Kevin Curtis	<kevin.curtis@farsite.co.uk>
 */
/*	Revision History
 *	$Log: fardsltest.c,v $
 *	Revision 1.2  2009/07/10 08:23:24  kevinc
 *	Merging 1.09.04 and async branches
 *	
 *	Revision 1.1.1.1  2009/05/06 12:45:30  kevinc
 *	
 *	
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>

#ifdef GEN_HDLC
#include <linux/hdlc.h>
#include <linux/sockios.h>
#endif

#include "farsync.h"


/* Define FALSE & TRUE 
 */
#ifndef TRUE
# define TRUE   1
# define FALSE  0
typedef int bool;
#endif

#define NUM_INTERFACES	1	/* Max no of interfaces we will support */

#define FST_MAX_MTU     8*1024
#define TEST_PACKETS	100	/* Number of packets to test with */
#define TEST_SIZE	500	/* Default size of test packet */

/* We're passing data around in global structures for simplicity.
 * OK for simple programs but we should probably be doing a lot more parameter
 * passing in larger programs.
 */
static char *progname;
static int duration = 60;
static int txsize = TEST_SIZE;
static int vpi = 0;
static int vci = 17;
static int test_mode = TEST_MODE_ANALOG_TRANSPARENT_LOOP;
static struct fstioc_info orig_info[NUM_INTERFACES];
static unsigned char rxbuf[ FST_MAX_MTU ];
static unsigned char txbuf[ FST_MAX_MTU ];

static int orig_mtu;		/* Detected MTU. */

#ifdef GEN_HDLC
static bool is_gen_hdlc = FALSE;
#endif


/* Structure for the line stats */
struct line_info {
                                /* Interface stats */
        long    rx_errs;
        long    rx_drop;
        long    rx_fifo;
        long    rx_frame;
        long    tx_errs;
        long    tx_drop;
        long    tx_fifo;
        long    tx_carrier;
};


/* Structure to contain per interface information */
struct if_info {
        int     index;
	int     sock;		/* Socket handle */
	char   *ifname;		/* Interface name */

	long    rx_pkts;	/* Input stats */
	unsigned long long    rxbytes;
	long    rxerrs;

	long    tx_pkts;	/* Output stats */
	unsigned long long    txbytes;
	long    txerrs;
	long    txpause;	/* Delays due to full ouput queue */

        time_t  tx_xfer_start;  /* Tx Start time */
        time_t  tx_xfer_end;    /* and Tx End time */
        time_t  rx_xfer_start;  /* Tx Start time */
        time_t  rx_xfer_end;    /* and Tx End time */

        struct line_info line_errs_start;
        struct line_info line_errs_end;
        unsigned int send_seq;
        unsigned int recv_seq;
        unsigned int lost_packets;
        unsigned int kick_tx;

} if_info[ NUM_INTERFACES ];



/*
 *      Print an error message and exit
 */
static void
error_exit ( char *msg, ...)
{
  va_list ap;
  
  va_start ( ap, msg );
  fprintf ( stderr,"%s: ", progname );
  vfprintf ( stderr, msg, ap );
  va_end ( ap );
  
  exit ( 1 );
}


/*
 * 	Initialise an interface
 *
 *	Print an error message and exit on first error (real programs may wish
 *	to do something better here).
 */
static void
fst_init ( struct if_info *iface, char *name )
{
  struct fstioc_info info;
  struct ifreq req;
#ifdef GEN_HDLC
  static bool probed_gen_hdlc = FALSE;
#endif

  
  /* Sanity check and record the interface name */
  if ( strlen ( name ) >= IFNAMSIZ )
    error_exit ("Interface name %s too long\n", name );
  
  iface->ifname = name;
  
  if (( iface->sock = socket ( PF_INET, SOCK_DGRAM, IPPROTO_IP )) < 0 )
    error_exit ("Unable to create socket. %s\n", strerror ( errno ));
  
  /* Setup the request structure with the interface name */
  memset ( &req, 0, sizeof ( req ));
  strcpy ( req.ifr_name, name );

#ifdef GEN_HDLC
  if ( ! probed_gen_hdlc )
    {
      /* Probe to see if the driver supports the "generic HDLC" system */
      memset ( &info, 0, sizeof ( info ));
      req.ifr_data = (char *)&info;
      if ( ioctl ( iface->sock, FSTGETCONF, &req ))
        {
          error_exit ("Problem getting information for interface %s.\n"
                      "%s\n", name, strerror ( errno ));
        }
      is_gen_hdlc = ( info.proto == FST_GEN_HDLC );
      probed_gen_hdlc = TRUE;
      
      if ( verbose && is_gen_hdlc )
        printf ("System is using generic-HDLC interface\n");
    }
  /*
   * Save the current config so that we can restore it later
   */
  orig_info[iface->index] = info;
  
  
  /* Switch interface to raw mode if not already set */
  if ( is_gen_hdlc )
    {
      req.ifr_settings.type = IF_GET_PROTO;
      req.ifr_settings.data_length = 0;
      if ( ioctl ( iface->sock, SIOCDEVICE, &req ))
        {
          error_exit ("Problem getting protocol information for"
                      " interface %s.\n%s\n",
                      name, strerror ( errno ));
          exit ( 1 );
        }
      
      if ( req.ifr_settings.type != IF_PROTO_HDLC )
        {
          if ( verbose )
            printf ("Setting HDLC mode on %s\n", name );
          
          req.ifr_settings.type = IF_PROTO_RAW;
          req.ifr_settings.data_length = 0;
          
          if ( ioctl ( iface->sock, SIOCDEVICE, &req ))
            {
              error_exit ("Problem setting protocol for interface %s.\n"
                          "%s\n", name, strerror ( errno ));
            }
        }
    }
  else	/* Tricksy drop through to non-ifdef case */
#endif
    {  
      
      /* Check the driver is set to the "raw" operating mode.
       * You could skip all of the code from here down to END SKIP if you
       * chose to rely on some external configuration setup utility such
       * as farutil.  We rely on farutil for line type and clocks anyway.
       */
      memset ( &info, 0, sizeof ( info ));
      req.ifr_data = (char *)&info;
      if ( ioctl ( iface->sock, FSTGETCONF, &req ))
        {
          error_exit ("Problem getting protcol information for interface"
                      " %s.\n%s\n", name, strerror ( errno ));
        }
      
      /*
       * Save the current config so that we can restore it later
       */
      orig_info[iface->index] = info;
      
      /*
       * Now set what we want
       */
      info.proto = FST_RAW;
      info.lineSpeed = 64000;
      info.lineInterface = SHDSL;
      info.transparentMode = 0;
      info.numTxBuffers = 8;
      info.numRxBuffers = 8;
      info.testMode = test_mode;
      info.terminalType =  EQUIP_TYPE_CENTRAL;
      info.vpi = vpi;
      info.vci = vci;

      info.valid = FSTVAL_ALL;
      
      req.ifr_data = (char *)&info;
      if ( ioctl ( iface->sock, FSTSETCONF, &req ))
        {
          error_exit ("Problem setting config for interface %s.\n"
                      "%s\n", name, strerror ( errno ));
        }
    }
  
  /* Bring the interface UP if necessary */
  if ( ioctl ( iface->sock, SIOCGIFFLAGS, &req ) < 0 )
    {
      error_exit ("Error getting interface %s flags.\n%s\n", name,
		  strerror ( errno ));
    }
  
  if ( ! ( req.ifr_flags & IFF_UP ))
    {
      req.ifr_flags |= IFF_UP;
      if ( ioctl ( iface->sock, SIOCSIFFLAGS, &req ) < 0 )
	{
	  error_exit ("Failed to bring interface %s up.\n%s\n", name,
		      strerror ( errno ));
	}
    }
  else
    {
      error_exit ("Please make sure interfaces are down before running this test - %s\n", name);
    }
  
  /* END SKIP (code to setup "raw" protocol) */
  
  
  /* Record interface MTU. */
  if ( ioctl ( iface->sock, SIOCGIFMTU, &req ) < 0 )
    {
      error_exit ("Error getting interface %s MTU.\n%s\n", name,
		  strerror ( errno ));
    }
  orig_mtu = req.ifr_mtu;
  
#if 0
  /*
   * Set the MTU to the max size
   */
  req.ifr_mtu = FST_MAX_MTU;
  if ( ioctl ( iface->sock, SIOCSIFMTU, &req ) < 0 )
    {
      error_exit ("Error setting interface %s MTU.\n%s\n", name,
		  strerror ( errno ));
    }
#endif
  
  /* Done setting the mode. Close the socket so we can reopen in packet
   * mode.
   */
  if ( close ( iface->sock ) < 0 )
    {
      error_exit ("Error closing interface %s\n.%s\n", name,
		  strerror ( errno ));
    }
}


/*
 *	Close an interface down
 */
static void
fst_uninit ( struct if_info *iface )
{
  struct fstioc_info info;
  struct ifreq req;
  
  if ( close ( iface->sock ) < 0 )
    {
      fprintf ( stderr,"%s: Problem closing %s\n%s\n", progname,
		iface->ifname, strerror ( errno ));
    }
  /*
   * We should be nice and bring the interface down again
   */
  
  if (( iface->sock = socket ( PF_INET, SOCK_DGRAM, IPPROTO_IP )) < 0 )
    error_exit ("Unable to create socket. %s\n", strerror ( errno ));
  
  memset ( &info, 0, sizeof ( info ));
  req.ifr_data = (char *)&info;
  strcpy ( req.ifr_name, iface->ifname );
  
  /* Bring the interface DOWN */
  if ( ioctl ( iface->sock, SIOCGIFFLAGS, &req ) < 0 )
    {
      error_exit ("Error getting interface %s flags.\n%s\n", iface->ifname,
		  strerror ( errno ));
    }
  if (( req.ifr_flags & IFF_UP ))
    {
      req.ifr_flags &= !IFF_UP;
      if ( ioctl ( iface->sock, SIOCSIFFLAGS, &req ) < 0 )
	{
	  error_exit ("Failed to bring interface %s down.\n%s\n", 
		      iface->ifname, strerror ( errno ));
	}
    }
  
  /*
   * Restore the line config the way it was
   */
  req.ifr_data = (char *)&orig_info[iface->index];
  if ( ioctl ( iface->sock, FSTSETCONF, &req ))
    {
      error_exit ("Problem restoring config for interface %s.\n"
		  "%s\n", iface->ifname, strerror ( errno ));
    }
  
  /*
   * Reset the MTU to the original size
   */
  req.ifr_mtu = orig_mtu;
  if ( ioctl ( iface->sock, SIOCSIFMTU, &req ) < 0 )
    {
      error_exit ("Error restoring interface %s MTU.\n%s\n", 
		  iface->ifname, strerror ( errno ));
    }

}

/*
 * 	Open and Bind interface in raw sockets mode.
 *
 *	Print an error message and exit on first error (real programs may wish
 *	to do something better here).
 */
static void
fst_open ( struct if_info *iface )
{
  struct ifreq req;
  struct sockaddr_ll sll;
  int ifindex;
  
  /* 
   * Now that the driver is set into "raw" mode open the packet interface
   */
  
  /* Setup the request structure with the interface name */
  memset ( &req, 0, sizeof ( req ));
  strcpy ( req.ifr_name, iface->ifname );
  
  /* The original FarSync driver marks raw packets with ETH_P_CUST */
  if (( iface->sock = socket ( PF_PACKET, SOCK_RAW,
			       htons(ETH_P_CUST) )) < 0 )
    {
      error_exit ("Failed to create raw socket for interface %s.\n"
		  "%s\n", iface->ifname, strerror ( errno ));
    }
  
  /* Bind socket to a single interface. Start by getting the index */
  if ( ioctl ( iface->sock, SIOCGIFINDEX, &req ) < 0 )
    {
      error_exit ("Error getting interface %s index.\n%s\n", 
		  iface->ifname, strerror ( errno ) );
    }
  ifindex = req.ifr_ifindex;
  
  /* Now bind it */
  bzero ( &sll, sizeof ( sll ));
  sll.sll_family = AF_PACKET;
  sll.sll_hatype = ARPHRD_RAWHDLC;
  sll.sll_ifindex = ifindex;
  
  if ( bind ( iface->sock, (struct sockaddr *)&sll, sizeof ( sll )) < 0 )
    {
      error_exit ("Problem binding to interface %s.\n%s\n", 
		  iface->ifname, strerror ( errno ));
    }
  
  /* Clear internal stats counters */
  iface->rx_pkts  = 0;
  iface->rxbytes = 0;
  iface->rxerrs  = 0;
  iface->tx_pkts  = 0;
  iface->txbytes = 0;
  iface->txerrs  = 0;
  iface->txpause = 0;
}


/*
 *	Print the internal stats for the interface
 */
static void
print_rawhdlc_stats ( struct if_info *iface )
{
  unsigned long rx_duration;
  unsigned long tx_duration;

  rx_duration = iface->rx_xfer_end - iface->rx_xfer_start;
  tx_duration = iface->tx_xfer_end - iface->tx_xfer_start;
  
  printf ("%s: Rx: %lld bytes - %ld errs. ",
	  iface->ifname, iface->rxbytes, iface->rxerrs );
  printf("in %ld s ", rx_duration);
  printf("pps = %ld ", iface->rx_pkts/rx_duration);
  printf("bps = %lld\n", (iface->rxbytes/rx_duration)*8);
  printf ("%s: Tx: %lld bytes - %ld errs & %ld"
	  " pauses. ",
	  iface->ifname, iface->txbytes, iface->txerrs, iface->txpause );
  printf("in %ld s ", tx_duration);
  printf("pps = %ld ", iface->tx_pkts/tx_duration);
  printf("bps = %lld\n", (iface->txbytes/tx_duration)*8);
  printf("%s: Rx Line errors: errs = %ld  drop = %ld  fifo = %ld  frame = %ld\n",
	 iface->ifname,
	 iface->line_errs_end.rx_errs - iface->line_errs_start.rx_errs,
	 iface->line_errs_end.rx_drop - iface->line_errs_start.rx_drop,
	 iface->line_errs_end.rx_fifo - iface->line_errs_start.rx_fifo,
	 iface->line_errs_end.rx_frame - iface->line_errs_start.rx_frame);
  printf("%s: Tx Line errors: errs = %ld  drop = %ld  fifo = %ld  carrier = %ld\n\n",
	 iface->ifname,
	 iface->line_errs_end.tx_errs - iface->line_errs_start.tx_errs,
	 iface->line_errs_end.tx_drop - iface->line_errs_start.tx_drop,
	 iface->line_errs_end.tx_fifo - iface->line_errs_start.tx_fifo,
	 iface->line_errs_end.tx_carrier - iface->line_errs_start.tx_carrier);
  
}

static void
do_mini_stats (struct if_info *iface)
{
  unsigned long rx_duration;
  unsigned long tx_duration;
  time_t time_now;
  char *time_str;
  static int txbytes;

  if (txbytes == iface->txbytes)
    {
      iface->kick_tx++;
      printf("Stutter in rx, kicking into life again\n");
    }
  time_now = time(NULL);
  time_str =  ctime( &time_now);
  time_str[24] = '\0';
  rx_duration = time_now - iface->rx_xfer_start;
  tx_duration = time_now - iface->tx_xfer_start;
  printf ("%s %s: Rx: %lld bytes. ", time_str,
	  iface->ifname, iface->rxbytes );
  printf("bps = %lld", (iface->rxbytes/rx_duration)*8);
  printf (" Tx: %lld bytes. ", iface->txbytes );
  printf("bps = %lld lost = %d\n", (iface->txbytes/tx_duration)*8, iface->lost_packets);
  txbytes = iface->txbytes;
  
}

char *
skip_space (char *str)
{
  while (*str == ' ')
    str++;
  return str;
}

static void
get_dev_stats (struct line_info *istats, char *name)
{
  /*
   * Open the file /proc/net/dev and look for our devices
   * print out the error stats
   */
  
#define LINESIZE 1024
  FILE *fd;
  char line[LINESIZE];
  char *chrptr;
  char *lineptr;
  char tokens[] = " :\n";
  char myname[] = "hdlccxx\0";
  
  fd = fopen("/proc/net/dev", "r");
  if (fd == NULL)
    {
      perror("Cannot open proc file system\n");
      exit (1);
    }
  
  strcpy(myname,name);
  strcat (myname, ":");
  while (fgets(line, LINESIZE, fd) != NULL)
    {
      if (strstr(line, myname) != NULL)
	{
	  /*
	   * feel our way along the line for the data
	   */
	  lineptr = line;
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);   /* iface name */
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);   /* bytes */
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);   /* packets */
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);  
	  istats->rx_errs = atol(chrptr);
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);
	  istats->rx_drop = atol(chrptr);
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);
	  istats->rx_fifo = atol(chrptr);
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, chrptr);
	  istats->rx_frame = atol(chrptr);
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);   /* Compressed */
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);   /* Multicast */
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);   /* Tx bytes */
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);   /* Tx packets */
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);   
	  istats->tx_errs = atol(chrptr);
	  
	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);
	  istats->tx_drop = atol(chrptr);

	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);
	  istats->tx_fifo = atol(chrptr);

	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);   /* Collisions */

	  lineptr = skip_space(lineptr);
	  chrptr = strsep(&lineptr, tokens);
	  istats->tx_carrier = atol(chrptr);
	}
    }
  fclose(fd);
}


/*
 *	Read a raw packet from an open socket returning the length of data
 *      received
 */
static int
read_rawhdlc ( struct if_info *iface, void *buf, int len )
{
  int retval;
  int saved_errno;
  
  retval = recvfrom ( iface->sock, buf, len, 0, NULL, 0 );
  
  if ( retval == -1 )
    {
      saved_errno = errno;
      if (saved_errno != EAGAIN)
        {
          fprintf ( stderr,"%s: %s recvfrom error.%d\n%s\n", progname,
                    iface->ifname,  errno, strerror ( errno ));
          
          /* Restore errno destroyed by the printf in case parent wants to
           * check it */
          errno = saved_errno;
          iface->rxerrs++;
        }
    }
  else
    {
      iface->rx_pkts++;
      iface->rxbytes += retval;
    }
  return retval;
}


/*
 *	Send a raw packet out an open socket returning the length of data
 *      sent
 */
static int
write_rawhdlc ( struct if_info *iface, void *buf, int len )
{
  int retval;
  int saved_errno;
  
  retval = sendto ( iface->sock, buf, len, 0, NULL, 0 );
  //retval = send (iface->sock, buf, len, 0);
  saved_errno = errno;
  //usleep (2);
  /* If the output queue is full, pause a while (100ms) and try again */
  while ( (retval == -1) && (saved_errno == ENOBUFS) )
    {
      iface->txpause++;
      usleep ( 100 );
      retval = sendto ( iface->sock, buf, len, 0, NULL, 0 );
    }

  if ( retval == -1 )
    {
      saved_errno = errno;
      fprintf ( stderr,"%s: %s sendto error.\n%s\n", progname,
		iface->ifname,  strerror ( errno ));
      
      /* Restore errno destroyed by the printf in case parent wants to
       * check it */
      errno = saved_errno;
      
      iface->txerrs++;
    }
  else
    {
      iface->tx_pkts++;
      iface->txbytes += retval;
    }
  return retval;
}



/*
 *	Send a single test packet of data through the link
 */
static bool
send_one_message ( struct if_info *port )
{
  int size;
  int xfer;
  unsigned int *seq_ptr;

  size = txsize;

  seq_ptr = (unsigned int *)&txbuf;
  *seq_ptr = port->send_seq++;

  /* Transmit the buffer */
  xfer = write_rawhdlc ( port, txbuf, size );

  if ( xfer != size )
    {
      if ( xfer >= 0 )
	{
	  printf ("Short transmission. Expected to send %d bytes, write"
		  " reported %d\n", size, xfer );
	}
      return FALSE;
    }
  return TRUE;
}

/*
 *	get a single test packet of data 
 */
static bool
get_one_message ( struct if_info *port )
{
  int size;
  int xfer;
#ifndef FST_TRANSPARENT
  int i;
  unsigned char j;
  unsigned int *seq_ptr;
  int loop_count = 0;
#endif
  
  size = txsize;
  
  /* Read all queued frames */
  while (( xfer = read_rawhdlc ( port, rxbuf, size )) > 0)
    {
      if ( xfer != size )
        {
          if ( xfer >= 0 )
            {
              printf ("%s: Bad reception size. Expected %d bytes, saw %d\n",
                      port->ifname, size, xfer );
            }
          return FALSE;
        }

      /* Buffer is the right size. Check the contents */
      j = 4;
      seq_ptr = (unsigned int *)&rxbuf;
      if (*seq_ptr != port->recv_seq)
        {
          printf("%s: Error in recv seq number. Got %d expected %d sent %ld\n",
                 port->ifname, *seq_ptr, port->recv_seq, port->tx_pkts);
          port->lost_packets += (*seq_ptr - port->recv_seq);
          port->recv_seq = *seq_ptr+1;
        }
      else
        {
          port->recv_seq++;
        }

      for ( i = 4 ; i < size ; i++ )
        {
          if ( rxbuf[ i ] != j)
            {
              printf ("Data error at byte %d. Expected %02x, got %02x\n",
                      i, j, rxbuf[ i ]);
              return FALSE;
            }
          j++;
        }
      loop_count++;
    }
  if(loop_count > 1)
    {
      //printf("Found %d messages in one call to get_one_message\n", loop_count);
    }
  return TRUE;
}


/*
 *	Send TEST_PACKETS packets through the link
 */

static void
test_xfer ( struct if_info *port)
{
  fd_set read_set;
  fd_set write_set;
  time_t start_time;
  time_t last_time;
  int retval;
  int flags;
  int rcv_buf_size;
  int opt_len;
  struct timeval timeout;


  opt_len = sizeof(int);
  flags = getsockopt(port->sock, SOL_SOCKET, SO_RCVBUF, &rcv_buf_size, &opt_len);
  rcv_buf_size *= 16; 
  flags = setsockopt(port->sock, SOL_SOCKET, SO_RCVBUF, &rcv_buf_size, opt_len); 
  if (flags)
    perror("Error setting receive buffer size\n");

  printf ("Reading/writing packets on %s (%d) (%d)\n", 
  	  port->ifname, port->index, port->sock);
  port->rx_pkts = 0;
  port->tx_pkts = 0;
  last_time = time (NULL);
  start_time = time(NULL);
  port->rx_xfer_start = time (NULL);
  port->tx_xfer_start = time (NULL);
  port->lost_packets = 0;
  port->send_seq = 0;
  port->recv_seq = 0;
  
  /*
   * Set the sockets into non blocking mode
   */
  flags = fcntl(port->sock, F_GETFL, 0);
  flags |= O_NONBLOCK;
  retval = fcntl(port->sock, F_SETFL, flags);
  flags = fcntl(port->sock, F_GETFL, 0);
  
  while ( (time (NULL) - start_time ) < duration )
    {
      FD_ZERO(&read_set);
      FD_ZERO(&write_set);
      FD_SET(port->sock, &write_set);
      FD_SET(port->sock, &read_set);

      if ((port->tx_pkts == 0) || (port->kick_tx))
	{
	  send_one_message(port);
	  port->kick_tx = 0;
	}
      timeout.tv_sec = 5;
      timeout.tv_usec = 0;
      retval = select(10, &read_set, &write_set, NULL, &timeout);
      if (retval > 0)
	{
	  if (FD_ISSET(port->sock, &read_set))
	    {
              //printf("Select says readable\n");
	      get_one_message(port);
	    }
#if 1
	  if (FD_ISSET(port->sock, &write_set))
	    {
              //printf("Select says writable\n");
	      if ( (port->send_seq - port->recv_seq) < 5)
		send_one_message(port);
	    }
#endif
	}
      else
	{
	  if (retval == 0)
	    {
	      printf("Select timeout\n");
	    }
	  else
	    {
	      printf("%s: Error in select %d %d\n", 
		     port->ifname, retval ,errno);
	    }
	}
      if ((time(NULL) - last_time) > 10)
	{
	  do_mini_stats(port);
	  last_time = time(NULL);
	}
    }
  port->rx_xfer_end = time(NULL);
  port->tx_xfer_end = time(NULL);
  printf ("%s: Duration complete tx = %ld rx = %ld lost = %d\n", port->ifname,
	  port->tx_pkts, port->rx_pkts, port->lost_packets);
}


/*
 *	Print a simple usage message
 */
static void
usage ( void )
{
  fprintf ( stderr,"usage: %s int1 vpi vci [duration] [tx size]\n", 
	    progname );
}

static void cleanup(void)
{
  printf("Exit handler\n");
  if (if_info[0].ifname != NULL)
    {
      fst_uninit ( &if_info[0]);
    }
  }

void signal_terminate(int signo)
{
  /*
   * This will force the exit handler to run
   */
  printf("Signal handler\n");
  sleep(1);
  exit(1);
}


static char *activation_strings[] = {
        "SDSL Power Down",
        "Send Single Pulses",
        "Initialisation",
        "Processing CL",
        "TR Analog Loop Closed",
        "TR Analog Loop Active",
        "Non TR Analog Loop Closed",
        "Non TR Analog Loop Active",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "GHS Startup",
        "GHS Transaction Auto",
        "Line Probing",
        "GHS Finished",
        "GHS Silence",
        "GHS Startup Failed",
        "GHS Transaction Failed",
        "Line Probing Failed",
        "GHS No Common Mode",
        "Time Out",
        "Idle",
        "Wait STU",
        "LP Finished",
        "Time viloation Host",
        "Invalid",
        "Invalid",
        "Invalid",
        "Wait For CR",
        "Transmit SC",
        "Transmit TC",
        "Transmit FC",
        "Invalid",
        "Invalid",
        "Invalid",
        "Transmit CR",
        "CRSR Silence",
        "Transmit SR",
        "Transmit TR",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "SR Not Detected",
        "TR Not Detected",
        "Data Axception",
        "Not Detected",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "TC Not Detected",
        "FC Not Detected",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Data",
        "Data + CRC Anomaly",
        "Data + LOSW Defect",
        "Invalid",
        "Data + Loop Atten Defect",
        "Invalid",
        "Invalid",
        "Invalid",
        "Data + SNR Margin Defect",
        "Data + ???",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Data + LOSW Failure",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Data CIE",
        "Invalid"
};

#define do_fromtable(P,T,V)			\
	if ( V >= sizeof (T) / sizeof (char *))	\
	    printf ("%sUnknown (%d)\n", P, V );	\
	else					\
	    printf ("%s%s\n", P, T[V] )

#define do_bit(V,B,S)				\
	if ( V & B ) {				\
	    if ( done_bit )			\
		printf (", ");			\
	    printf ( S );			\
	    done_bit = 1;			\
	} else /* Swallow ; */


static void
show_shdsl_state (struct fstioc_info info)
{
  static int first_time = 0;
  static int activation_status = 0;

  if (first_time == 0)
    {
      printf("Tracking the Activation Status:\n");
      first_time++;
    }
  if (activation_status != info.activationStatus)
    {
      do_fromtable("\t  Activation Status:\t", 
                   activation_strings, info.activationStatus);
      activation_status = info.activationStatus;
    }
}

static void
do_info ( struct if_info *port  )
{
	struct fstioc_info info;
        static struct ifreq req;
	int sync_status;

#ifdef GEN_HDLC
	union {
		fr_proto fr;
		cisco_proto cisco;
	} proto;

	bool proto_set = TRUE;
        unsigned int cmd;
#endif
	/* Zero structure before use. Good practice anyway but in this case it
	 * allows for future expansion where various fields (valid for instance)
	 * might be used to indicate an extended structure.
	 */
	memset ( &info, 0, sizeof ( info ));
	req.ifr_data = (char *)&info;
	strcpy ( req.ifr_name, port->ifname );
	if (test_mode == 0)
	  sync_status = 0x40;
	else
	  sync_status = 0x05;
        while ( info.activationStatus < sync_status)
          {
            if ( ioctl ( port->sock, FSTGETCONF, &req ))
              {
                fprintf ( stderr,"%s: Problem getting port information. %s\n",
                          progname, strerror ( errno ));
                exit ( 1 );
              }
            show_shdsl_state(info);
          }
        printf("Loopback data state reached\n");
        sleep(5);
}

/*
 *	Open the two interfaces and transfer TEST_PACKETS packets in each
 *	direction
 */
int
main ( int argc, char *argv[])
{
  pid_t child_pid[4];
  int status;
  int nports;
  int i;
  unsigned char j;

  /* Strip and record the program name for later use */
  if (( progname = strrchr ( argv[0],'/')) != NULL )
    ++progname;
  else
    progname = argv[0];
  
  nports = 1;
  if ( argc < 4 )
    {
      usage ();
      exit ( 1 );
    }
  vpi = atoi(argv[2]);
  printf("Setting vpi to %d\n", vpi);
  vci = atoi(argv[3]);
  printf("Setting vci to %d\n", vci);
  test_mode = atoi(argv[4]);
  printf("Setting test mode to %d\n", test_mode);

  if ( argc == 6 )
    {
      duration = atoi(argv[5]);
      printf("Setting duration to %d\n", duration);
    }
  if ( argc == 7 )
    {
      duration = atoi(argv[5]);
      printf("Setting duration to %d\n", duration);
      txsize = atoi(argv[6]);
      printf("Setting tx size to %d\n", txsize);
    }
  if ( argc > 7 )
    {
      usage ();
      exit ( 1 );
    }

  /* Initialise and Open the interfaces */
  for (i = 0; i < nports; i++)
    {
      if_info[i].index = i;
      fst_init ( &if_info[i], argv[i+1]);
      printf("Initialising port %s\n", if_info[i].ifname);
      fst_open( &if_info[i]);
      do_info(&if_info[i]);
    }

  printf("Interfaces are now up\n");
  sleep (1);
  
  j = 0;
  for (i = 0 ; i < txsize; i++)
    {
      txbuf[i] = j;
      j++;
    }
  
  /*
   * To get full duplex operation we  are going to spawn child 
   * processes to do the xfer.  We then wait for them to expire.
   * and calculate the stats
   */
  printf("Starting child processes\n");
  for ( i = 0; i < nports; i++)
    {
      child_pid[i] = fork();
      if (child_pid[i] == 0)
	{
	  /*
	   * This is the child process
	   */
	  get_dev_stats(&(if_info[i].line_errs_start), if_info[i].ifname);
	  test_xfer ( &if_info[i]);
	  get_dev_stats(&(if_info[i].line_errs_end), if_info[i].ifname);
	  print_rawhdlc_stats ( &if_info[i]);
	  fst_uninit ( &if_info[i]);
	  exit (0);
	}
    }
  /*
   * Parent continues here.
   * Set up signal handlers for program termination
   */
  if (signal(SIGTERM, signal_terminate) == SIG_ERR)
    {
      perror("Can't register signal handler ");
      exit (1);
    }
  if (signal(SIGINT, signal_terminate) == SIG_ERR)
    {
      perror("Can't register signal handler ");
      exit (1);
    }
  
  /*
   * Set up exit handlers to clean up
   */
  
  printf("Setting up exit handlers\n");
  if (atexit(cleanup) != 0)
    {
      printf("Can't register exit handler\n");
      exit (1);
    }
  
  /* 
   *If we get here we are the parent
   * so wait for the children to exit
   */
  
  for ( i = 0; i < nports; i++)
    {
      child_pid[i] = waitpid(child_pid[i], &status, 0);
    }
  exit ( 0 );
}

/* End $Id: fardsltest.c,v 1.2 2009/07/10 08:23:24 kevinc Exp $ */
