/*
 *	FarSync OEM driver for Linux
 *
 *	Simple port monitoring evaluation tool
 *
 *	Copyright (C) 2003 FarSite Communications Ltd.
 *	www.farsite.co.uk
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	File:	farmon.c
 *	Author:	Kevin Curtis	<kevin.curtis@farsite.co.uk>
 */
/*	Revision History
 *	$Log: farmon.c,v $
 *	Revision 1.2  2009/07/10 08:23:24  kevinc
 *	Merging 1.09.04 and async branches
 *	
 *	Revision 1.1.1.1  2009/05/06 12:45:30  kevinc
 *	
 *	
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>
#include <fcntl.h>

#include "farsync.h"


/* Define FALSE & TRUE 
 */
#ifndef TRUE
# define TRUE   1
# define FALSE  0
typedef int bool;
#endif

#define NUM_INTERFACES	16	/* Max no of interfaces we will support */

#define TEST_PACKETS	100	/* Number of packets to test with */
#define TEST_SIZE	1400	/* Default size of test packet */

/*
 * Change these #defines to suit your interface requirements
 */
#define FST_MAX_MTU     32*1024 /* Max frame size */
#define FST_LINE_TYPE   X21     /* E1, V24 or V35 as alternatives */
#define FST_LINE_SPEED  8192000 /* 8 Meg */

#define RX_OK        0
#define RX_ERROR     1
#define RX_CARRIER   2

/* We're passing data around in global structures for simplicity.
 * OK for simple programs but we should probably be doing a lot more parameter
 * passing in larger programs.
 */
static char *progname;
static int txsize = TEST_SIZE;
static unsigned char rxbuf[ FST_MAX_MTU ];

/* Structure for the line stats */
struct line_info {
                                /* Interface stats */
        long    rx_errs;
        long    rx_drop;
        long    rx_fifo;
        long    rx_frame;
        long    tx_errs;
        long    tx_drop;
        long    tx_fifo;
        long    tx_carrier;
};


/* Structure to contain per interface information */
struct if_info {
        int     index;
	int     sock;		/* Socket handle */
	char   *ifname;		/* Interface name */

	long    rx_pkts;	/* Input stats */
	unsigned long long    rxbytes;
	long    rxerrs;

	long    tx_pkts;	/* Output stats */
	unsigned long long    txbytes;
	long    txerrs;
	long    txpause;	/* Delays due to full ouput queue */

        time_t  tx_xfer_start;  /* Tx Start time */
        time_t  tx_xfer_end;    /* and Tx End time */
        time_t  rx_xfer_start;  /* Tx Start time */
        time_t  rx_xfer_end;    /* and Tx End time */

        struct line_info line_errs_start;
        struct line_info line_errs_end;
        unsigned int send_seq;
        unsigned int recv_seq;
        unsigned int lost_packets;

} if_info[ NUM_INTERFACES ];


void dump_hex(void* Data, unsigned int NumBytes)
{
  char* pCharData=(char*)Data;
  unsigned int  i,j;

  for (i = 0; i < NumBytes; i += 16)
    {
      printf("%08x: ", pCharData+i);
      // Output the hex bytes
      for (j = i; j < (i+16); j++)
        {
          if (j < NumBytes)
            { 
              printf("%02x ",(unsigned int)((unsigned char)*(pCharData+j)));
            }
          else
            {
              printf("   ");
            }
          /*fi*/
        }
      /*rof*/  

      printf("  ");

      // Output the ASCII bytes
      for (j = i; j < (i+16); j++)
        {
          if (j < NumBytes)
            {
              char c = *(pCharData+j);

              if (c < ' ' || c > 'z')
                {
                  c = '.';
                }
              /*fi*/
              printf("%c", (unsigned int)c);
            }
          else
            {
              printf(" ");
            }
          /*fi*/
        }
      /*rof*/
      printf("\r\n");
    }
  /*rof*/
}

/*
 *      Print an error message and exit
 */
static void
error_exit ( char *msg, ...)
{
  va_list ap;
  
  va_start ( ap, msg );
  fprintf ( stderr,"%s: ", progname );
  vfprintf ( stderr, msg, ap );
  va_end ( ap );
  
  exit ( 1 );
}



/*
 *	Close an interface down
 */
static void
fst_uninit ( struct if_info *iface )
{
  struct fstioc_info info;
  struct ifreq req;
  int monitor_flag = 0;

  /* 
   * It's important to set monitoring off in the driver
   * to avoid performance impact
   */
  memset ( &info, 0, sizeof ( info ));
  strcpy ( req.ifr_name, iface->ifname );
  monitor_flag = 0;
  req.ifr_data = (char *) &monitor_flag;
  if ( ioctl ( iface->sock, FSTSETMON, &req ))
    {
      error_exit ("Problem setting monitor mode %s.\n"
                  "%s\n", iface->ifname, strerror ( errno ));
    }

  if ( close ( iface->sock ) < 0 )
    {
      fprintf ( stderr,"%s: Problem closing %s\n%s\n", progname,
		iface->ifname, strerror ( errno ));
    }

}

/*
 *      Check Signal State
 */
static int
check_state ( int sock, char * name )
{
  /*
   * Select has signalled an exception on the socket
   * Check the signal states
   */
  struct ifreq req;
  struct fstioc_status state;

  strcpy ( req.ifr_name, name );
  req.ifr_data = (char *)&state;
  if ( ioctl ( sock, FSTGSTATE, &req ))
    {
      printf ("Problem getting status for interface"
	      " %s.\n%s\n", name, strerror ( errno ));
    }
  printf("%s: Signal state is %x\n", name, state.carrier_state); 
  printf("%s: Tx Queue Length is %x\n", name, state.txq_length);   
  return state.carrier_state;
}

static void
fst_init ( struct if_info *iface, char *name )
{
  struct ifreq req;
  int monitor_flag = 0;

  
  /* Sanity check and record the interface name */
  if ( strlen ( name ) >= IFNAMSIZ )
    error_exit ("Interface name %s too long\n", name );
  
  iface->ifname = name;
  
  if (( iface->sock = socket ( PF_INET, SOCK_DGRAM, IPPROTO_IP )) < 0 )
    error_exit ("Unable to create socket. %s\n", strerror ( errno ));
  
  /* Setup the request structure with the interface name */
  memset ( &req, 0, sizeof ( req ));
  strcpy ( req.ifr_name, name );
      
  monitor_flag = 1;
  req.ifr_data = (char *) &monitor_flag;
  if ( ioctl ( iface->sock, FSTSETMON, &req ))
    {
      error_exit ("Problem setting monitor mode %s.\n"
                  "%s\n", name, strerror ( errno ));
    }
  if ( close ( iface->sock ) < 0 )
    {
      error_exit ("Error closing interface %s\n.%s\n", name,
		  strerror ( errno ));
    }
}

/*
 * 	Open and Bind interface in raw sockets mode.
 *
 *	Print an error message and exit on first error (real programs may wish
 *	to do something better here).
 */
static void
fst_open ( struct if_info *iface )
{
  struct ifreq req;
  struct sockaddr_ll sll;
  int ifindex;
  
  /* 
   * Now that the driver is set into "raw" mode open the packet interface
   */
  
  /* Setup the request structure with the interface name */
  memset ( &req, 0, sizeof ( req ));
  strcpy ( req.ifr_name, iface->ifname );
  
  /* The FarSync driver marks monitor packets with ETH_P_DIAG */
  if (( iface->sock = socket ( PF_PACKET, SOCK_RAW,
			       htons(ETH_P_DIAG) )) < 0 )
    {
      error_exit ("Failed to create raw socket for interface %s.\n"
		  "%s\n", iface->ifname, strerror ( errno ));
    }
  
  /* Bind socket to a single interface. Start by getting the index */
  if ( ioctl ( iface->sock, SIOCGIFINDEX, &req ) < 0 )
    {
      error_exit ("Error getting interface %s index.\n%s\n", 
		  iface->ifname, strerror ( errno ) );
    }
  ifindex = req.ifr_ifindex;
  
  /* Now bind it */
  bzero ( &sll, sizeof ( sll ));
  sll.sll_family = AF_PACKET;
  sll.sll_hatype = ARPHRD_RAWHDLC;
  sll.sll_ifindex = ifindex;
  if ( bind ( iface->sock, (struct sockaddr *)&sll, sizeof ( sll )) < 0 )
    {
      error_exit ("Problem binding to interface %s.\n%s\n", 
		  iface->ifname, strerror ( errno ));
    }
  
  /* Clear internal stats counters */
  iface->rx_pkts  = 0;
  iface->rxbytes = 0;
  iface->rxerrs  = 0;
  iface->tx_pkts  = 0;
  iface->txbytes = 0;
  iface->txerrs  = 0;
  iface->txpause = 0;
}

/*
 *	Read a raw packet from an open socket returning the length of data
 *      received
 */
static int
read_rawhdlc ( struct if_info *iface, void *buf, int len )
{
  int retval;
  int saved_errno;
  
  retval = recvfrom ( iface->sock, buf, len, 0, NULL, 0 );
  
  if ( retval == -1 )
    {
      saved_errno = errno;
      if (saved_errno != EAGAIN)
        {
          fprintf ( stderr,"%s: %s recvfrom error.%d\n%s\n", progname,
                    iface->ifname,  errno, strerror ( errno ));
          
          /* Restore errno destroyed by the printf in case parent wants to
           * check it */
          errno = saved_errno;
          iface->rxerrs++;
        }
    }
  else
    {
      iface->rx_pkts++;
      iface->rxbytes += retval;
    }
  return retval;
}



/*
 *	get any monitor frames
 */
static int
get_one_message ( struct if_info *port )
{
  int size;
  int xfer;
  unsigned int seq_no;
  int loop_count = 0;
  struct fstioc_mon *header;
  char *data_ptr;
  
  size = txsize;
  header = (struct fstioc_mon*) &rxbuf;
  data_ptr = (char*) (&rxbuf)+sizeof(struct fstioc_mon);
  
  /* Read all queued frames */
  while (( xfer = read_rawhdlc ( port, rxbuf, size )) >= 0)
    {
      seq_no = header->sequence;
      if (seq_no != port->recv_seq)
        {
          printf("%s: **** Error in recv seq number. Got %d expected %d sent %ld\n",
                 port->ifname, seq_no, port->recv_seq, port->tx_pkts);
          port->lost_packets += (seq_no - port->recv_seq);
          port->recv_seq = seq_no+1;
	  //exit(1);  /* Should we exit or not ? */
        }
      else
        {
          port->recv_seq++;
        }
      if (header->tx_rx_ind == FST_MON_TX)
        printf("\n%s TX: Seq: %d: Time %d: Len %d\n", if_info->ifname, 
               header->sequence, header->timestamp, header->length);
      else
        printf("\n%s RX: Seq: %d: Time %d: Len %d\n", if_info->ifname, 
               header->sequence, header->timestamp, header->length);
      dump_hex(data_ptr, header->length);

      loop_count++;
    }
  if(loop_count > 1)
    {
      //printf("Found %d messages in one call to get_one_message\n", 
      //loop_count);
    }
  return RX_OK;
}


/*
 *	Send TEST_PACKETS packets through the link
 */

static void
test_xfer ( struct if_info *port)
{
  fd_set read_set;
  fd_set excpt_set;
  time_t start_time;
  time_t last_time;
  int retval;
  int flags;
  struct timeval timeout;
  int i =1;
  
  port->rx_pkts = 0;
  port->tx_pkts = 0;
  last_time = time (NULL);
  start_time = time(NULL);
  port->rx_xfer_start = time (NULL);
  port->tx_xfer_start = time (NULL);
  port->lost_packets = 0;
  port->send_seq = 0;
  port->recv_seq = 0;
  
  /*
   * Set the sockets into non blocking mode
   */
  flags = fcntl(port->sock, F_GETFL, 0);
  flags |= O_NONBLOCK;
  retval = fcntl(port->sock, F_SETFL, flags);
  flags = fcntl(port->sock, F_GETFL, 0);
  
  while (i)
    {
      FD_ZERO(&read_set);
      FD_ZERO(&excpt_set);
      FD_SET(port->sock, &read_set);
      FD_SET(port->sock, &excpt_set);

      timeout.tv_sec = 1;
      timeout.tv_usec = 0;
      retval = select(10, &read_set, NULL, &excpt_set, &timeout);
      if (retval > 0)
	{
	  if (FD_ISSET(port->sock, &read_set))
	    {
	      //printf("Select says readable\n");
	      if (get_one_message(port) == RX_CARRIER)
		break;
	    }
	  if(FD_ISSET(port->sock, &excpt_set))
	  {
	    printf("Select says exception\n");
	    check_state(port->sock, port->ifname);
	  }
	}
      else
	{
	  if (retval == 0)
	    {
	      //printf("Timeout in select\n");
	    }
	  else
	    {
	      printf("%s: Error in select %d %d\n", 
		     port->ifname, retval ,errno);
	    }
	}
    }
  port->rx_xfer_end = time(NULL);
  port->tx_xfer_end = time(NULL);
  printf ("%s: Duration complete tx = %ld rx = %ld lost = %d\n", port->ifname,
	  port->tx_pkts, port->rx_pkts, port->lost_packets);
}


/*
 *	Print a simple usage message
 */
static void
usage ( void )
{
  fprintf ( stderr,"usage: %s device\n", progname );
  fprintf ( stderr,"e.g. %s sync0\n", progname);
}
static void cleanup(void)
{
  printf("Exit handler\n");
  if (if_info[0].ifname != NULL)
    {
      fst_uninit ( &if_info[0]);
    }
}

void signal_terminate(int signo)
{
  /*
   * This will force the exit handler to run
   */
  printf("Signal handler\n");
  sleep(1);
  exit(1);
}


/*
 *	Open the two interfaces and transfer TEST_PACKETS packets in each
 *	direction
 */
int
main ( int argc, char *argv[])
{
  int nports;

  /* Strip and record the program name for later use */
  if (( progname = strrchr ( argv[0],'/')) != NULL )
    ++progname;
  else
    progname = argv[0];

  if (argc !=2)
    {
      usage();
      exit(1);
    }
  nports = 1;  
  /* Initialise and Open the interfaces */
  fst_init ( &if_info[0], argv[1]);
  if_info[0].index = 0;
  printf("Opening port %s to monitor\n", if_info[0].ifname);
  fst_open( &if_info[0]);
  sleep (1);
  
  /*
   * Set up signal handlers for program termination
   */
  printf("Setting up Signal and Exit handlers\n");
  if (signal(SIGTERM, signal_terminate) == SIG_ERR)
    {
      perror("Can't register signal handler ");
      exit (1);
    }
  if (signal(SIGINT, signal_terminate) == SIG_ERR)
    {
      perror("Can't register signal handler ");
      exit (1);
    }
  
  /*
   * Set up exit handlers to clean up
   */
  
  if (atexit(cleanup) != 0)
    {
      printf("Can't register exit handler\n");
      exit (1);
    }
  
  /*
   * Just keep posting reads to get trace data
   */
  test_xfer(&if_info[0]); /* Monitor Line zero only for now */
  exit ( 0 );
}

/* End $Id: farmon.c,v 1.2 2009/07/10 08:23:24 kevinc Exp $ */
