/*
 *	FarSync X21 driver for Linux
 *
 *	Null modem cable loopback test
 *
 *	Copyright (C) 2001 FarSite Communications Ltd.
 *	www.farsite.co.uk
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	File:	fartest.c
 *	Author:	R.J.Dunlop	<bob.dunlop@farsite.co.uk>
 */
/*	Revision History
 *	$Log: fartest.c,v $
 *	Revision 1.2  2009/07/10 08:23:24  kevinc
 *	Merging 1.09.04 and async branches
 *	
 *	Revision 1.1.1.1  2009/05/06 12:45:30  kevinc
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <stdarg.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/if.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>

/* Defining GEN_HDLC includes code for detecting and using Krzysztof Halasa's
 * generic HDLC driver module.  Only define it if you have the Krzysztof's
 * modified headers.  The resultant binary will then be able to run on both
 * patched and unpatched kernels.
 */
#ifdef GEN_HDLC
#include <linux/hdlc.h>
#include <linux/sockios.h>
#endif

#include "farsync.h"


/* Pet peeve, why something like this never went into a consistent standard
 * header. Or even "enum bool { FALSE, TRUE };"
 */
#ifndef TRUE
# define TRUE   1
# define FALSE  0
  typedef int bool;
#endif

#define NUM_INTERFACES	2	/* This is a loopback test */

#define TEST_PACKETS	100	/* Number of packets to test with */
#define TEST_SIZE	1000	/* Size of test packet */


/* We're passing data around in global structures for simplicity.
 * OK for simple programs but we should probably be doing a lot more parameter
 * passing in larger programs.
 */
static char *progname;
static bool verbose = FALSE;
static int txcount = TEST_PACKETS;

#ifdef GEN_HDLC
static bool is_gen_hdlc = FALSE;
#endif

static int mtu;			/* Detected MTU. This is important when using
				 * a raw interface as the system won't frag
				 * for us.
				 */

/* Structure to contain per interface information */
struct if_info {
	int sock;		/* Socket handle */
	char *ifname;		/* Interface name */

	long  rxpkts;		/* Input stats */
	long  rxbytes;
	long  rxerrs;

	long  txpkts;		/* Output stats */
	long  txbytes;
	long  txerrs;
	long  txpause;		/* Delays due to full ouput queue */

} if_info[ NUM_INTERFACES ];



/*
 *      Print an error message and exit
 */
static void
error_exit ( char *msg, ...)
{
	 va_list ap;

	 va_start ( ap, msg );
	 fprintf ( stderr,"%s: ", progname );
	 vfprintf ( stderr, msg, ap );
	 va_end ( ap );

	 exit ( 1 );
}


/*
 * 	Open an interface and set to raw HDLC mode.
 *
 *	Print an error message and exit on first error (real programs may wish
 *	to do something better here).
 *	If this is the first socket opened and we have support compiled in
 *	check to see if we should use the generic-HDLC module ioctls.
 */
static void
setup_rawhdlc ( struct if_info *iface, char *name )
{
	struct fstioc_info info;
	struct ifreq req;
	struct sockaddr_ll sll;
	int ifindex;

#ifdef GEN_HDLC
	static bool probed_gen_hdlc = FALSE;
#endif


	/* Sanity check and record the interface name */
	if ( strlen ( name ) >= IFNAMSIZ )
	    error_exit ("Interface name %s too long\n", name );

	iface->ifname = name;

	if (( iface->sock = socket ( PF_INET, SOCK_DGRAM, IPPROTO_IP )) < 0 )
	    error_exit ("Unable to create socket. %s\n", strerror ( errno ));

	/* Setup the request structure with the interface name */
	memset ( &req, 0, sizeof ( req ));
	strcpy ( req.ifr_name, name );


	/* Check the driver is set to the "raw" operating mode.  This is driver
	 * specific code using either the FarSync driver interface or the soon
	 * to be (hopefully) common driver interface "generic-HDLC".
	 * You could skip all of the code from here down to END SKIP if you
	 * chose to rely on some external configuration setup utility such
	 * as farutil.  We rely on farutil for line type and clocks anyway.
	 */
#ifdef GEN_HDLC
	if ( ! probed_gen_hdlc )
	{
	    /* Probe to see if the driver supports the "generic HDLC" system */
	    memset ( &info, 0, sizeof ( info ));
	    req.ifr_data = (char *)&info;
	    if ( ioctl ( iface->sock, FSTGETCONF, &req ))
	    {
		error_exit ("Problem getting information for interface %s.\n"
				"%s\n", name, strerror ( errno ));
	    }
	    is_gen_hdlc = ( info.proto == FST_GEN_HDLC );
	    probed_gen_hdlc = TRUE;

	    if ( verbose && is_gen_hdlc )
	    	printf ("System is using generic-HDLC interface\n");
	}

	/* Switch interface to raw mode if not already set */
	if ( is_gen_hdlc )
	{
	    req.ifr_settings.type = IF_GET_PROTO;
	    req.ifr_settings.data_length = 0;
	    if ( ioctl ( iface->sock, SIOCDEVICE, &req ))
	    {
		error_exit ("Problem getting protocol information for"
				" interface %s.\n%s\n",
				name, strerror ( errno ));
		exit ( 1 );
	    }

	    if ( req.ifr_settings.type != IF_PROTO_HDLC )
	    {
	    	if ( verbose )
		    printf ("Setting HDLC mode on %s\n", name );

		req.ifr_settings.type = IF_PROTO_HDLC;
		req.ifr_settings.data_length = 0;

		if ( ioctl ( iface->sock, SIOCDEVICE, &req ))
		{
		    error_exit ("Problem setting protocol for interface %s.\n"
		    		"%s\n", name, strerror ( errno ));
		}
	    }
	}
	else	/* Tricksy drop through to non-ifdef case */
#endif
	{
	    /* Check and set raw HDLC mode using the older (non generic-HDLC)
	     * ioctl
	     */
	    memset ( &info, 0, sizeof ( info ));
	    req.ifr_data = (char *)&info;
	    if ( ioctl ( iface->sock, FSTGETCONF, &req ))
	    {
		error_exit ("Problem getting protcol information for interface"
				" %s.\n%s\n", name, strerror ( errno ));
	    }

	    if ( info.proto != FST_RAW )
	    {
	    	if ( verbose )
		    printf ("Setting raw mode on %s\n", name );

	        info.proto = FST_RAW;
		info.valid = FSTVAL_PROTO;

		req.ifr_data = (char *)&info;
		if ( ioctl ( iface->sock, FSTSETCONF, &req ))
		{
		    error_exit ("Problem setting protocol for interface %s.\n"
		    		"%s\n", name, strerror ( errno ));
		}
	    }
	}

	/* Bring the interface UP if necessary */
	if ( ioctl ( iface->sock, SIOCGIFFLAGS, &req ) < 0 )
	{
	    error_exit ("Error getting interface %s flags.\n%s\n", name,
	    						strerror ( errno ));
	}

	if ( ! ( req.ifr_flags & IFF_UP ))
	{
	    if ( verbose )
		printf ("Bringing interface %s up\n", name );

	    req.ifr_flags |= IFF_UP;
	    if ( ioctl ( iface->sock, SIOCSIFFLAGS, &req ) < 0 )
	    {
		error_exit ("Failed to bring interface %s up.\n%s\n", name,
							strerror ( errno ));
	    }
	}
	else
	{
	  error_exit ("Please make sure interfaces are down before running this test - %s\n", name);
	}

	/* END SKIP (code to setup "raw" protocol) */


	/* Record interface MTU.
	 * Since we are using a raw interface the system won't frag for us. */
	if ( ioctl ( iface->sock, SIOCGIFMTU, &req ) < 0 )
	{
		error_exit ("Error getting interface %s MTU.\n%s\n", name,
							strerror ( errno ));
	}
	if ( mtu == 0 )
	{
	    mtu = req.ifr_mtu;
	    if ( verbose )
		printf ("Interface %s MTU is %d\n", name, mtu );
	}
	else if ( req.ifr_mtu < mtu )
	{
	    mtu = req.ifr_mtu;
	    if ( verbose )
	    	printf ("Interface %s reduces MTU to %d\n", name, mtu );
	}


	/* Done setting the mode. Close the socket so we can reopen in packet
	 * mode.
	 */
	if ( close ( iface->sock ) < 0 )
	{
	    error_exit ("Error closing interface %s\n.%s\n", name,
							strerror ( errno ));
	}


	/* Now that the driver is set into "raw" mode open the packet interface
	 */
#if GEN_HDLC
	if ( is_gen_hdlc )
	{
	    /* ETH_P_IP is how packets are marked by the generic HDLC driver in
	     * raw HDLC mode */
	    if (( iface->sock = socket ( PF_PACKET, SOCK_RAW,
						htons(ETH_P_IP) )) < 0 )
	    {
		error_exit ("Failed to create raw socket for interface %s.\n"
				"%s\n", name, strerror ( errno ));
	    }
	}
	else	/* Drop though to non-generic driver case */
#endif
	{
	    /* The original FarSync driver marks raw packets with ETH_P_CUST */
	    if (( iface->sock = socket ( PF_PACKET, SOCK_RAW,
						htons(ETH_P_CUST) )) < 0 )
	    {
		error_exit ("Failed to create raw socket for interface %s.\n"
				"%s\n", name, strerror ( errno ));
	    }
	}

        /* Bind socket to a single interface. Start by getting the index */
	if ( ioctl ( iface->sock, SIOCGIFINDEX, &req ) < 0 )
	{
	    error_exit ("Error getting interface %s index.\n%s\n", name,
	    					strerror ( errno ) );
	}
	ifindex = req.ifr_ifindex;
#if 0
	/* An almost useless piece of info */
	if ( verbose )
	    printf ("Interface %s index is %d\n", name, ifindex );
#endif

	/* Now bind it */
	bzero ( &sll, sizeof ( sll ));
	sll.sll_family = AF_PACKET;
	sll.sll_hatype = ARPHRD_RAWHDLC;
	sll.sll_ifindex = ifindex;

	if ( bind ( iface->sock, (struct sockaddr *)&sll, sizeof ( sll )) < 0 )
	{
	    error_exit ("Problem binding to interface %s.\n%s\n", name,
						strerror ( errno ));
	}

	/* Clear internal stats counters */
	iface->rxpkts  = 0;
	iface->rxbytes = 0;
	iface->rxerrs  = 0;
	iface->txpkts  = 0;
	iface->txbytes = 0;
	iface->txerrs  = 0;
	iface->txpause = 0;
}


/*
 *	Close an interface down
 */
static void
close_rawhdlc ( struct if_info *iface, char *name )
{
	struct fstioc_info info;
	struct ifreq req;

	if ( verbose )
	    printf ("Closing %s\n", iface->ifname );

	if ( close ( iface->sock ) < 0 )
	{
	    fprintf ( stderr,"%s: Problem closing %s\n%s\n", progname,
	    				iface->ifname, strerror ( errno ));
	}
	/*
	 * We should be nice and bring the interface down again
	 */

	if (( iface->sock = socket ( PF_INET, SOCK_DGRAM, IPPROTO_IP )) < 0 )
	    error_exit ("Unable to create socket. %s\n", strerror ( errno ));

	memset ( &info, 0, sizeof ( info ));
	req.ifr_data = (char *)&info;
	strcpy ( req.ifr_name, name );

	/* Bring the interface DOWN */
	if ( ioctl ( iface->sock, SIOCGIFFLAGS, &req ) < 0 )
	{
	    error_exit ("Error getting interface %s flags.\n%s\n", name,
	    						strerror ( errno ));
	}
	if (( req.ifr_flags & IFF_UP ))
	{
	    if ( verbose )
		printf ("Bringing interface %s down\n", name );

	    req.ifr_flags &= !IFF_UP;
	    if ( ioctl ( iface->sock, SIOCSIFFLAGS, &req ) < 0 )
	    {
		error_exit ("Failed to bring interface %s down.\n%s\n", name,
							strerror ( errno ));
	    }
	}
}


/*
 *	Print the internal stats for the interface
 */
static void
print_rawhdlc_stats ( struct if_info *iface )
{
	printf ("Transfer stats for %s\n", iface->ifname );
	printf ("Received    %6ld bytes in %4ld packets with %ld errors.\n",
		iface->rxbytes, iface->rxpkts, iface->rxerrs );
	printf ("Transmitted %6ld bytes in %4ld packets with %ld errors & %ld"
		" pauses.\n",
		iface->txbytes, iface->txpkts, iface->txerrs, iface->txpause );
}


/*
 *	Read a raw packet from an open socket returning the length of data
 *      received
 */
static int
read_rawhdlc ( struct if_info *iface, void *buf, int len )
{
        int retval;
	int saved_errno;
	 
	retval = recvfrom ( iface->sock, buf, len, 0, NULL, 0 );

        if ( retval == -1 )
	{
	    saved_errno = errno;
	    fprintf ( stderr,"%s: %s recvfrom error.\n%s\n", progname,
	    				iface->ifname,  strerror ( errno ));

	    /* Restore errno destroyed by the printf in case parent wants to
	     * check it */
	    errno = saved_errno;

	    iface->rxerrs++;
	}
	else
	{
	    iface->rxpkts++;
	    iface->rxbytes += retval;
	}

	return retval;
}


/*
 *	Send a raw packet out an open socket returning the length of data
 *      sent
 */
static int
write_rawhdlc ( struct if_info *iface, void *buf, int len )
{
        int retval;
	int saved_errno;
	 
	retval = sendto ( iface->sock, buf, len, 0, NULL, 0 );

	/* If the output queue is full, pause a while (10ms) and try again */
	while ( retval == -1 && errno == ENOBUFS )
	{
	    iface->txpause++;
	    usleep ( 10000 );
	    retval = sendto ( iface->sock, buf, len, 0, NULL, 0 );
	}

	if ( retval == -1 )
	{
	    saved_errno = errno;
	    fprintf ( stderr,"%s: %s sendto error.\n%s\n", progname,
	    				iface->ifname,  strerror ( errno ));

	    /* Restore errno destroyed by the printf in case parent wants to
	     * check it */
	    errno = saved_errno;

	    iface->txerrs++;
	}
	else
	{
	    iface->txpkts++;
	    iface->txbytes += retval;
	}

	return retval;
}


/*
 *	Send a single test packet of random data through the link
 */
static bool
test_xfer_one ( struct if_info *from, struct if_info *to )
{
	unsigned char txbuf[ TEST_SIZE ];
	unsigned char rxbuf[ TEST_SIZE ];
	int i;
	int size;
	int xfer;

	size = ( mtu < TEST_SIZE ) ? mtu : TEST_SIZE;

	/* Fill the buffer with pseudo random data or incrementing data */
	for ( i = 0 ; i < size ; i++ )
	  //txbuf[ i ] = rand ();
	    txbuf[ i ] = i;

	/* Transmit the buffer */
	xfer = write_rawhdlc ( from, txbuf, size );
	if ( xfer != size )
	{
	    if ( xfer >= 0 )
	    {
	    	printf ("Short transmission. Expected to send %d bytes, write"
					" reported %d\n", size, xfer );
	    }
	    return FALSE;
	}

	/* Read the result */
	xfer = read_rawhdlc ( to, rxbuf, TEST_SIZE );
	if ( xfer != size )
	{
	    if ( xfer >= 0 )
	    {
	    	printf ("Bad reception size. Expected %d bytes, saw %d\n",
					size, xfer );
	    }
	    return FALSE;
	}

	/* Buffer is the right size. Check the contents */
	for ( i = 0 ; i < size ; i++ )
	{
	    if ( txbuf[ i ] != rxbuf[ i ])
	    {
	    	printf ("Data error at byte %d. Expected %02x, got %02x\n",
					i, txbuf[ i ], rxbuf[ i ]);
	    	return FALSE;
	    }
	}

	return TRUE;
}


/*
 *	Send TEST_PACKETS packets through the link
 */
static void
test_xfer ( struct if_info *from, struct if_info *to )
{
	time_t last_report = 0;
	time_t time_now;
	int pkt;

	printf ("Sending %d packets from %s to %s\n", txcount,
					from->ifname, to->ifname );

	for ( pkt = 0 ; pkt < txcount ; pkt++ )
	{
	    /* Update a packet count display once per second. On slow line 100
	     * packets can take a time
	     */
	    if ( verbose && ( time_now = time ( NULL )) != last_report )
	    {
	    	fprintf ( stderr,"Packet %d\r", pkt + 1 );
		last_report = time_now;
	    }

	    /* Send one test packet */
	    if ( ! test_xfer_one ( from, to ))
		break;
	}

	if ( pkt == txcount )
	    printf ("Transfer completed OK\n");
	else
	    printf ("Transfer failed after %d packets\n", pkt );
}


/*
 *	Print a simple usage message
 */
static void
usage ( void )
{
	fprintf ( stderr,"usage: %s [-v] interface1 interface2 [count]\n", 
		  progname );
}


/*
 *	Open the two interfaces and transfer TEST_PACKETS packets in each
 *	direction
 */
int
main ( int argc, char *argv[])
{
	/* Strip and record the program name for later use */
	if (( progname = strrchr ( argv[0],'/')) != NULL )
	    ++progname;
	else
	    progname = argv[0];

	/* Set verbose mode and skip over flag */
	if ( argc >= 2 && strcmp ( argv[1],"-v") == 0 )
	{
	    verbose = TRUE;
	    --argc;
	    ++argv;
	}

	/* Remaining two args should be the interface names */
	if ( argc < 3 )
	{
	    usage ();
	    exit ( 1 );
	}

	if ( argc == 4 )
	{
	  txcount = atoi(argv[3]);
	  printf("Setting tx count to %d\n", txcount);
	}

	/* Open the interfaces and set to raw mode */
	setup_rawhdlc ( &if_info[0], argv[1]);
	setup_rawhdlc ( &if_info[1], argv[2]);

	printf("Interfaces are now be up\n");
        sleep (1);

	/* Test data transfer in each direction */
	test_xfer ( &if_info[0], &if_info[1]);
	test_xfer ( &if_info[1], &if_info[0]);

	/* Print some additional stats */
	if ( verbose )
	{
	    print_rawhdlc_stats ( &if_info[0]);
	    print_rawhdlc_stats ( &if_info[1]);
	}

	/* Orderly shutdown (not strictly necessary) */
	close_rawhdlc ( &if_info[0], argv[1]);
	close_rawhdlc ( &if_info[1], argv[2]);
	exit ( 0 );
}

/* End $Id: fartest.c,v 1.2 2009/07/10 08:23:24 kevinc Exp $ */
