/*
 *	FarSync X21 driver for Linux
 *
 *	Actually sync driver for X.21, V.35 and V.24 on FarSync T-series cards
 *
 *	Copyright (C) 2001 FarSite Communications Ltd.
 *	www.farsite.co.uk
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *
 *	File:	farutil.c
 *	Author:	R.J.Dunlop	<bob.dunlop@farsite.co.uk>
 */
/*	Revision History
 *	$Log: farutil.c,v $
 *	Revision 1.3  2009/12/17 11:31:22  kevinc
 *	Make the release version 1.09.05 - build 143
 *	
 *	Revision 1.2  2009/07/10 08:23:25  kevinc
 *	Merging 1.09.04 and async branches
 *	
 *	Revision 1.1.1.1  2009/05/06 12:45:30  kevinc
 *	
 */

#include "farutil.h"

/*
 *	Allocate the download buffer
 */
static void
alloc_buf ( void )
{
	buf = (struct fstioc_write *) malloc ( sizeof ( struct fstioc_write )
			 				+ CBUFSIZ );
	if ( buf == NULL )
	{
	    fprintf ( stderr,"%s: Failed to allocate download buffer\n",
							progname );
	    exit ( 1 );
	}
}

static void
write_buf ( unsigned long size, unsigned long offset, char *name )
{
	buf->size = size;
	buf->offset = offset;

	req.ifr_data = (char *)buf;
	if ( ioctl ( sock, FSTWRITE, &req ))
	{
	    fprintf ( stderr,"%s: Problem writing %s to card. %s\n",
					progname, name, strerror ( errno ));
	    exit ( 1 );
	}
}

static void
zero_memory ( void )
{
	memset ( buf->data, 0, CBUFSIZ );
	write_buf ( CBUFSIZ, 0, "Zeroed memory");
}


/*
 * 	Download one .CDE file to card
 */
static void
load_file ( char *filename, char *name, bool load_header, int offset )
{
	struct exehdr exec;             /* Fixed part of exec header */
	int exec_fd;
	int read_sofar;			/* Number of bytes we have read */
	int read_thistime;		/* work it out for yourself */

	if (( exec_fd = open ( filename, O_RDONLY )) < 0 )
	{
	    fprintf ( stderr,"%s: Failed to open %s\n", progname, filename);
	    exit ( 1 );
	}

        /* Read in the fixed part of the executive header */
        if ( read ( exec_fd, &exec, sizeof ( exec )) != sizeof ( exec ))
        {
	    fprintf ( stderr,"%s: Cannot read header from %s file %s\n",
			    		progname, name, filename );
	    exit ( 1 );
	}

	/* Check the magic. Byte order is going to hit us on a non-Intel
	 * platform one day, but I'd rather not modify the code until we have
	 * a machine to test on
	 */
	if ( exec.exe_sig_mz != (('Z' << 8 ) | 'M'))
	{
	    fprintf ( stderr,"%s: Bad magic in %s header\n", progname, name );
	    if ( exec.exe_sig_mz == (('M' << 8 ) | 'Z'))
		fprintf ( stderr,"%s: Incompatible byte ordering ?\n",progname);

	    exit ( 1 );
	}

	/* Don't support relocation */
	if ( exec.num_relocs != 0 )
	{
	    fprintf ( stderr,"%s: Cannot relocate %s\n", progname, name );
	    exit ( 1 );
	}

	if ( verbose )
	{
	    printf ("%s start address: %04X:%04X\n", name, exec.code_disp,
			    		exec.ip_offset );
	    printf ("%s header size  : %d bytes\n", name, exec.hdr_size * 16 );
	}

	if ( load_header )
	{
	    /* Copy fixed header info into the write buffer */
	    memcpy ( buf->data, &exec, sizeof ( exec ));
	    read_sofar = sizeof ( exec );
	}
	else
	{
	    /* Not loading the header implies BIOS. Record start address for
	     * use later
	     */
	    bios_start_seg = exec.code_disp;
	    bios_start_off = exec.ip_offset;

	    if ( verbose )
	    {
		printf ("Skipping %d header bytes\n",
                        (int)(exec.hdr_size * 16 - sizeof (exec)));
	    }

	    if ( lseek ( exec_fd, exec.hdr_size * 16, SEEK_SET ) !=
                                        exec.hdr_size * 16 )
	    {
		fprintf ( stderr,"%s: Skipping %s header failed\n", progname,
					name );
		exit ( 1 );
	    }
	    read_sofar = 0;
	}

	/* Read in the rest of the file for download. Should never be greater
	 * than CBUFSIZ bytes
	 */
        read_thistime = read ( exec_fd, buf->data + read_sofar,
					CBUFSIZ - read_sofar  );
        if ( read_thistime < 0 )
        {
            fprintf ( stderr,"%s: Read of %s failed\n", progname, name );
            exit ( 1 );
        }
	read_sofar += read_thistime;
        if ( read_sofar >= CBUFSIZ )
        {
            fprintf ( stderr,"%s: %s too large\n", progname, name );
            exit ( 1 );
        }
        if ( verbose )
        {
            printf ("Read %d bytes of %s\n", read_sofar, name );
        }

        /* Done with the file */
        (void) close ( exec_fd );

        /* Improvement. We could checksum the code here, but that's getting
         * really paranoid.
         */

	/* Write the code to the card */
	write_buf ( read_sofar, offset, name );
}

/*
 *	Card start and stop
 */
static void
reset_card ( void )
{
	if ( ioctl ( sock, FSTCPURESET, &req ))
	{
	    fprintf ( stderr,"%s: Problem reseting card. %s\n", progname,
						strerror ( errno ));
	    exit ( 1 );
	}
}


static void
start_card ( void )
{
	unsigned char *rv;

	if ( verbose && ( bios_start_off != 0x0550 || bios_start_seg != 0x0000))
	{
	    printf ("WARNING: Unusual BIOS start address %04X:%04X\n",
					bios_start_seg, bios_start_off );
	}

	/* Setup and write the reset vector */
	rv = buf->data;

        *rv++ = 0xBA;               /* 00: mov dx,IPCB.LMCS     */
        *rv++ = 0xA2;
        *rv++ = 0xFF;
        *rv++ = 0xB8;               /* 03: mox ax,7F9Ch         */
        *rv++ = 0x9C;
        *rv++ = 0x7F;
        *rv++ = 0xEF;               /* 06: out dx,ax            */
        *rv++ = 0xEA;               /* 07: jmp far ptr start... */
        *rv++ =   bios_start_off        & 0xFF;
        *rv++ = ( bios_start_off >> 8 ) & 0xFF;
        *rv++ =   bios_start_seg        & 0xFF;
        *rv   = ( bios_start_seg >> 8 ) & 0xFF;

	write_buf ( 12, 0xFFFF0,"Reset vector");

	if ( ioctl ( sock, FSTCPURELEASE, &req ))
	{
	    fprintf ( stderr,"%s: Problem releasing card. %s\n", progname,
						strerror ( errno ));
	    exit ( 1 );
	}
}


/*
 *	Check card is running
 */
static void
check_running ( void )
{
	struct fstioc_info info;

	memset ( &info, 0, sizeof ( info ));
	req.ifr_data = (char *)&info;
	if ( ioctl ( sock, FSTGETCONF, &req ))
	{
	    fprintf ( stderr,"%s: Problem getting port information. %s\n",
				progname, strerror ( errno ));
	    exit ( 1 );
	}

	if ( info.state != FST_RUNNING )
	{
		fprintf ( stderr,"%s: Card failed to start. State %d\n",
				progname, info.state );
	}

	if ( verbose )
	{
	    printf ("Card running\n");
	}
}

/*
 *	Download firmware to card and start
 */
static void
do_load ( char *bios_name, char *fw_name )
{
	alloc_buf ();
	reset_card ();
	zero_memory ();
	load_file ( bios_name,"BIOS code", FALSE, 0x00000 );
	load_file ( fw_name,  "Firmware",  TRUE,  0x08000 );
	start_card ();
	sleep ( 1 );
	check_running ();
}

/*
 * Routines to set the synthesiser value and check the reply
 */
int
getdigit( char digit)
{
      if ((digit >= '0') && (digit <= '9'))
	{
	  return digit - 0x30;
	}
      if ((digit >= 'A') && (digit <= 'F'))
	{
	  return digit - 0x41+0xa;
	}
      if ((digit >= 'a') && (digit <= 'f'))
	{
	  return digit - 0x61+0xa;
	}
      printf("Shouldn't be here %d\n", digit);
      return -1;
}

int
set_synth ( char *config_data, int command, int port )
{
  struct fstioc_req request;
  int i;
  int len;

  len = strlen(config_data)/2;
  req.ifr_data = (char *)&request;
  memset(&request, 0, sizeof(struct fstioc_req));
  request.msg_type = command;
  request.msg_len = sizeof(struct fstioc_req);
  request.i_reg_idx = port;
  /*
   * Set incrementing data as the data to hash
   */
  for (i=0; i<len; i++)
    {
      request.u_msg[i] = (getdigit(config_data[i*2])*16) +
	(getdigit(config_data[i*2+1]));
    }
  if ( ioctl ( sock, FSTSYSREQ, &req ))
    {
      fprintf ( stderr,"Problem sending ioctl to card. %s\n", strerror ( errno ));
      exit ( 1 );
    }
  if (request.ret_code)
    {
      printf("Error in setting synthesizer value\n");
      exit (1);
    }
  printf("Synth setting set ok\n");
  return ( 0 );
}

/* End $Id: farutil.c,v 1.3 2009/12/17 11:31:22 kevinc Exp $ */

static void
do_num ( int *result, char *arg, int lower, int upper, char *descrip )
{
	long val;
	char *end;

	val = strtol ( arg, &end, 0 );

	if ( val < lower || val > upper || *end != '\0')
	{
	    fprintf ( stderr,"%s: %s must be a number between %d and %d\n",
	    				progname, descrip, lower, upper );
	    exit ( 1 );
	}

	*result = val;
}

#ifdef GEN_HDLC
/*
 *	Set the protocol to be used or print an error if an incompatable mode
 *	is already set
 */
static void
set_proto ( int proto, char *err )
{
	if ( protocol == 0 )
	{
	    protocol = proto;
	    return;
	}
	if ( protocol == proto )
	{
	    return;
	}

	fprintf ( stderr,"%s: %s\n", progname, err );
}
#endif

/*
 *	Convert a string to line speed
 *	Print error and help then exit if it's not valid
 */

static unsigned int
lookup_async_speed ( char *str )
{
  struct speeds *sp;

  for ( sp = async_speeds ; sp->text != NULL ; sp++ )
    {
      if ( strcasecmp ( sp->text, str ) == 0 )
        return sp->val;
    }
  fprintf ( stderr,"%s: unknown async line speed \"%s\".\n"
            "\n"
            "The nearest to this speed will be selected:\n"
            , progname, str );
  return atoi(str);
}

static unsigned int
lookup_speed ( char *str, int card_type )
{
	struct speeds *sp;

        /*
         * With the introduction of the T4E and the clock synthisizer
         * almost any speed is valid.  In this case allow the user to
         * select a value not in the table above.
         */
        if ( (card_type == FST_TYPE_T4E) || 
             (card_type == FST_TYPE_T4Ep) || 
             (card_type == FST_TYPE_T2UE))
          {
            for ( sp = t4e_speeds ; sp->text != NULL ; sp++ )
              {
                if ( strcasecmp ( sp->text, str ) == 0 )
                  return sp->val;
              }
            fprintf ( stderr,"%s: unknown T4E/T2UE line speed \"%s\".\n"
                      "\n"
                      "The nearest to this speed will be selected:\n"
                      , progname, str );
            return atoi(str);
          }
        else if ( card_type == FST_TYPE_FLEX1 )
          {
            for ( sp = flex_speeds ; sp->text != NULL ; sp++ )
              {
                if ( strcasecmp ( sp->text, str ) == 0 )
                  return sp->val;
              }
            fprintf ( stderr,"%s: unknown Flex line speed \"%s\".\n"
                      "\n"
                      "The nearest to this speed will be selected:\n"
                      , progname, str );
            exit ( 1 );
          }
        else
          {
            for ( sp = speed_table ; sp->text != NULL ; sp++ )
              {
                if ( strcasecmp ( sp->text, str ) == 0 )
                  return sp->val;
              }
            fprintf ( stderr,"%s: unknown line speed \"%s\".\n"
                      "\n"
                      "Recognised speeds are:\n"
                      "\t9600, 19200, 38400, 64000, 128000, 256000, 512000\n"
                      "\t1024000, 2048000, 4096000, 8192000\n"
                      , progname, str );
            exit ( 1 );
          }
}

/*
 *     Special function for T4E cards to work out the clock settings
 */

static int
set_t4e_clock(char *mode, struct fstioc_info *info)
{
        //printf("mode is %s\n", mode);
        //info->internalClock = 0;
        if (strcasecmp ( mode,"mode0") == 0)
	  {
            info->extendedClocking = 0x85;
	  }
        else if (strcasecmp ( mode,"mode1") == 0)
	  {
            info->extendedClocking = 0x8B;
	  }
        else if (strcasecmp ( mode,"mode2") == 0)
	  {
            info->extendedClocking = 0x81;
	  }
        else if (strcasecmp ( mode,"erx_etx") == 0)
	  {
            info->extendedClocking = 0x80;
	  }
        else if (strcasecmp ( mode,"erx_itx") == 0)
	  {
            info->extendedClocking = 0x81;
	  }
        else if (strcasecmp ( mode,"irx_etx") == 0)
	  {
            info->extendedClocking = 0x82;
	  }
        else if (strcasecmp ( mode,"irx_itx") == 0)
	  {
            info->extendedClocking = 0x83;
	  }
        else if (strcasecmp ( mode,"none") == 0)
	  {
            info->extendedClocking = 0x0;
	  }
        else
          return 0;
        return 1;
}


static unsigned int
lookup_shdsl_speed ( char *str )
{
	struct speeds *sp;

	for ( sp = shdsl_speed_table ; sp->text != NULL ; sp++ )
	{
	    if ( strcasecmp ( sp->text, str ) == 0 )
		return sp->val;
	}

	fprintf ( stderr,"%s: unknown line speed \"%s\" for SHDSL interface.\n"
			"\n"
			"Recognised speeds are from 192000 in multiples of 64000:\n"
			"\te.g. 192000, 704000, 1024000, 2048000 etc "
			"up to a maximum of 2304000\n"
			, progname, str );
	exit ( 1 );
}

static unsigned int
lookup_e1_speed ( char *str )
{
	struct speeds *sp;

	for ( sp = e1_speed_table ; sp->text != NULL ; sp++ )
	{
	    if ( strcasecmp ( sp->text, str ) == 0 )
		return sp->val;
	}

	fprintf ( stderr,"%s: unknown line speed \"%s\" for E1 interface.\n"
			"\n"
			"Recognised speeds are multiples of 64000:\n"
			"\te.g. 64000, 128000, 192000, 256000 etc "
			"up to a maximum of 2048000\n"
			, progname, str );
	exit ( 1 );
}

static unsigned int
lookup_t1_speed ( char *str )
{
	struct speeds *sp;

	for ( sp = t1_speed_table ; sp->text != NULL ; sp++ )
	{
	    if ( strcasecmp ( sp->text, str ) == 0 )
		return sp->val;
	}

	fprintf ( stderr,"%s: unknown line speed \"%s\" for T1 interface.\n"
			"\n"
			"Recognised speeds are multiples of 64000:\n"
			"\te.g. 64000, 128000, 192000, 256000 etc "
			"up to a maximum of 1536000\n"
			, progname, str );
	exit ( 1 );
}

/*
 *      Simple routine to convert  an input string into an integer
 *      where the input string can be decimal or hex
 *      The hex must start with 0x
 */
static int
get_value(char * value_string)
{
  int value_int = -1;

  if ((value_string[0] == '0') && (value_string[1] == 'x'))
    { 
      /*
       * Hex conversion required
       */
      sscanf(value_string, "%x", &value_int);
    }
  else
    {
      if (isdigit(value_string[0]))
	value_int = atoi(value_string);
    }
  return value_int;
}

static void 
check_valid_card_te1(int card_type, char *param)
{
  if (card_type != FST_TYPE_TE1)
    {
      fprintf(stderr, "Cannot configure parameter [%s] on %s\n", 
              param, card_strings[card_type]);
      exit (1);
    }
}

static void 
check_valid_card_dsl(int card_type)
{
  if (card_type != FST_TYPE_DSL_S1)
    {
      fprintf(stderr, "Cannot configure DSL parameters on %s\n", 
              card_strings[card_type]);
      exit (1);
    }
}

/*
 * 	Set various configuration options
 */
static void
do_set ( int argc, char *argv[] )
{
	struct fstioc_info info;
	char *last;
	long i;
	int card_type;
        int line_interface;
        int line_speed;
        int slot ;
        int can_do_async;
        int can_do_synth;
#ifdef GEN_HDLC
	sync_serial_settings synch;
	bool is_gen_hdlc;
        unsigned int cmd;
#endif
	/*
	 * Get the basic configuration, we will need to refer to it
	 */
	memset ( &info, 0, sizeof ( info ));
	req.ifr_data = (char *)&info;
	if ( ioctl ( sock, FSTGETCONF, &req ))
	{
	    fprintf ( stderr,"%s: Problem getting port information. %s\n",
				progname, strerror ( errno ));
            exit ( 1 );
        }
	card_type = info.type;
        line_interface = info.lineInterface;
        line_speed = info.lineSpeed;
        slot = info.startingSlot;
        can_do_async = info.asyncAbility;
        can_do_synth = info.synthAbility;

#ifdef GEN_HDLC
	/* Probe to see if the driver supports the "generic HDLC" system */
	is_gen_hdlc = ( info.proto == FST_GEN_HDLC );

	/* Get clock setting is case user changes physical interface without
	 * respecifying the clock
	 */
	if ( is_gen_hdlc )
	{
	    req.ifr_settings.type = IF_GET_IFACE;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	    req.ifr_settings.size = sizeof ( synch );
	    req.ifr_settings.ifs_ifsu.sync = &synch;
            cmd = SIOCWANDEV;
#else
	    req.ifr_settings.data_length = sizeof ( synch );
	    req.ifr_settings.data = &synch;
            cmd = SIOCDEVICE;
#endif
	    if ( ioctl ( sock, cmd, &req ))
	    {
		if ( errno == ENOSYS )
		{
		    /* No interface set. Set clock from FSTGETCONF values */
		    synch.clock_type = ( info.internalClock == INTCLK )
		    				? CLOCK_INT : CLOCK_EXT;
		    synch.clock_rate = info.lineSpeed;
		}
		else
		{
		    fprintf ( stderr,"%s: Problem getting clock information."
			    	" %s\n", progname, strerror ( errno ));
		    exit ( 1 );
		}
	    }
	}
#else
	//memset ( &info, 0, sizeof ( info ));
#endif
        info.valid = 0;

	/* Gather parameters in the info structure.  With the exception of dce
	 * parameters come in pairs of keyword/value
	 */
	while ( argc >= 2 || ( argc == 1 && strcasecmp ( argv[0],"dce") == 0 ))
	{
	    if ( strcmp ( argv[0],"media") == 0 )
	    {
              if (card_type != FST_TYPE_DSL_S1)
                {
                  if ( strcasecmp ( argv[1],"x21") == 0 )
		    info.lineInterface = X21;
                  else if ( strcasecmp ( argv[1],"x21d") == 0 )
		    info.lineInterface = X21D;
                  else if ( strcasecmp ( argv[1],"v35") == 0 )
		    info.lineInterface = V35;
                  else if ( strcasecmp ( argv[1],"v24") == 0
                            || strcasecmp ( argv[1],"rs232") == 0 )
		    info.lineInterface = V24;
                  else if ( strcasecmp ( argv[1],"rs449") == 0 
                               ||  strcasecmp ( argv[1],"rs530") == 0 )
                    info.lineInterface = RS530_449;
                  else if ( strcasecmp ( argv[1],"rs485") == 0 )
                    info.lineInterface = RS485;
                  else if ( strcasecmp ( argv[1],"ux35c") == 0 )
                    info.lineInterface = UX35C;
                  else
                    {
                      fprintf ( stderr,"%s: Media type \"%s\" not allowed on %s \n",
                                progname, argv[1], card_strings[card_type] );
                      exit ( 1 );
                    }
                }
              else
                {
                  if ( strcasecmp ( argv[1],"e1") == 0 )
                    {
                      info.lineInterface = E1;
                      info.framing = E1;
                    }
                  else if ( strcasecmp ( argv[1],"t1") == 0 )
                    {
                      info.lineInterface = T1;
                      info.framing = T1;
                    }
#ifdef FSTCONFJ1
                  else if ( strcasecmp ( argv[1],"j1") == 0 )
                    {
                      info.lineInterface = J1;
                      info.framing = J1;
                    }
#endif
                  else if ( strcasecmp ( argv[1],"shdsl") == 0 )
                    {
                      info.lineInterface = SHDSL;
                    }
                  else
                    {
                      fprintf ( stderr,"%s: Media type \"%s\" not allowed on %s \n",
				progname, argv[1], card_strings[card_type] );
                      exit ( 1 );
                    }
                }

		if ((info.lineInterface == X21D && card_type == FST_TYPE_T2P) 
                    || (info.lineInterface == X21D && card_type == FST_TYPE_T4P)
                    || (info.lineInterface == X21D && card_type == FST_TYPE_FLEX1))
		  {
		    fprintf(stderr, "Dual Clocking only allowed on TXU\n");
		    exit (1);
		  }
		info.valid |= FSTVAL_CABLE;
                line_interface = info.lineInterface;
	    }
	    else if ( strcmp ( argv[0],"termination") == 0 )
	    {
		if ( strcasecmp ( argv[1],"none") == 0 )
		    info.termination = 0;
		else if ( strcmp ( argv[1],"resistive") == 0 )
		    info.termination = 1;
		else
		{
		    fprintf ( stderr,"%s: Unknown termination value \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
		}
		info.valid |= FSTVAL_CABLE;
            }
	    else if ( strcasecmp ( argv[0],"NRZIClocking") == 0 )
              {
                if ((card_type == FST_TYPE_T4Ep) || 
                    (card_type == FST_TYPE_FLEX1))
                  {
                    if ( strcasecmp ( argv[1],"enable") == 0 )
                      info.enableNRZIClocking = 1;
                    else if ( strcmp ( argv[1],"disable") == 0 )
                      info.enableNRZIClocking = 0;
                    else
                      {
                        fprintf ( stderr,"%s: Unknown NRZIClocking mode \"%s\"\n",
                                  progname, argv[1] );
                        exit ( 1 );
                      }
                    info.valid |= FSTVAL_SPEED;
                  }
                else
                  {
                    fprintf(stderr, "%s: NRZIClocking only valid on T4E+\n",
                            progname);
                  }
              }
            else if ( strcmp ( argv[0],"carrier") == 0 )
            {
              if ( strcasecmp ( argv[1],"ignore") == 0 )
                info.ignoreCarrier = 1;
              else if ( strcmp ( argv[1],"required") == 0 )
                info.ignoreCarrier = 0;
              else
		{
                  fprintf ( stderr,"%s: Unknown carrier mode \"%s\"\n",
                            progname, argv[1] );
                  exit ( 1 );
		}
              info.valid |= FSTVAL_CABLE;
            }
#if !defined(__FreeBSD__)
	    else if ( strcmp ( argv[0],"proto") == 0 )
	    {
#ifdef GEN_HDLC
		if ( is_gen_hdlc )
		{
		    if ( strcasecmp ( argv[1],"ppp") == 0 )
		    {
			set_proto ( IF_PROTO_PPP,
					"Conflicting protocol selection");
		    }
		    else if ( strcasecmp ( argv[1],"cisco") == 0 )
		    {
			set_proto ( IF_PROTO_CISCO,
					"Conflicting protocol selection");
		    }
		    else if ( strcasecmp ( argv[1],"hdlc") == 0 )
		    {
			set_proto ( IF_PROTO_HDLC,
					"Conflicting protocol selection");
		    }
		    else if ( strcasecmp ( argv[1],"fr") == 0 )
		    {
			set_proto ( IF_PROTO_FR,
					"Conflicting protocol selection");
		    }
		    else if ( strcasecmp ( argv[1],"x25") == 0
		    			|| strcasecmp ( argv[1],"x.25") == 0 )
		    {
			set_proto ( IF_PROTO_X25,
					"Conflicting protocol selection");
		    }
		    else if ( strcasecmp ( argv[1],"raw") == 0)
		    {
                      /*
                       * Because RAW mode is a FarSync only function
                       * we need to deal with it in a private way
                       */
                        info.proto = FST_RAW;
                        protocol = 0; 
		    }
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,22)
		    else if ( strcasecmp ( argv[1],"hdlc_eth") == 0)
		    {
			set_proto ( IF_PROTO_HDLC_ETH,
                                    "Conflicting protocol selection");
		    }
#endif
		    else
		    {
			fprintf ( stderr,"%s: Unknown protocol \"%s\"\n",
				    	progname, argv[1] );
			exit ( 1 );
		    }
		}
		else /* Tricksy drop through to non-ifdef case */
#endif
		{
		    /* Note non-gen HDLC code up to V0.08 supported "hdlc" as
		     * an undocumented pseudonym for "cisco", this has been
		     * dropped for obvious reasons.
		     */
		    if ( strcasecmp ( argv[1],"ppp") == 0 )
			info.proto = FST_PPP;
		    else if ( strcasecmp ( argv[1],"cisco") == 0 )
			info.proto = FST_HDLC;
		    else if ( strcasecmp ( argv[1],"monitor") == 0 )
			info.proto = FST_MONITOR;
		    else if ( strcasecmp ( argv[1],"raw") == 0 )
			info.proto = FST_RAW;
		    else
		    {
#ifdef GEN_HDLC
			/* Warning: two Ifs sharing the same Else here */
			if ( strcasecmp ( argv[1],"hdlc") == 0
				    	|| strcasecmp ( argv[1],"fr") == 0
				    	|| strcasecmp ( argv[1],"x25") == 0 )
			{
			    fprintf ( stderr,"%s: Protocol \"%s\" requires the"
				    " generic HDLC device driver varient.\n",
				    	progname, argv[1] );
			}
#else
			if ( strcasecmp ( argv[1],"hdlc") == 0 )
			{
			    fprintf ( stderr,"%s: \"proto hdlc\" has been"
				    " depreciated. Use \"proto cisco\".\n",
				    	progname );
			}
#endif
			else
			{
			    fprintf ( stderr,"%s: Unknown protocol \"%s\"\n",
					progname, argv[1] );
			}
			exit ( 1 );
		    }
		}
		info.valid |= FSTVAL_PROTO;
	    }
#endif
            else if ( strcasecmp ( argv[0], "lineMode") == 0 )
            {
		if ( strcmp ( argv[1],"hdlc") == 0 )
		    info.transparentMode = FST_MODE_HDLC;
		else if ( strcmp ( argv[1],"transparent") == 0 )
		    info.transparentMode = FST_MODE_TRANSPARENT;
		else if ( strcmp ( argv[1],"async") == 0 )
		    info.transparentMode = FST_MODE_ASYNC;
		else
		{
		    fprintf ( stderr,"%s: Unknown Line Mode mode \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
		}
		info.valid |= FSTVAL_SPEED | FSTVAL_ASYNC;
            }
	    else if ( strcasecmp ( argv[0],"led") == 0 )
	    {
		if ( strcmp ( argv[1],"flash") == 0 )
		    info.cardMode = CARD_MODE_IDENTIFY;
		else if ( strcmp ( argv[1],"normal") == 0 )
		    info.cardMode = 0;
		else
		{
		    fprintf ( stderr,"%s: Unknown LED mode \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
		}
		info.valid |= FSTVAL_MODE;
	    }
	    else if ( strcasecmp ( argv[0],"phase") == 0 )
	    {
	        if (strcasecmp (argv[1], "inverted") == 0 )
		    info.invertClock = 1;
		else if ( strcasecmp ( argv[1],"normal") == 0 )
		    info.invertClock = 0;
		else
		{
		    fprintf ( stderr,"%s: Unknown PHASE mode \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
		}
		info.valid |= FSTVAL_PHASE;
	    }
	    else if ( strcasecmp ( argv[0],"synth") == 0 )
	    {
              if (!can_do_synth)
                {
                  fprintf ( stderr, "Cannot set clock synthesizer on %s\n",
                            card_strings[card_type]);
                  exit (1);

                }
              if (info.coding != CODING_MANCHESTER)
                {
                  set_synth(argv[1], 1, info.index);
                }
              else
                {
                  set_synth(argv[1], 2, info.index);
                }
            }
	    else if ( strcasecmp ( argv[0],"clock") == 0 )
	    {
                if ( strcmp ( argv[1],"ext") == 0 )
		  {
		    if (info.lineInterface == X21D)
		     {
		       fprintf ( stderr, "Only internal clocking allowed with X21D\n");
		       exit (1);
		     } 
		    info.internalClock = EXTCLK;
		    info.lineSpeed = 0;
		  }
		  else 
		  {
                    info.internalClock = INTCLK;
                    if (info.transparentMode == FST_MODE_ASYNC)
                      info.lineSpeed = lookup_async_speed( argv[1] );
                    else
                      info.lineSpeed = lookup_speed ( argv[1], info.type );
		    if (card_type == FST_TYPE_T1U)
		      {
                        if (info.lineSpeed > 38400)
                          {
                            fprintf(stderr, "Max internal clock is 38400 for T1U\n");
                            exit (1);
                          }
		      }
                    if (line_interface == E1)
                      {
		        fprintf(stderr, "Please do not use clock to set the E1 data rate.  Use datarate or startslot and numslots\n");
		        exit (1);
                      }
		  }
		info.valid |= FSTVAL_SPEED;
	    }
	    else if ( strcasecmp ( argv[0],"extended") == 0 )
	    {
              if ((card_type == FST_TYPE_T4E) || 
                  (card_type == FST_TYPE_T4Ep) || 
                  (card_type == FST_TYPE_FLEX1))
                {
                  if ((info.lineInterface == V24) || 
                      (info.lineInterface == RS530_449))
                    {
                      /*
                       * T4E card can support additional clock configurations
                       * So I have hived it off into a function rather than do it
                       * in line
                       */
                      if (set_t4e_clock(argv[1], &info))
                        {
                          //printf("extended clocking is %x\n", info.extendedClocking);
                          //printf("line speed set as %d\n", info.lineSpeed);
                          info.valid |= FSTVAL_T4E;
                        }
                      else
                        {
                          printf("Extended clocking value %s is not valid\n",
                                 argv[1]);
                          exit (1);
                        }
                    }
                  else
                    {
                      printf("Can only use Extended Clocking on V24,RS449 or RS530 interfaces\n");
                    }
                }
              else
                {
                  printf("Can only use Extended Clocking on T4E or Flex interfaces\n");
                }
	    }
	    else if ( strcasecmp ( argv[0],"numbuffers") == 0 )
	    {
              info.numRxBuffers = atoi(argv[1]);
              info.numTxBuffers = atoi(argv[1]);
              info.valid |= FSTVAL_BUFFERS;
            }
	    else if ( strcasecmp ( argv[0],"numrxbuffers") == 0 )
	    {
              info.numRxBuffers = atoi(argv[1]);
              info.valid |= FSTVAL_BUFFERS;
            }
	    else if ( strcasecmp ( argv[0],"numtxbuffers") == 0 )
	    {
              info.numTxBuffers = atoi(argv[1]);
              info.valid |= FSTVAL_BUFFERS;
            }
	    else if ( strcasecmp ( argv[0],"sizebuffers") == 0 )
	    {
              info.rxBufferSize = atoi(argv[1]);
              info.txBufferSize = atoi(argv[1]);
              info.valid |= FSTVAL_BUFFERS;
            }
	    else if ( strcasecmp ( argv[0],"sizerxbuffers") == 0 )
	    {
              info.rxBufferSize = atoi(argv[1]);
              info.valid |= FSTVAL_BUFFERS;
            }
	    else if ( strcasecmp ( argv[0],"sizetxbuffers") == 0 )
	    {
              info.txBufferSize = atoi(argv[1]);
              info.valid |= FSTVAL_BUFFERS;
            }	    else if ( strcasecmp ( argv[0],"lowlatency") == 0 )
	    {
              if ( card_type != FST_TYPE_FLEX1 )
                {
                  if ( strcasecmp ( argv[1],"rx") == 0 )
                    info.lowLatency = lowLatencyRx;
                  else if ( strcasecmp ( argv[1],"tx") == 0 )
                    info.lowLatency = lowLatencyTx;
                  else if ( strcasecmp ( argv[1],"txrx") == 0 )
                    info.lowLatency = lowLatencyRxTx;
                  else if ( strcmp ( argv[1],"disable") == 0 )
                    info.lowLatency = lowLatencyDisable;
                  else
                    {
                      fprintf ( stderr,"%s: Unknown lowLatency setting \"%s\"\n",
                                progname, argv[1] );
                      exit ( 1 );
                    }
                  info.valid |= FSTVAL_CARD;
                }
              else
                {
                  fprintf ( stderr,"%s: Cannot configure lowLatency on Flex\n",
                            progname );
                  exit ( 1 );

                }
            }
	    else if ( strcasecmp ( argv[0],"debug") == 0 )
	    {
		i = strtol ( argv[1], &last, 0 );
		if ( last == argv[1] || *last != '\0' || i == LONG_MIN
				|| i == LONG_MAX )
		{
		    fprintf ( stderr,"%s: Debug flags must be a number\n",
				progname );
		    exit ( 1 );
		}
		info.debug = i;
		info.valid |= FSTVAL_DEBUG;
	    }
            /*
             * TE1 parameters
             */
	    else if ( strcasecmp ( argv[0],"datarate") == 0 )
	    {
              if (card_type == FST_TYPE_TE1)
                {
                  if (line_interface == E1)
                    {
                      info.lineSpeed = lookup_e1_speed(argv[1]);
                      if (slot >= (32 - (line_speed/640000)))
                        {
                          fprintf(stderr, "Not enough slots for requested bandwidth, i.e. cannot fit %d into %d slots\n", line_speed, slot);
                          exit (1);
                        }
                    }
                  if (line_interface == T1)
                    {
                      info.lineSpeed = lookup_t1_speed(argv[1]);
                    }
                  line_speed = info.lineSpeed;
                  info.valid |= FSTVAL_TE1;
                }
              else
                {
                  /*
                   * This is a shared parameter with shdsl
                   */
                  check_valid_card_dsl(card_type);
                  info.lineSpeed = lookup_shdsl_speed(argv[1]);
                  line_speed = info.lineSpeed;
                  info.valid |= FSTVAL_DSL_S1;
                }
            }
	    else if ( strcasecmp ( argv[0],"numslots") == 0 )
	    {
              int my_num_slots;
              /*
               * This is an alternatuve way of specifying the line rate.
               * We need to check that there are enough slots for
               * the required bandwidth.  e.g. he cannot request 1024k after
               * slot 16.  Note we will need a different calculation when we
               * support T1.
               */
              check_valid_card_te1(card_type, argv[0]);
              my_num_slots = atoi(argv[1]);
              if (info.lineInterface == E1)
                {
                  if ((my_num_slots > 32) || (my_num_slots < 1))
                    {
                      fprintf(stderr, "E1: numslots must be between 1 and 32\n");
                      exit (1);
                    }
                  if (my_num_slots > (32 - slot))
                    {
                      fprintf(stderr, "E1: Not enough slots for requested bandwidth, i.e. cannot fit %d into %d slots\n", line_speed, slot);
                      exit (1);
                    }
                }
              else
                {
                  if ((my_num_slots > 24) || (my_num_slots < 1))
                    {
                      fprintf(stderr, "T1: numslots must be between 1 and 24\n");
                      exit (1);
                    }
                  if (my_num_slots > (24 - slot))
                    {
                      fprintf(stderr, "T1: Not enough slots for requested bandwidth, i.e. cannot fit %d into %d slots\n", line_speed, slot);
                      exit (1);
                    }
                }
              info.lineSpeed = my_num_slots* 64000;
              line_speed = info.lineSpeed;
              info.valid |= FSTVAL_TE1;
            }
	    else if ( strcasecmp ( argv[0],"startslot") == 0 )
            {
              /*
               * This is the starting timeslot that the user wishes
               * to use for his bandwidth selected.
               * We need to check that there are enough slots for
               * the required bandwidth.  e.g. he cannot request 1024k after
               * slot 16.  Note we will need a different calculation when we
               * support T1.
               */
              check_valid_card_te1(card_type, argv[0]);
              info.startingSlot = atoi(argv[1]);
              if (info.lineInterface == E1)
                {
                  if (info.startingSlot > 31)
                    {
                      fprintf ( stderr,"%s: E1: Invalid SLOT value \"%s\", use 0-31\n",
                                progname, argv[1] );
                      exit ( 1 );
                    }
                  if (info.startingSlot > (32 - (line_speed/64000)))
                    {
                      fprintf (stderr,"%s: E1: Starting slot %d is too large for bandwidth selected %d\n",
                               progname, info.startingSlot, line_speed );
                      exit (1);
                    }
                }
              else
                {
                  if (info.startingSlot > 24)
                    {
                      fprintf ( stderr,"%s: T1: Invalid SLOT value \"%s\", use 0-31\n",
                                progname, argv[1] );
                      exit ( 1 );
                    }
                  if (info.startingSlot > (24 - (line_speed/64000)))
                    {
                      fprintf (stderr,"%s: T1: Starting slot %d is too large for bandwidth selected %d\n",
                               progname, info.startingSlot, line_speed );
                      exit (1);
                    }
                }
              info.valid |= FSTVAL_TE1;
            }
	    else if ( strcasecmp ( argv[0],"mode") == 0 )
	    {
              check_valid_card_te1(card_type, argv[0]);
              if (strcasecmp (argv[1], "master") == 0 )
                info.clockSource = CLOCKING_MASTER;
              else if ( strcasecmp ( argv[1],"slave") == 0 )
                info.clockSource = CLOCKING_SLAVE;
              else
		{
                  fprintf ( stderr,"%s: Unknown MODE value \"%s\"\n",
                            progname, argv[1] );
                  exit ( 1 );
		}
              info.valid |= FSTVAL_TE1;
	    }
	    else if ( strcasecmp ( argv[0],"structure") == 0 )
            {
              check_valid_card_te1(card_type, argv[0]);
              if (info.lineInterface == E1)
                {
                  if (strcasecmp (argv[1], "unframed") == 0 )
                    info.structure = STRUCTURE_UNFRAMED;
                  else if ( strcasecmp ( argv[1],"double") == 0 )
                    info.structure = STRUCTURE_E1_DOUBLE;
                  else if ( strcasecmp ( argv[1],"crc4") == 0 )
                    info.structure = STRUCTURE_E1_CRC4;
                  else if ( strcasecmp ( argv[1],"crc4m") == 0 )
                    info.structure = STRUCTURE_E1_CRC4M;
                  else 
                    {
                      fprintf(stderr, "%s: Unknown STRUCTURE value \"%s\"\n",
                              progname, argv[1] );
                      exit ( 1 );
                    }
                }
              else
                {
                  if (strcasecmp (argv[1], "unframed") == 0 )
                    info.structure = STRUCTURE_UNFRAMED;
                  else if ( strcasecmp ( argv[1],"f4") == 0 )
                    info.structure = STRUCTURE_T1_4;
                  else if ( strcasecmp ( argv[1],"f12") == 0 )
                    info.structure = STRUCTURE_T1_12;
                  else if ( strcasecmp ( argv[1],"f24") == 0 )
                    info.structure = STRUCTURE_T1_24;
                  else if ( strcasecmp ( argv[1],"f72") == 0 )
                    info.structure = STRUCTURE_T1_72;
                  else 
                    {
                      fprintf(stderr, "%s: Unknown MODE value \"%s\"\n",
                              progname, argv[1] );
                      exit ( 1 );
                    }
                }
		info.valid |= FSTVAL_TE1;
            }
	    else if ( strcasecmp ( argv[0],"interface") == 0 )
	    {
              check_valid_card_te1(card_type, argv[0]);
              if (strcasecmp (argv[1], "bnc") == 0 )
                info.interface = INTERFACE_BNC;
              else if ( strcasecmp ( argv[1],"rj48c") == 0 )
                info.interface = INTERFACE_RJ48C;
              else
		{
                  fprintf ( stderr,"%s: Unknown INTERFACE value \"%s\"\n",
                            progname, argv[1] );
                  exit ( 1 );
		}
              info.valid |= FSTVAL_TE1;
	    }
	    else if ( strcasecmp ( argv[0],"coding") == 0 )
            {
              if ((card_type == FST_TYPE_FLEX1) || 
                  (card_type == FST_TYPE_T4Ep))
                {
                  if ( strcasecmp ( argv[1],"nrz") == 0 )
                    info.coding = CODING_NRZ;
                  else if ( strcasecmp ( argv[1],"nrzi") == 0 )
                    info.coding = CODING_NRZI;
                  else if ( strcasecmp ( argv[1],"fm0") == 0 )
                    info.coding = CODING_FM0;
                  else if ( strcasecmp ( argv[1],"fm1") == 0 )
                    info.coding = CODING_FM1;
                  else if ( strcasecmp ( argv[1],"manchester") == 0 )
                    info.coding = CODING_MANCHESTER;
                  else 
                    {
                      fprintf(stderr, "%s: Flex/T4E+: Unknown CODING value \"%s\"\n",
                              progname, argv[1] );
                      exit ( 1 );
                    }
                  info.valid |= FSTVAL_TE1;
                }
              else
                {
                  check_valid_card_te1(card_type, argv[0]);
                  if (info.lineInterface == E1)
                    {
                      if (strcasecmp (argv[1], "hdb3") == 0 )
                        info.coding = CODING_HDB3;
                      else 
                        {
                          fprintf(stderr, "%s: E1: Unknown CODING value \"%s\"\n",
                                  progname, argv[1] );
                          exit ( 1 );
                        }
                    }
                  else 
                    {
                      if ( strcasecmp ( argv[1],"ami") == 0 )
                        info.coding = CODING_AMI;
                      else if ( strcasecmp ( argv[1],"ami_zcs") == 0 )
                        info.coding = CODING_AMI_ZCS;
                      else if ( strcasecmp ( argv[1],"b8zs") == 0 )
                        info.coding = CODING_B8ZS;
                      else 
                        {
                          fprintf(stderr, "%s: T1: Unknown CODING value \"%s\"\n",
                                  progname, argv[1] );
                          exit ( 1 );
                        }
                    }
                  info.valid |= FSTVAL_TE1;
                }
            }
	    else if ( strcasecmp ( argv[0],"lbo") == 0 )
            {
              check_valid_card_te1(card_type, argv[0]);
              if (info.lineInterface == T1)
                {
                  if (strcasecmp (argv[1], "0") == 0 )
                    info.lineBuildOut = LBO_0dB;
                  else if ( strcasecmp ( argv[1],"-7.5") == 0 )
                    info.lineBuildOut = LBO_7dB5;
                  else if ( strcasecmp ( argv[1],"-15") == 0 )
                    info.lineBuildOut = LBO_15dB;
                  else if ( strcasecmp ( argv[1],"-22") == 0 )
                    info.lineBuildOut = LBO_22dB5;
                  else 
                    {
                      fprintf(stderr, "%s: Unknown LBO value \"%s\"\n",
                              progname, argv[1] );
                      exit ( 1 );
                    }
                  info.valid |= FSTVAL_TE1;
                }
              else
                {
                  fprintf(stderr, "%s: LBO value only valid for T1 links\n",
                          progname);
                  exit(1);
                }
            }
	    else if ( strcasecmp ( argv[0],"equalizer") == 0 )
	    {
              check_valid_card_te1(card_type, argv[0]);
              if (strcasecmp (argv[1], "short") == 0 )
                info.equalizer = EQUALIZER_SHORT;
              else if ( strcasecmp ( argv[1],"long") == 0 )
                info.equalizer = EQUALIZER_LONG;
              else
		{
                  fprintf ( stderr,"%s: Unknown EQUALIZER value \"%s\"\n",
                            progname, argv[1] );
                  exit ( 1 );
		}
              info.valid |= FSTVAL_TE1;
	    }
	    else if ( strcasecmp ( argv[0],"loop") == 0 )
            {
              check_valid_card_te1(card_type, argv[0]);
              if (strcasecmp (argv[1], "none") == 0 )
                info.loopMode = LOOP_NONE;
              else if ( strcasecmp ( argv[1],"local") == 0 )
                info.loopMode = LOOP_LOCAL;
              else if ( strcasecmp ( argv[1],"remote") == 0 )
                info.loopMode = LOOP_REMOTE;
              else if ( strcasecmp ( argv[1],"pload+ts0") == 0 )
                info.loopMode = LOOP_PAYLOAD_INC_TS0;
              else if ( strcasecmp ( argv[1],"pload-ts0") == 0 )
                info.loopMode = LOOP_PAYLOAD_EXC_TS0;
              else 
                {
                  fprintf(stderr, "%s: Unknown LOOP value \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
                }
		info.valid |= FSTVAL_TE1;
            }
	    else if ( strcasecmp ( argv[0],"txBuffer") == 0 )
            {
              check_valid_card_te1(card_type, argv[0]);
              if (strcasecmp (argv[1], "none") == 0 )
                info.txBufferMode = BUFFER_NONE;
              else if ( strcasecmp ( argv[1],"96Bit") == 0 )
                info.txBufferMode = BUFFER_96_BIT;
              else if ( strcasecmp ( argv[1],"1Frame") == 0 )
                info.txBufferMode = BUFFER_1_FRAME;
              else if ( strcasecmp ( argv[1],"2Frame") == 0 )
                info.txBufferMode = BUFFER_2_FRAME;
              else 
                {
                  fprintf(stderr, "%s: Unknown TX BUFFER value \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
                }
		info.valid |= FSTVAL_TE1;           
            }
	    else if ( strcasecmp ( argv[0],"rxBuffer") == 0 )
            {
              check_valid_card_te1(card_type, argv[0]);
              if (strcasecmp (argv[1], "none") == 0 )
                info.rxBufferMode = BUFFER_NONE;
              else if ( strcasecmp ( argv[1],"96Bit") == 0 )
                info.rxBufferMode = BUFFER_96_BIT;
              else if ( strcasecmp ( argv[1],"1Frame") == 0 )
                info.rxBufferMode = BUFFER_1_FRAME;
              else if ( strcasecmp ( argv[1],"2Frame") == 0 )
                info.rxBufferMode = BUFFER_2_FRAME;
              else 
                {
                  fprintf(stderr, "%s: Unknown RX BUFFER value \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
                }
		info.valid |= FSTVAL_TE1;
            }
	    else if ( strcasecmp ( argv[0],"idleCode") == 0 )
            {
	      int idle_code;
	      check_valid_card_te1(card_type, argv[0]);
              idle_code = get_value(argv[1]);
              if ((idle_code < 0) || (idle_code > 0xff))
		{
                  fprintf(stderr, "%s: idleCode must be between 0 and 0xff\n",
				progname );
		    exit ( 1 );
		}
	      info.idleCode = idle_code;
	      info.valid |= FSTVAL_TE1;
	    }
	    else if ( strcasecmp ( argv[0],"range") == 0 )
            {
              check_valid_card_te1(card_type, argv[0]);
              if (info.lineInterface == T1)
                {
                  if ( (strcasecmp (argv[1], "0-133ft") == 0 ) ||
                       (strcasecmp (argv[1], "0-40m") == 0 ) )
                    info.range = RANGE_0_133_FT;
                  else if ( (strcasecmp ( argv[1],"133-266ft") == 0 ) ||
                            (strcasecmp ( argv[1],"40-81m") == 0) )
                    info.range = RANGE_133_266_FT;
                  else if ( (strcasecmp ( argv[1],"266-399ft") == 0 ) ||
                            (strcasecmp ( argv[1],"81-122m") == 0) )
                    info.range = RANGE_266_399_FT;
                  else if ( (strcasecmp ( argv[1],"399-533ft") == 0 ) ||
                            (strcasecmp ( argv[1],"122-162m") == 0) )
                    info.range = RANGE_399_533_FT;
                  else if ( (strcasecmp ( argv[1],"533-655ft") == 0 ) ||
                            (strcasecmp ( argv[1],"162-200m") == 0) )
                    info.range = RANGE_533_655_FT;
                  else 
                    {
                      fprintf(stderr, "%s: Unknown RANGE value \"%s\"\n",
                              progname, argv[1] );
                      exit ( 1 );
                    }
                  info.valid |= FSTVAL_TE1;           
                }
              else
                {
                      fprintf(stderr, "%s: Range only valid on T1 links\n",
                              progname);
                      exit ( 1 );
                }
            }
            /*
             * SHDSL parmaters
             */
	    else if ( strcasecmp ( argv[0],"termType") == 0 )
            {
              check_valid_card_dsl(card_type);
              if ( (strcasecmp (argv[1], "central") == 0 ) )
                info.terminalType = EQUIP_TYPE_CENTRAL;
              else if ( (strcasecmp ( argv[1],"remote") == 0 ) )
                info.terminalType = EQUIP_TYPE_REMOTE;
              else 
                {
                  fprintf(stderr, "%s: Unknown TERM TYPE value \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
                }
		info.valid |= FSTVAL_DSL_S1;
            }
	    else if ( strcasecmp ( argv[0],"annex") == 0 )
            {
              check_valid_card_dsl(card_type);
              if ( (strcasecmp (argv[1], "a") == 0 ) ||
                   (strcasecmp ( argv[1],"A") == 0) )
                info.annexType = ANNEX_A;
              else if ( (strcasecmp ( argv[1],"b") == 0 ) ||
                        (strcasecmp ( argv[1],"B") == 0) )
                info.annexType = ANNEX_B;
              else 
                {
                  fprintf(stderr, "%s: Unknown ANNEX value \"%s\" Use A or B\n",
                          progname, argv[1] );
		    exit ( 1 );
                }
		info.valid |= FSTVAL_DSL_S1;
            }
#if defined(GEN_HDLC) || defined(__FreeBSD__)
	    else if ( strcasecmp ( argv[0],"encap") == 0 )
            {
              check_valid_card_dsl(card_type);
              if ( (strcasecmp (argv[1], "ppp") == 0 ) ||
                   (strcasecmp ( argv[1],"PPP") == 0) )
                info.encap = ENCAP_PPP;
              else if ( (strcasecmp ( argv[1],"mpoa") == 0 ) ||
                        (strcasecmp ( argv[1],"MPOA") == 0) )
                {
                  info.encap = ENCAP_MPOA;
                }
              else 
                {
                  fprintf(stderr, "%s: Unknown encap value \"%s\" Use ppp or mpoa\n",
                          progname, argv[1] );
		    exit ( 1 );
                }
		info.valid |= FSTVAL_DSL_S1;
            }
#else
	    else if ( strcasecmp ( argv[0],"encap") == 0 )
            {
              fprintf(stderr, "%s: Warning encap = \"%s\" has no effect unless Generic HDLC module used\n",
                          progname, argv[1] );
            }
#endif
	    else if ( strcasecmp ( argv[0],"testMode") == 0 )
            {
              check_valid_card_dsl(card_type);
              info.testMode = atoi(argv[1]);
              info.valid |= FSTVAL_DSL_S1;
            }
	    else if ( strcasecmp ( argv[0],"backoff") == 0 )
            {
              check_valid_card_dsl(card_type);
              info.backoff = atoi(argv[1]);
              if (info.backoff > 31)
                {
                  fprintf(stderr, "%s: Invalid backoff range.  Use 0 - 31\n",
                          progname);
                  exit (1);
                }
              info.valid |= FSTVAL_DSL_S1;
            }
	    else if ( strcasecmp ( argv[0],"vpi") == 0 )
            {
              check_valid_card_dsl(card_type);
              info.vpi = atoi(argv[1]);
              if (info.vpi > 255)
                {
                  fprintf(stderr, "%s: Invalid VPI range.  Use 0 - 255\n",
                          progname);
                  exit (1);
                }
              info.valid |= FSTVAL_DSL_S1;
            }
	    else if ( strcasecmp ( argv[0],"vci") == 0 )
            {
              check_valid_card_dsl(card_type);
              info.vci = atoi(argv[1]);
              if (info.vci > 65534)
                {
                  fprintf(stderr, "%s: Invalid VCI range.  Use 0 - 65535\n",
                          progname);
                  exit (1);
                }
              info.valid |= FSTVAL_DSL_S1;
            }
	    else if ( strcasecmp ( argv[0],"snrth") == 0 )
            {
              check_valid_card_dsl(card_type);
              info.snrth = atoi(argv[1]);
              if (info.snrth > 35)
                {
                  fprintf(stderr, "%s: Invalid snrth range.  Use 0 - 35\n",
                          progname);
                  exit (1);
                }
              info.valid |= FSTVAL_DSL_S1;
            }
	    else if ( strcasecmp ( argv[0],"lpath") == 0 )
            {
              check_valid_card_dsl(card_type);
              info.lpath = atoi(argv[1]);
              if (info.lpath > 35)
                {
                  fprintf(stderr, "%s: Invalid lpath.  Use 0 - 35\n",
                          progname);
                  exit (1);
                }
              info.valid |= FSTVAL_DSL_S1;
            }
#ifdef GEN_HDLC
            /*
             * Frame Relay Configuration
             */
	    else if ( strcasecmp ( argv[0],"lmi") == 0 )
	    {
		set_proto ( IF_PROTO_FR,"lmi only valid for Frame Relay");

		if ( strcasecmp ( argv[1],"none") == 0 )
		    fr.lmi = LMI_NONE;
		else if ( strcasecmp ( argv[1],"ansi") == 0 )
		    fr.lmi = LMI_ANSI;
		else if ( strcasecmp ( argv[1],"ccitt") == 0 )
		    fr.lmi = LMI_CCITT;
		else
		{
		    fprintf ( stderr,"%s: Unknown LMI type \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
		}
	    }
	    else if ( strcasecmp ( argv[0],"dce") == 0 )
	    {
		set_proto ( IF_PROTO_FR,"dce only valid for Frame Relay");
		fr.dce = 1;

	    	--argc;
		++argv;
		continue;
	    }
	    else if ( strcasecmp ( argv[0],"t391") == 0 )
	    {
		set_proto ( IF_PROTO_FR,"t391 only valid for Frame Relay");
	    	do_num ( (int *) &fr.t391, argv[1], 1, 1000,
					"Polling verification timer (t391)");
	    }
	    else if ( strcasecmp ( argv[0],"t392") == 0 )
	    {
		set_proto ( IF_PROTO_FR,"t392 only valid for Frame Relay");
	    	do_num ( (int *) &fr.t392, argv[1], 1, 1000,
					"Link integrity polling timer (t392)");
	    }
	    else if ( strcasecmp ( argv[0],"n391") == 0 )
	    {
		set_proto ( IF_PROTO_FR,"n391 only valid for Frame Relay");
	    	do_num ( (int *) &fr.n391, argv[1], 1, 1000,
					"Full status polling counter (n391)");
	    }
	    else if ( strcasecmp ( argv[0],"n392") == 0 )
	    {
		set_proto ( IF_PROTO_FR,"n392 only valid for Frame Relay");
	    	do_num ( (int *) &fr.n392, argv[1], 1, 1000,
					"Error threshold (n392)");
	    }
	    else if ( strcasecmp ( argv[0],"n393") == 0 )
	    {
		set_proto ( IF_PROTO_FR,"n393 only valid for Frame Relay");
	    	do_num ( (int *) &fr.n393, argv[1], 1, 1000,
					"Monitored events count (n393)");
	    }
	    else if ( strcasecmp ( argv[0],"interval") == 0 )
	    {
		set_proto ( IF_PROTO_CISCO,
					"interval only valid for Cisco HDLC");
	    	do_num ( (int *) &cisco.interval, argv[1], 1, 100, "Interval");
	    }
	    else if ( strcasecmp ( argv[0],"timeout") == 0 )
	    {
		set_proto ( IF_PROTO_CISCO,
					"timeout only valid for Cisco HDLC");
	    	do_num ( (int *) &cisco.timeout, argv[1], 1, 100, "Timeout");
	    }
#endif
            /*
             * Async Configuration
             */
	    else if ( strcasecmp ( argv[0],"flow_control") == 0 )
	    {
              if (!can_do_async)
                {
                  fprintf ( stderr,"%s: Async parameter \"%s\" not allowed on %s \n",
                            progname, argv[0], card_strings[card_type] );
                      exit ( 1 );
                }
		if ( strcasecmp ( argv[1],"none") == 0 )
                  info.async_conf.flow_control = COM_FLOW_CONTROL_NONE;
                else if(strcasecmp ( argv[1],"hardware") == 0 )
                  info.async_conf.flow_control = COM_FLOW_CONTROL_RTSCTS;
                else if(strcasecmp ( argv[1],"software") == 0 )
                  info.async_conf.flow_control = COM_FLOW_CONTROL_XONXOFF;
                else
                  {
		    fprintf ( stderr,"%s: Unknown flow control type \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
                  }
		info.valid |= FSTVAL_ASYNC;
            }
	    else if ( strcasecmp ( argv[0],"stop_bits") == 0 )
	    {
              if (!can_do_async)
                {
                  fprintf ( stderr,"%s: Async parameter \"%s\" not allowed on %s \n",
                            progname, argv[0], card_strings[card_type] );
                      exit ( 1 );
                }
		if ( strcasecmp ( argv[1],"1") == 0 )
                  info.async_conf.stop_bits = COM_STOP_BITS_1;
                else if(strcasecmp ( argv[1],"1.5") == 0 )
                  info.async_conf.stop_bits = COM_STOP_BITS_1_5;
                else if(strcasecmp ( argv[1],"2") == 0 )
                  info.async_conf.stop_bits = COM_STOP_BITS_2;
                else
                  {
		    fprintf ( stderr,"%s: Unknown stop bits setting \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
                  }
		info.valid |= FSTVAL_ASYNC;
            }
	    else if ( strcasecmp ( argv[0],"parity") == 0 )
	    {
              if (!can_do_async)
                {
                  fprintf ( stderr,"%s: Async parameter \"%s\" not allowed on %s \n",
                            progname, argv[0], card_strings[card_type] );
                      exit ( 1 );
                }
		if ( strcasecmp ( argv[1],"none") == 0 )
                  info.async_conf.parity = COM_NO_PARITY;
                else if(strcasecmp ( argv[1],"odd") == 0 )
                  info.async_conf.parity = COM_ODD_PARITY;
                else if(strcasecmp ( argv[1],"even") == 0 )
                  info.async_conf.parity = COM_EVEN_PARITY;
                else
                  {
		    fprintf ( stderr,"%s: Unknown parity setting \"%s\"\n",
				progname, argv[1] );
		    exit ( 1 );
                  }
		info.valid |= FSTVAL_ASYNC;
            }
	    else if ( strcasecmp ( argv[0],"word_len") == 0 )
	    {
              if (!can_do_async)
                {
                  fprintf ( stderr,"%s: Async parameter \"%s\" not allowed on %s \n",
                            progname, argv[0], card_strings[card_type] );
                      exit ( 1 );
                }
	    	do_num ( (int *) &info.async_conf.word_length, argv[1], 5, 8,
					"Word length");
		info.valid |= FSTVAL_ASYNC;
            }
	    else if ( strcasecmp ( argv[0],"xon_char") == 0 )
	    {
              if (!can_do_async)
                {
                  fprintf ( stderr,"%s: Async parameter \"%s\" not allowed on %s \n",
                            progname, argv[0], card_strings[card_type] );
                      exit ( 1 );
                }
	    	do_num ( (int *) &info.async_conf.xon_char, argv[1], 0, 255,
					"XON character");
		info.valid |= FSTVAL_ASYNC;
            }
	    else if ( strcasecmp ( argv[0],"xoff_char") == 0 )
	    {
              if (!can_do_async)
                {
                  fprintf ( stderr,"%s: Async parameter \"%s\" not allowed on %s \n",
                            progname, argv[0], card_strings[card_type] );
                      exit ( 1 );
                }
	    	do_num ( (int *) &info.async_conf.xoff_char, argv[1], 0, 255,
					"XOFF character");
		info.valid |= FSTVAL_ASYNC;
            }
	    else
	    {
		fprintf ( stderr,"%s: Unknown configuration option \"%s\"\n",
				progname, argv[0] );
	        exit ( 1 );
	    }

	    argc -= 2;
	    argv += 2;
	}

	if ( argc != 0 )
	{
	    fprintf ( stderr,"%s: Spurious set option or missing argument\n",
			    	progname );
	    exit ( 1 );
	}

#ifdef GEN_HDLC
	if ( is_gen_hdlc )
	{
	    /* Set physical line type and speed */
	    if ( info.valid & ( FSTVAL_CABLE | FSTVAL_SPEED ))
	    {
		if ( info.valid & FSTVAL_SPEED )
		{
		    synch.clock_type = ( info.internalClock == INTCLK )
		    				? CLOCK_INT : CLOCK_EXT;
		    synch.clock_rate = info.lineSpeed;
		}
		req.ifr_settings.type = IF_IFACE_SYNC_SERIAL;
		if ( info.valid & FSTVAL_CABLE )
		{
		    switch ( info.lineInterface )
		    {
		    case V35:	req.ifr_settings.type = IF_IFACE_V35;	break;
		    case V24:	req.ifr_settings.type = IF_IFACE_V24;	break;
		    case X21:	req.ifr_settings.type = IF_IFACE_X21;	break;
		    case X21D:	req.ifr_settings.type = IF_IFACE_X21D;	break;
		    case T1:	req.ifr_settings.type = IF_IFACE_T1;	break;
		    case E1:	req.ifr_settings.type = IF_IFACE_E1;	break;
		    case SHDSL:	req.ifr_settings.type = IF_IFACE_SHDSL;	break;
                    //case J1:	req.ifr_settings.type = IF_IFACE_J1;	break;
                    case RS530_449: req.ifr_settings.type = IF_IFACE_RS530_449;	break;
                    case RS485:	req.ifr_settings.type = IF_IFACE_RS485;	break;
		    }
		}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	    req.ifr_settings.size = sizeof ( synch );
	    req.ifr_settings.ifs_ifsu.sync = &synch;
            cmd = SIOCWANDEV;
#else
		req.ifr_settings.data_length = sizeof ( synch );
		req.ifr_settings.data = &synch;
            cmd = SIOCDEVICE;
#endif
		if ( ioctl ( sock, cmd, &req ))
		{
		    fprintf ( stderr,"%s: Problem setting line and clock type."
			    	" %s\n", progname, strerror ( errno ));
		    exit ( 1 );
		}
	    }

            /* Set protocol */
	    if  ( protocol != 0 )
	    {
		req.ifr_settings.type = protocol;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
		req.ifr_settings.size = 0;
#else
		req.ifr_settings.data_length = 0;
#endif

		/* Add in protocol specific information defaulting any fields
		 * as necessary
		 */
		switch ( protocol )
		{
		case IF_PROTO_FR:
			if ( fr.t391 == 0 ) fr.t391 = 10;
			if ( fr.t392 == 0 ) fr.t392 = 15;
			if ( fr.n391 == 0 ) fr.n391 = 6;
			if ( fr.n392 == 0 ) fr.n392 = 3;
			if ( fr.n393 == 0 ) fr.n393 = 4;
			if ( fr.lmi  == 0 ) fr.lmi  = LMI_ANSI;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
			req.ifr_settings.ifs_ifsu.sync = (sync_serial_settings*) &fr;
		    	req.ifr_settings.size = sizeof ( fr );
#else
			req.ifr_settings.data =  &fr;
		    	req.ifr_settings.data_length = sizeof ( fr );
#endif
			break;

		case IF_PROTO_CISCO:
			if ( cisco.interval == 0 ) cisco.interval = 10;
			if ( cisco.timeout  == 0 ) cisco.timeout  = 25;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
			req.ifr_settings.ifs_ifsu.sync = (sync_serial_settings*) &cisco;
			req.ifr_settings.size = sizeof ( cisco );
#else
			req.ifr_settings.data = &cisco;
			req.ifr_settings.data_length = sizeof ( cisco );
#endif
			break;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,22)
                  case IF_PROTO_HDLC:
                  case IF_PROTO_HDLC_ETH:
                        raw_hdlc.encoding = ENCODING_DEFAULT;
                        raw_hdlc.parity = PARITY_DEFAULT;
                        req.ifr_settings.ifs_ifsu.raw_hdlc = &raw_hdlc;
                        req.ifr_settings.size = sizeof(raw_hdlc);
                        break;
#endif
                }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
                cmd = SIOCWANDEV;
#else
                cmd = SIOCDEVICE;
#endif
		if ( ioctl ( sock, cmd, &req ))
		{
		    fprintf ( stderr,"%s: Problem setting protcol. %s\n",
			    	progname, strerror ( errno ));
		    fprintf ( stderr,"%s: Have you enabled this protocol in the Kernel configuration?\n", progname);
		    exit ( 1 );
		}
	    }

	    /* Set debug and card location mode */
	    info.valid &= ~( FSTVAL_CABLE | FSTVAL_SPEED | FSTVAL_PROTO );
	    if ( info.valid != 0 )
	    {
		req.ifr_data = (char *)&info;
		if ( ioctl ( sock, FSTSETCONF, &req ))
		{
		    fprintf ( stderr,"%s: Problem setting port mode. %s\n",
				    progname, strerror ( errno ));
		    exit ( 1 );
		}
	    }
	}
	else /* Tricksy drop through to non-ifdef case */
#endif
	{
	    req.ifr_data = (char *)&info;
	    if ( ioctl ( sock, FSTSETCONF, &req ))
	    {
	        fprintf ( stderr,"%s: Problem setting port configuration. %s\n",
				progname, strerror ( errno ));
	        exit ( 1 );
	    }

            if ( info.valid & ( FSTVAL_CABLE | FSTVAL_SPEED ))
            {
                if ( ioctl ( sock, SIOCGIFFLAGS, &req ) < 0 )
                {
                    fprintf ( stderr,"%s: Warning checking interface flags. %s\n",
                                progname, strerror ( errno ));
                }

                if ( req.ifr_flags & IFF_UP )
                {
                  printf ("%s: Warning: Interface %s is up.\n"
                          "Speed and media changes may not take effect until"
                          " interface is restarted.\n", progname, ifname );
                }
            }
	}
}


#if GEN_HDLC
/*
 *	Add or remove a PVC from the master device
 */
static void
do_pvc ( int command, int ether, char *arg )
{
	fr_proto_pvc pvc;
        unsigned int cmd;

	do_num ( (int *) &pvc.dlci, arg, 0, 1023, "PVC DLCI");

	req.ifr_settings.type = command;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	req.ifr_settings.ifs_ifsu.sync = (sync_serial_settings*) &pvc;
	req.ifr_settings.size = sizeof ( pvc );
        cmd = SIOCWANDEV;
#else
	req.ifr_settings.data = &pvc;
	req.ifr_settings.data_length = sizeof ( pvc );
        cmd = SIOCDEVICE;
#endif

	if ( ioctl ( sock, cmd, &req ))
	{
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,22)
          if (ether)
	    fprintf ( stderr,"%s: Problem %s Ethernet PVC. %s\n", progname,
		    command == IF_PROTO_FR_ADD_ETH_PVC ? "creating":"deleting",
		    strerror ( errno ));
          else
#endif
	    fprintf ( stderr,"%s: Problem %s PVC. %s\n", progname,
		    command == IF_PROTO_FR_ADD_PVC ? "creating":"deleting",
		    strerror ( errno ));
          exit ( 1 );
	}
}
#endif


/*
 *	Display status info for port
 */

#define do_fromtable(P,T,V)			\
	if ( V >= sizeof (T) / sizeof (char *))	\
	    printf ("%sUnknown (%d)\n", P, V );	\
	else					\
	    printf ("%s%s\n", P, T[V] )

#define do_fromtableX(P,T,V)			\
	if ( V >= sizeof (T) / sizeof (char *))	\
	    printf ("%sUnknown (%d)\n", P, V );	\
	else					\
	    printf ("%s%s ", P, T[V] )

#define do_bit(V,B,S)				\
	if ( V & B ) {				\
	    if ( done_bit )			\
		printf (", ");			\
	    printf ( S );			\
	    done_bit = 1;			\
	} else /* Swallow ; */

static void
do_te1_info (struct fstioc_info info)
{
  if (info.type != FST_TYPE_TE1)
    {
      do_fromtable ("  cable status:\t\t", cable_strings, info.cableStatus );
    }
  printf ("  active alarms:\t");
  if ( (info.lossOfSignal) || 
       (info.receiveRemoteAlarm) || 
       (info.alarmIndicationSignal) )
    {
      if (info.lossOfSignal)
        printf("LOS ");
      if (info.receiveRemoteAlarm)
        printf("RRA  ");
      if (info.alarmIndicationSignal)
        printf("AIS ");
    }
  else
    printf("None");
  printf("\n");

  printf ("  clock mode:\t\t");
  if (info.clockSource == CLOCKING_SLAVE)
    printf ("Slave\n");
  else
    printf("Master\n");

  printf ("  datarate:\t\t%d\n", info.lineSpeed );
  printf ("  start slot:\t\t%d\n", info.startingSlot);
  do_fromtable ("  structure:\t\t", structure_strings, info.structure );

  if (info.interface == INTERFACE_RJ48C)
    printf("  interface:\t\tRJ48C\n");
  else
    printf("  interface:\t\tBNC\n");

  do_fromtable ("  coding:\t\t", coding_strings, info.coding );

  if (info.equalizer == EQUALIZER_SHORT)
    printf("  equalizer:\t\tShort Haul\n");
  else
    {
      printf("  equalizer:\t\tLong Haul\n");
      if (info.lineInterface == T1)
        {
          do_fromtable ("  line build out:\t", lbo_strings, info.lineBuildOut );
        }
    }

  do_fromtable ("  loop mode:\t\t", loop_strings, info.loopMode );
  if (info.lineInterface == T1)
    {
      do_fromtable ("  range:\t\t", range_strings, info.range );
    }
  do_fromtable ("  tx buff mode:\t\t", buff_strings, info.txBufferMode );
  do_fromtable ("  rx buff mode:\t\t", buff_strings, info.rxBufferMode );
  if (info.idleCode)
    printf("  idle code:\t\t0x%X\n", info.idleCode);
  else
    printf("  idle code:\t\tDefault\n");
}

static void
do_te1_stats (struct fstioc_info info)
{
  printf("Stats:\n");
  printf("  Receive Buffer Delay:\t%d\n", info.receiveBufferDelay);
  printf("  Framing Error Count:\t%d\n", info.framingErrorCount);
  printf("  Code Violation Count:\t%d\n", info.codeViolationCount);
  printf("  CRC Error Count:\t%d\n", info.crcErrorCount);
  if (info.lineAttenuation < -2147483647)
    printf("  Line Attenuation:\tUnknown\n");
  else
    printf("  Line Attenuation:\t%d\n", info.lineAttenuation);
}

static void
do_shdsl_info (struct fstioc_info info)
{
  if (info.terminalType == EQUIP_TYPE_REMOTE)
    {
      printf ("  Term Type:\tRemote Office");
    }
  else
    {
      printf ("  Term Type:\tCentral Office");
    }
  if (info.annexType == ANNEX_B)
    {
      printf ("\t\t  Annex:\t\t\tEurope\n");
    }
  else
    {
      printf ("  Annex:\t\t\tUS\n");
    }
  if (info.encap == ENCAP_PPP)
    {
      printf ("  Encap:\tPPP\n");
    }
  else
    {
      printf ("  Encap:\tMPOA\n");
    }
  printf ("  ATM VPI:\t%d\t\t\t", info.vpi);
  printf ("  ATM VCI:\t\t\t%d\n", info.vci);
  printf ("  snrth:\t%d\t\t\t", info.snrth);
  printf ("  lpath:\t\t\t%d\n", info.lpath);
  printf ("  Test Mode:\t%d\t\t\t", info.testMode);
  printf ("  Power Backoff:\t\t%d\n", info.backoff);
  printf ("  Datarate:\t%d\t\t\t", info.lineSpeed );
  printf ("  Actual Datarate:\t\t%d\n", (info.negotiatedBChannels*64000) +
	  (info.negotiatedZBits*8*1000));

}

static void
do_shdsl_stats (struct fstioc_info info)
{
  printf("Status:\n");
  if (info.activationStatus >= 0x40)
    printf("  Activation Status:  Data Xfer (0x%x)", info.activationStatus);
  else
    printf("  Activation Status:  Training (0x%x)", info.activationStatus);
  do_fromtable("\t  Last Status:\t", activation_strings, info.activationStatus);
  printf("  No Common Mode Status:\t%d\n", info.noCommonModeStatus);
  printf("  Transceiver Status1:\t\t%d\t", info.transceiverStatus1);
  printf("  Transceiver Status2:\t\t%d\n", info.transceiverStatus2);
  printf("  Line Loss:\t\t\t%d dB\t", info.lineLoss);
  printf("  Signal Quality:\t\t%d dB\n", info.signalQuality);

  printf("Stats:\n");
  printf("  Near End Block Err Cnt:\t%d\t", info.nearEndBlockErrorCount);
  printf("  Signal To Noise Ratio:\t%d dB\n", info.signalToNoiseRatio);
  printf("  Code Violation Count:\t\t%d\t", info.codeViolationCount);
  printf("  Errored Second Count:\t\t%d\n", info.erroredSecondCount);
  printf("  Severely Errored Second Cnt:\t%d\t", info.severelyErroredSecondCount);
  printf("  Loss of Sync Word Secnd Cnt:\t%d\n", info.lossOfSyncWordSecondCount);
  printf("  Unavailable Second Cnt:\t%d\t", info.unavailableSecondCount);
  printf("  Frequency deviation:\t\t%d ppm\n", info.frequencyDeviation);
  printf("  ATM Cells transmitted:\t%d\t", info.atmTxCellCount);
  printf("  ATM Cells received:\t\t%d\n", info.atmRxCellCount);
  printf("  ATM Cell HEC count:\t\t%d\t", info.atmHecErrorCount);
  printf("  ATM Cells Dropped:\t\t%d\n", info.atmCellsDropped);
  printf("Negotiated Parameters\n");
  printf("  Negotiated Power Backoff:\t%d dB\t", info.negotiatedPowerBackOff);
  printf("  Negotiated PSD:\t\t%d\n", info.negotiatedPSD);
  printf("  Negotiated B Channels:\t%d\t", info.negotiatedBChannels);
  printf("  Negotiated Z Bits:\t\t%d\n", info.negotiatedZBits);
  printf("  Negotiated Sync Word:\t\t0x%x\t", info.negotiatedSyncWord);
  printf("  Negotiated Stuff Bits:\t%d\n", info.negotiatedStuffBits);
  printf("  Chip Version:\t\t\t%d\t", info.chipVersion);
  printf("  Firmware Version:\t\t0x%x\n", info.firmwareVersion&0x7f);
  printf("  ROM Version:\t\t\t%d\t", info.romVersion);
  printf("  XPLD Version:\t\t\t%x.%x\n", info.xpldVersion>>4, 
	 info.xpldVersion&0xf);
  printf("Far End Vendor Information\n");
  printf("  Country Code:\t\t\t0x%02x%02x\t", info.farEndCountryCode[0], 
         info.farEndCountryCode[1]);
  printf("  Vendor Info:\t\t\t0x%02x%02x\n", info.farEndVendorInfo[0], 
         info.farEndVendorInfo[1]);
  printf("  Provider Code:\t\t0x%02x%02x%02x%02x", info.farEndProviderCode[0],
	 info.farEndProviderCode[1], info.farEndProviderCode[2],
	 info.farEndProviderCode[3]);
  printf("  (%s)\n", info.farEndProviderCode);
  printf("  Utopia ATM Status byte:\t%d\n", info.utopiaAtmStatus);
}

/*
 * Print port stats
 */
static void
do_port_stats ( struct fstioc_info info )
{
  printf ("Port Statistics:\n");
  printf ("\tRx packets\t%lu", info.stats.rx_packets);
  printf ("\t\tRx bytes\t%lu\n", info.stats.rx_bytes);
  printf ("\tTx packets\t%lu", info.stats.tx_packets);
  printf ("\t\tTx bytes\t%lu\n", info.stats.tx_bytes);
  printf ("\tRx errors\t%lu", info.stats.rx_errors);
  printf ("\t\tTx errors\t%lu\n", info.stats.tx_errors);
  printf ("\tRx dropped\t%lu", info.stats.rx_dropped);
  printf ("\t\tTx dropped\t%lu\n", info.stats.tx_dropped);

  printf("\nRx errors detail:\t\tTx errors detail:\n");
  printf ("\tRx length\t%lu", info.stats.rx_length_errors);
  printf ("\t\tTx aborted\t%lu\n", info.stats.tx_aborted_errors);
  printf ("\tRx over run\t%lu", info.stats.rx_over_errors);
  printf ("\t\tTx carrier\t%lu\n", info.stats.tx_carrier_errors);
  printf ("\tRx crc \t\t%lu", info.stats.rx_crc_errors);
  printf ("\t\tTx fifo\t\t%lu\n", info.stats.tx_fifo_errors);
  printf ("\tRx frame\t%lu", info.stats.rx_frame_errors);
  printf ("\t\tTx under run\t%lu\n", info.stats.tx_underrun_errors);
  printf ("\tRx fifo\t\t%lu\n", info.stats.rx_fifo_errors);
}

static void
do_info ( int full )
{
	struct fstioc_info info;
	int done_bit;

#ifdef GEN_HDLC
	union {
		fr_proto fr;
		cisco_proto cisco;
	} proto;

	bool proto_set = TRUE;
        unsigned int cmd;
#endif
        fprintf(stderr, "\nFarutil for FarSync version %s\n\n", FST_USER_VERSION);
	/* Zero structure before use. Good practice anyway but in this case it
	 * allows for future expansion where various fields (valid for instance)
	 * might be used to indicate an extended structure.
	 */
	memset ( &info, 0, sizeof ( info ));
	req.ifr_data = (char *)&info;
	if ( ioctl ( sock, FSTGETCONF, &req ))
	{
	    fprintf ( stderr,"%s: Problem getting port information. %s\n",
				progname, strerror ( errno ));
	    exit ( 1 );
	}
#ifdef GEN_HDLC
	/* If port is flagged as using generic HDLC ioctls then we need another
	 * ioctl to get the protocol type. Everything else we use in still in
	 * the old config structure.
	 */
	if ( info.proto == FST_GEN_HDLC )
	{
	    req.ifr_settings.type = IF_GET_PROTO;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
	    req.ifr_settings.ifs_ifsu.sync = (sync_serial_settings*) &proto;
	    req.ifr_settings.size = sizeof ( proto );
            cmd = SIOCWANDEV;
#else
	    req.ifr_settings.data = &proto;
	    req.ifr_settings.data_length = sizeof ( proto );
            cmd = SIOCDEVICE;
#endif
	    if ( ioctl ( sock, cmd, &req ))
	    {
		if ( errno == ENOSYS )
		{
		    printf("Warning: No line protocol set\n");
		    proto_set = FALSE;
		}
		else
		{
		    fprintf ( stderr,"%s: Problem getting generic HDLC"
				" information. %s\n", progname,
				strerror ( errno ));
		    exit ( 1 );
		}
	    }
	}
#endif

	if ( full && ( info.valid & FSTVAL_ALL ) != FSTVAL_ALL )
	{
	    printf ("Warning: Some data has not be marked as valid and may"
				" be out of date\n\n");
	}

	/* Display basic card info */
	if ( ! full && ( info.valid & FSTVAL_CARD ) == 0 )
	{
	    printf ("Card type is invalid\n");
	    exit ( 1 );
	}

	do_fromtable ("card:\t\t", type_strings, info.type );
	printf ("ports:\t\t%d\n", info.nports );
	do_fromtable ("state:\t\t", state_strings, info.state );

	if ( ! full )
	    return;
        if ( full == 2)
          goto stats;

	printf ("\nfirmware id:\t%d",
				( info.smcFirmwareVersion >> 24 ) & 0xFF );
	printf ("\tfirmware vers:\t%d.%02d.%02d\n",
				( info.smcFirmwareVersion >> 16 ) & 0xFF,
				( info.smcFirmwareVersion >>  8 ) & 0xFF,
				( info.smcFirmwareVersion       ) & 0xFF );
        if (info.asyncAbility)
          printf("Async ports present");
        if (info.synthAbility)
          printf("\tClock Synthesiser present\n");

#ifdef GEN_HDLC
	if ( info.proto == FST_GEN_HDLC && verbose )
	{
	    printf ("driver feature:\tsupports generic HDLC interface\n");
	}
#endif

	printf ("\nConfiguration for port %d \n", info.index );

	do_fromtableX ("physical:\t", phys_strings, info.lineInterface);
        if (info.lineInterface == UX35C)
          printf("(UX35C converter cable is required)");
        if ((info.type != FST_TYPE_FLEX1) && 
            (info.type != FST_TYPE_T4E) && 
            (info.type != FST_TYPE_T4Ep))
          {
            if ((info.lineInterface ==  RS530_449) &&
                (info.internalClock == INTCLK))
              printf("(X21-449-INT converter cable is required)");
            if ((info.lineInterface ==  RS530_449) && 
                (info.internalClock != INTCLK))
              printf("(X21-449-EXT cable is required)");
          }
        printf("\n");
        if ( (info.lineInterface == T1) ||(info.lineInterface == E1) || 
             (info.lineInterface == J1) )
          {
            do_te1_info(info);
            do_te1_stats(info);
          }
        else if ( info.lineInterface == SHDSL )
          {
            do_shdsl_info(info);
            do_shdsl_stats(info);
          }
        else
          {
            if (info.type != FST_TYPE_FLEX1)
              {
                do_fromtable ("  cable status:\t\t", cable_strings, info.cableStatus );
              }
            printf ("  active inputs:\t");
            done_bit = 0;
            if ( ( info.lineInterface == X21 ) || 
                 (info.lineInterface == X21D) )
              do_bit ( info.v24IpSts, IPSTS_CTS,"Indicate");
            else
              do_bit ( info.v24IpSts, IPSTS_CTS,"CTS");
            do_bit ( info.v24IpSts, IPSTS_DSR,"DSR");
            do_bit ( info.v24IpSts, IPSTS_DCD,"DCD");
            do_bit ( info.v24IpSts, IPSTS_RI, "RI");
            do_bit ( info.v24IpSts, IPSTS_TMI,"Test mode");
            printf ( done_bit ? " ":"None");
            if (info.ignoreCarrier)
              {
                printf ("  Ignoring Modem Signals");
              }
            printf("\n");
            printf ("  active outputs:\t");
            done_bit = 0;
            if ( ( info.lineInterface == X21 ) || 
                 (info.lineInterface == X21D) )
              {
                do_bit ( info.v24OpSts, OPSTS_RTS,"Control");
              }
            else
              {
                do_bit ( info.v24OpSts, OPSTS_RTS,"RTS");
                do_bit ( info.v24OpSts, OPSTS_DTR, "DTR");
                do_bit ( info.v24OpSts, OPSTS_DSRS,"DSRS");
              }
            do_bit ( info.v24OpSts, OPSTS_SS,  "Standby");
            do_bit ( info.v24OpSts, OPSTS_LL,  "Test mode");
            printf ( done_bit ? "\n":"None\n");
                        
	    if (info.extendedClocking)
	      {
                printf("  extended clocking: \t");
		if (info.extendedClocking & 0x08)
		  {
		    printf("tt tx : ");
		  }
		if (info.extendedClocking & 0x04)
		  {
		    printf("tt rx : ");
		  }
		switch (info.extendedClocking & 0x03)
		  {
		  case 0x00:
		    printf("ext rx : ext tx ");
		    break;

		  case 0x01:
		    printf("ext rx : int tx ");
		    break;

		  case 0x02:
		    printf("int rx : ext tx ");
		    break;

		  case 0x03:
		    printf("int rx : int tx ");
		    break;

		  default:
		    printf("Invalid %x", info.extendedClocking);
		  }
		if (info.clockStatus)
		  printf("(not detected)\n");
		else
		  printf("(detected)\n");
	      }
	    else
	      {            
                if ( info.internalClock == INTCLK )
                  {
                    if ( info.clockStatus )
                      printf ("  clock:\t\tInternal but not detected ?\n");
                    else
                      if (info.invertClock)
                        printf ("  clock:\t\tInternal/Inverted detected\n");
                      else
                        printf ("  clock:\t\tInternal detected\n");
                  }
                else
                  {
                    if ( info.clockStatus )
                      printf ("  clock:\t\tExternal, Not detected\n");
                    else
                      {
                        if (info.invertClock)
                          printf ("  clock:\t\tExternal/Inverted, Detected\n");
                        else
                          printf ("  clock:\t\tExternal, Detected\n");
                      }
                  }
              }
            if ((!info.lineSpeed) && 
                ((info.type == FST_TYPE_T4E) ||
                 (info.type == FST_TYPE_T4Ep)))
              printf ("  Estimated Speed:\t%d\n", info.estLineSpeed);
            else
              printf ("  speed:\t\t%d\n", info.lineSpeed );
            if ( (info.type == FST_TYPE_FLEX1) || 
                 (info.type == FST_TYPE_T4Ep) )
              {
                printf ("  encoding:\t\t");
                switch (info.coding)
                  {
                  case CODING_NRZ:
                    printf("nrz\n");
                    break;
                  case CODING_NRZI:
                    printf("nrzi\n");
                    break;
                  case CODING_FM0:
                    printf("fm0\n");
                    break;
                  case CODING_FM1:
                    printf("fm1\n");
                    break;
                  case CODING_MANCHESTER:
                    printf("Manchester\n");
                    break;
                  default:
                    printf("Not known (%d)\n", info.coding);
                  }
                printf ("  NRZIClocking:\t\t");
                if (info.enableNRZIClocking)
                  printf("Enabled\n");
                else
                  printf("Disabled\n");
                printf ("  termination:\t\t");
                if (info.termination)
                  printf("Resistive\n");
                else
                  printf("None\n");
              }
            if (info.lowLatency)
              {
                do_fromtable("  Low Latency:\t\t", latency_strings, 
                             info.lowLatency);
              }
          }


#ifdef GEN_HDLC
	if ( info.proto == FST_GEN_HDLC )
	{
	    printf ("protocol:\t");
	    switch ( req.ifr_settings.type )
	    {
	    case IF_PROTO_FR:
		    printf ("Frame Relay\n");
		    switch ( proto.fr.lmi )
		    {
		    case LMI_ANSI:	printf ("  LMI:\tANSI");    break;
		    case LMI_CCITT:	printf ("  LMI:\tCCITT");   break;
		    case LMI_NONE:	printf ("  LMI:\tNONE");    break;
		    }
		    if ( proto.fr.dce )
		    	printf ("\tDCE");
		    printf ("\n");
		    printf ("  T391:\t%d\n", proto.fr.t391 );
		    printf ("  T392:\t%d\n", proto.fr.t392 );
		    printf ("  N391:\t%d\n", proto.fr.n391 );
		    printf ("  N392:\t%d\n", proto.fr.n392 );
		    printf ("  N393:\t%d\n", proto.fr.n393 );
		    break;

	    case IF_PROTO_CISCO:
	    	    printf ("Cisco compatible HDLC\n");
		    printf ("  Interval:\t%d\n", proto.cisco.interval );
		    printf ("  Timeout:\t%d\n", proto.cisco.timeout );
		    break;

	    case IF_PROTO_HDLC:  printf ("HDLC\n");			break;
	    case IF_PROTO_PPP:   printf ("Sync PPP\n");			break;
	    case IF_PROTO_X25:   printf ("X.25\n");			break;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,22)
	    case IF_PROTO_HDLC_ETH: printf ("HDLC_ETH\n");              break;
#endif
	    case IF_PROTO_RAW:      printf ("FarSync raw access\n");    break;

	    default:
		    if ( proto_set )
		    {
			 printf ("Unknown generic HDLC code 0x%04X\n",
					 req.ifr_settings.type );
	    	    }
		    else
		    {
			 printf ("Not set\n");
		    }
		    break;
	    }
	}
	else	/* Tricksy fall through to non-ifdef case */
#endif
	{
	    do_fromtable ("protocol:\t", proto_strings, info.proto );
	}
        do_fromtable ("line mode:\t", mode_strings, info.transparentMode);
        if ((info.asyncAbility) && (info.transparentMode == FST_MODE_ASYNC))
          {
            printf("Async Configuration:\n");
            do_fromtable("  flow control:\t\t", flow_strings, info.async_conf.flow_control);
            do_fromtable("  stop bits:\t\t", stop_strings, info.async_conf.stop_bits);
            do_fromtable("  parity:\t\t", parity_strings, info.async_conf.parity);
            printf("  word length:\t\t%d\n", info.async_conf.word_length);
            printf("  XON character:\t%d\t", info.async_conf.xon_char);
            printf("XOFF character:\t%d\n", info.async_conf.xoff_char);
          }
        printf("Buffer configuration:\n");
        printf("  no of rx buffers:\t%d\t", info.numRxBuffers);
        printf("size of rx buffers:\t%d\n", info.rxBufferSize);
        printf("  no of tx buffers:\t%d\t", info.numTxBuffers);
        printf("size of tx buffers:\t%d\n", info.txBufferSize);
	if ( info.valid & FSTVAL_DEBUG )
          {
	    printf ("\ndebug:\t\t0x%04X\n", info.debug );
            printf("Kernel Version:\t%lx\n", info.kernelVersion); 
          }
        if ( full == 1)
          return;
 stats:
        do_port_stats(info);
}

static int
get_config_device (char * device_name, char * next_arg)
{
        static struct fstioc_cmd my_cmd;
        unsigned char iocinfo_version;
        char char_dev_name[32];

        /* We come in with a socket, and we may leave with a socket
         * or an fd depending on what happens
         */
        strcpy(char_dev_name, device_name);
	my_cmd.command = FSTCMD_GET_VERSION;
        my_cmd.version = 1;
	my_cmd.output_data_len=1;
        my_cmd.status = 0x55555555;
	my_cmd.data_ptr = &iocinfo_version;
	req.ifr_data = (char *)&my_cmd;
	if (ioctl(sock, FSTCMD, &req))
	  {
            /*
             * Ok, so there probably wasn't a network device
             * Try fro the char device in /dev next
             */
            close(sock);
            strcpy(char_dev_name, "/dev/");
            strcat(char_dev_name, device_name);
            if ((sock = open (char_dev_name, O_RDWR)) < 0)
              {
                /*
                 * Only once choice left now
                 */
                strcpy(char_dev_name, "/dev/tty_");
                strcat(char_dev_name, device_name);
                if ((sock = open (char_dev_name, O_RDWR)) < 0)
                  {
                    if ( next_arg )
                      {
                        if (strcmp ( next_arg,"shellinfo") == 0)
                          return sock;
                      }
                    printf("Cannot find device to configure <%s>\n", 
                           device_name);
                    exit(0);
                  }
              }
	    if (ioctl(sock, FSTCMD, &req))
            {
              /*
               * One of the Char devices openend, but didn't accept the
               * command
               */
              printf("Unexpected ioctl result, errno is %d\n", errno);
            }
	  }
        if (my_cmd.status == 0xaaaaaaaa)
          {
            //printf("Using device %s for configuration\n", char_dev_name);
          }
        else
          {
            printf("Error in querying configuration device.  Status is %x\n", 
                   my_cmd.status);
            exit(0);
          }
        return sock;
}

static void
do_get_iocver (void)
{
        static struct fstioc_cmd my_cmd;
        unsigned char iocinfo_version;

	my_cmd.command = FSTCMD_GET_VERSION;
        my_cmd.version = 1;
	my_cmd.output_data_len=1;
        my_cmd.status = 0x55555555;
	my_cmd.data_ptr = &iocinfo_version;
	req.ifr_data = (char *)&my_cmd;
	if (ioctl(sock, FSTCMD, &req))
	  {
	    perror("Error in getting IOC Info structure version number\n");
	    exit(0);
	  }
        if (my_cmd.status == 0xaaaaaaaa)
          {
            printf("IOC Info structure version returned is %d\n", iocinfo_version);
          }
        else
          {
            printf("Error processing IOC Info version.  Status is %x\n", my_cmd.status);
          }
}

static void
do_set_iocver (char * argp)
{
        static struct fstioc_cmd my_cmd;
        unsigned char iocinfo_version;

        iocinfo_version = atoi(argp);
        printf("Requesting IOC Info structure version to be %d\n",
          iocinfo_version);
	my_cmd.command = FSTCMD_SET_VERSION;
        my_cmd.version = 1;
	my_cmd.input_data_len=1;
        my_cmd.status = 0x55555555;
	my_cmd.data_ptr = &iocinfo_version;
	req.ifr_data = (char *)&my_cmd;
	if (ioctl(sock, FSTCMD, &req))
	  {
	    perror("Error in setting IOC Info version number\n");
	    exit(0);
	  }
        if (my_cmd.status == 0xaaaaaaaa)
          {
            printf("IOC Info version set to %d\n", iocinfo_version);
          }
        else
          {
            printf("Error processing IOC Info version.  Status is %x\n", my_cmd.status);
          }
}

static void
do_reset_stats (void)
{
        static struct fstioc_cmd my_cmd;

	my_cmd.command = FSTCMD_RESET_STATS;
        my_cmd.version = 1;
	my_cmd.input_data_len=0;
        my_cmd.status = 0x55555555;
	my_cmd.data_ptr = NULL;
	req.ifr_data = (char *)&my_cmd;
	if (ioctl(sock, FSTCMD, &req))
	  {
	    perror("Error in resetting port stats \n");
	    exit(0);
	  }
        if (my_cmd.status == 0xaaaaaaaa)
          {
	    do_info ( 2 );
          }
        else
          {
            printf("Error processing stats reset.  Status is %x\n", my_cmd.status);
          }
}

struct fst_ints {
  unsigned long int_186;
  unsigned long int_txdma;
  unsigned long int_rxdma;
  unsigned long int_no_work;
};

static void
do_get_intstats (void)
{
        static struct fstioc_cmd my_cmd;
        struct fst_ints card_ints;
	struct fstioc_info info;

	memset ( &info, 0, sizeof ( info ));
	req.ifr_data = (char *)&info;
	if ( ioctl ( sock, FSTGETSHELL, &req ))
	{
	    printf ("CARD=error PORTS=0 STATE=0 INDEX=error\n");
	    exit ( 0 );
	}

	if ( info.type == FST_TYPE_FLEX1 )
          {
            printf("Sorry, Interrupt stats not available on FarSync Flex\n");
            exit(0);
          }
	my_cmd.command = FSTCMD_GET_INTS;
        my_cmd.version = 1;
	my_cmd.output_data_len=sizeof(struct fst_ints);
        my_cmd.status = 0x55555555;
	my_cmd.data_ptr = (unsigned char *)&card_ints;
	req.ifr_data = (char *)&my_cmd;
	if (ioctl(sock, FSTCMD, &req))
	  {
	    perror("Error in getting card interrupt stats\n");
	    exit(0);
	  }
        if (my_cmd.status == 0xaaaaaaaa)
          {
            printf("Card Interrupts returned as:\n");
            printf("186         %lu\n", card_ints.int_186);
            printf("txdma       %lu\n", card_ints.int_txdma);
            printf("rxdma       %lu\n", card_ints.int_rxdma);
            printf("no work     %lu\n", card_ints.int_no_work);
          }
        else
          {
            printf("Error processing interrupt stats.  Status is %x\n", my_cmd.status);
          }
}
static void
do_reset_intstats (void)
{
        static struct fstioc_cmd my_cmd;
	struct fstioc_info info;

	memset ( &info, 0, sizeof ( info ));
	req.ifr_data = (char *)&info;
	if ( ioctl ( sock, FSTGETSHELL, &req ))
	{
	    printf ("CARD=error PORTS=0 STATE=0 INDEX=error\n");
	    exit ( 0 );
	}

	if ( info.type == FST_TYPE_FLEX1 )
          {
            printf("Sorry, Interrupt stats not available on FarSync Flex\n");
            exit (0);
          }
	my_cmd.command = FSTCMD_RESET_INTS;
        my_cmd.version = 1;
	my_cmd.input_data_len=0;
        my_cmd.status = 0x55555555;
	my_cmd.data_ptr = NULL;
	req.ifr_data = (char *)&my_cmd;
	if (ioctl(sock, FSTCMD, &req))
	  {
	    perror("Error in resetting card interrupt stats\n");
	    exit(0);
	  }
}

/* Abrieviated card information in a format suitable for interpretation by a
 * script, with no error returns other than the information itself. This is
 * intended for scripts only and hence is not documented in the user interface
 * documents
 */
static void
do_shellinfo ( void )
{
	struct fstioc_info info;

	memset ( &info, 0, sizeof ( info ));
	req.ifr_data = (char *)&info;
	if ( ioctl ( sock, FSTGETSHELL, &req ))
	{
	    printf ("CARD=error PORTS=0 STATE=0 INDEX=error\n");
	    exit ( 0 );
	}

	if ( info.type == FST_TYPE_T2P )
	{
	    printf ("CARD=T2P PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_T4P )
	{
	    printf ("CARD=T4P PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_T1U )
	{
	    printf ("CARD=T1U PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_T2U )
	{
	    printf ("CARD=T2U PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_T4U )
	{
	    printf ("CARD=T4U PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_T4E )
	{
	    printf ("CARD=T4E PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_T4Ep )
	{
	    printf ("CARD=T4E PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_T2UE )
	{
	    printf ("CARD=T2UE PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_T4UE )
	{
	    printf ("CARD=T4UE PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_TE1 )
	{
	    printf ("CARD=TE1 PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_DSL_S1 )
	{
	    printf ("CARD=HDSL PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else if ( info.type == FST_TYPE_FLEX1 )
	{
	    printf ("CARD=Flex PORTS=%d STATE=%d INDEX=%d\n", info.nports,
			    		info.state, info.index );
	}
	else
	{
	    printf ("CARD=error PORTS=0 STATE=0 INDEX=error\n");
	}
}


static void
usage ( void )
{
        fprintf ( stderr, "\nFarutil for FarSync version %s\n\n", FST_USER_VERSION);
        fprintf ( stderr, "usage: %s [-v] interface command [parameters...]\n"
	"\n"
	"Commands:\n"
	"\tcard                       #Display card type and number of ports\n"
	"\tinfo                       #Display status info for port\n"
	"\tstats                      #Display statistics info for port\n"
	"\tstats_reset                #Reset statistics info for port\n"
	"\tload firmware1 firmware2   #Download firmware and start card\n"
#ifdef GEN_HDLC
	"\tcreate dlci                #Add PVC to master device\n"
	"\tdelete dlci                #Remove PVC from device\n"
	"\tcreate_eth dlci            #Add Ethernet bridged PVC to master device\n"
	"\tdelete_eth dlci            #Remove Ethernet bridged PVC from device\n"
#endif
	"\tset options...             #Set configurable options\n"
	"\n"
	"Configuration options:\n"
	"\tmedia {x21|v35|v24|rs530|  #Physical media type\n"
        "\t       rs449|E1|T1|SHDSL} \n"
#ifdef GEN_HDLC
	"\tproto {ppp|cisco|fr|x25|hdlc|hdlc_eth|raw}\n"
        "\t                           #Transport protocol\n"
#elif !defined(__FreeBSD__)
	"\tproto {ppp|cisco|raw}      #Transport protocol\n"
#endif
	"\tclock {ext|rate}           #External clock, or internal speed\n"
        "\tphase {normal|inverted}    #Invert receive clock\n"
	"\textended {value}           #Extended Clock values for T4E only\n"
        "\ttermination {none|resistive}\n"
        "\t                           #Flex and T4E+ only\n"
        "\tcoding {nrz|nrzi|FM0|FM1|manchester}\n"
        "\t                           #Flex and T4E+ only\n"
        "\tNRZIClocking {enable|disable}\n"
        "\t                           #T4E+ only\n"
        "\tcarrier {ignore|required}  #Carrier required or not\n"
	"\tled   {flash|normal}       #LED flash for card identification\n"
        "\tnumbuffers {number}        #Number of tx/rx buffers\n"
        "\tsizebuffers {size}         #Size of tx/rx buffers\n"
        "\tlowlatency {tx|rx|txrx|disable}\n"
        "\t                           #Set card into low latency mode\n"
        "\tlinemode {hdlc|async}\n"
        "\t                           #Set line mode for the port\n"
	"\n"
	"Async options:\n"
        "\tflow_control {none|hardware|software}\n"
        "\t                           #Type of flow control required\n"
        "\tstop_bits {1|1.5|2}        #Number of stop bits\n"
        "\tparity {none|odd|even}     #Type of parity checking in use\n"
        "\tword_len {5|6|7|8}         #Number of data bits per byte\n"
        "\txon_char {char}            #Character to turn software floe control on\n"
        "\txoff_char {char}           #Character to turn software flow control off\n"
	"\n"
	"SHDSL options:\n"
        "\tdatarate {192000-2304000}  #Required data rate\n"
        "\tvpi {0-255}                #ATM Virtual Path Identifier\n"
        "\tvci {0-65535}              #ATM Virtual Channel Identifier\n"
        "\ttermType {central|remote}  #Required number of timeslots\n"
        "\tannex {A|B}                #Region A (US) or B (EU)\n"
        "\tencap {ppp|mpoa}           #PPP or MPOA over ATM\n"
        "\ttestMode {0 .. 12}         #Test loop modes\n"
        "\tbackoff {0..??}            #Power backoff in dB\n"
        "\tsnrth {0..35}              #Signal/Noise Ratio threshold\n"
        "\tlpath {0..35}              #Loop Attenuation threshold\n"
        "\n"
	"Common T1/E1 options:\n"
	"\tmode {master|slave}        #Generate clock or recover from line\n"
        "\tinterface {bnc|rj48c}      #Interface connector to use\n"
        "\tequalizer {short|long}     #Short or long haul\n"
        "\tloop {none|local|remote|pload+ts0|pload-ts0}\n"
        "\t                           #Various loopback tests\n"
        "\trange {0-133ft||133-266ft|266-399ft|399-533ft|533-655ft}\n"
        "\t      {0-40m|40-81m|81-122m|122-162m|162-200m}\n"
        "\t                           #Range values in feet or meters\n"
        "\ttxBuffer {none|96bit|1Frame|2Frame}\n"
        "\t                           #Elastic TX buffer\n"
        "\trxBuffer {none|96bit|1Frame|2Frame}\n"
        "\t                           #Elastic RX buffer\n"
        "\tidlecode {0-0xFF}          #Value to transmit in idle timeslots\n"
	"Specific T1 options:\n"
        "\tdatarate {64000-1536000}   #Required data rate\n"
        "\tnumslots {1-24}            #Required number of timeslots\n"
        "\tstartslot {0-23}           #Starting timeslot # for fractional rates\n"
	"\tstructure {unframed|f4|f12|f24|f72}\n"
        "\t                           #Structure\n"
        "\tcoding {ami|ami_zcs|b8zs}\n"
        "\t                           #Line coding\n"
        "\tlbo {0|-7.5|-15|-22}       #Line build out\n"
	"Specific E1 options:\n"
        "\tdatarate {64000-2048000}   #Required data rate\n"
        "\tnumslots {1-32}            #Required number of timeslots\n"
        "\tstartslot {0-31}           #Starting timeslot # for fractional rates\n"
	"\tstructure {unframed|double|crc4|crc4m}\n"
        "\t                           #Structure\n"
        "\tcoding {hdb3}              #Line coding\n"

#ifdef GEN_HDLC
	"\n"
	"Cisco HDLC options:\n"
	"\tinterval #seconds          #Interval between keepalive packets\n"
	"\ttimeout  #seconds          #Timeout before link concidered down\n"
	"\n"
	"Frame Relay options:\n"
	"\tlmi   {none|ansi|ccitt}    #LMI type\n"
	"\tdce                        #Operate as a DCE\n"
	"\tt391  #seconds             #Polling verification timer\n"
	"\tt392  #seconds             #Link integrity polling timer\n"
	"\tn391  #                    #Full status polling counter\n"
	"\tn392  #                    #Error threshold\n"
	"\tn393  #                    #Monitored events count\n"
#endif
	, progname );
}


int
main ( int argc, char *argv[] )
{
	if (( progname = strrchr ( argv[0],'/')) != NULL )
	    ++progname;
	else
	    progname = argv[0];

	if ( argc >= 2 && strcmp ( argv[1],"-v") == 0 )
	{
	    verbose = TRUE;
	    --argc;
	    ++argv;
	}

	if ( argc <= 1 )
	{
	    usage ();
	    exit ( 1 );
	}

	if ( strlen ( argv[1] ) >= IFNAMSIZ )
	{
	    fprintf ( stderr,"%s: Interface name too long\n", progname );
	    exit ( 1 );
	}
	strcpy ( req.ifr_name, argv[1] );
	ifname = argv[1];

	if (( sock = socket ( PF_INET, SOCK_DGRAM, IPPROTO_IP )) < 0 )
	{
	    fprintf ( stderr,"%s: Unable to create socket. %s\n", progname,
					strerror ( errno ));
	    exit ( 1 );
	}
        /*
         * Try and open a device through which to do the configuration
         * Start with the network device, then the char then the tty
         */
        sock = get_config_device(argv[1], argv[2]);
        /*
         * note that when we return here, sock may actually be an fd
         */

	if ( argc == 2 )
	    do_info ( 1 );
	else if ( argc == 3 && strcmp ( argv[2],"stats") == 0 )
	    do_info ( 2 );
	else if ( argc == 3 && strcmp ( argv[2],"card") == 0 )
	    do_info ( 0 );
	else if ( argc == 3 && strcmp ( argv[2],"info") == 0 )
	    do_info ( 1 );
	else if ( argc == 3 && strcmp ( argv[2],"shellinfo") == 0 )
	    do_shellinfo ();
	else if ( argc == 3 && strcmp ( argv[2],"get_ints") == 0 )
	    do_get_intstats ();
	else if ( argc == 3 && strcmp ( argv[2],"reset_ints") == 0 )
	    do_reset_intstats ();
	else if ( argc == 3 && strcmp ( argv[2],"get_iocver") == 0 )
	    do_get_iocver ();
	else if ( argc == 4 && strcmp ( argv[2],"set_iocver") == 0 )
	    do_set_iocver (argv[3]);
	else if ( argc == 3 && strcmp ( argv[2],"stats_reset") == 0 )
	    do_reset_stats ();
#if GEN_HDLC
	else if ( argc == 4 && strcmp ( argv[2],"create") == 0 )
	    do_pvc ( IF_PROTO_FR_ADD_PVC, 0, argv[3]);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,22)
	else if ( argc == 4 && strcmp ( argv[2],"create_eth") == 0 )
	    do_pvc ( IF_PROTO_FR_ADD_ETH_PVC, 1, argv[3]);
#endif
	else if ( argc == 4 && strcmp ( argv[2],"delete") == 0 )
	    do_pvc ( IF_PROTO_FR_DEL_PVC, 0, argv[3]);
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,22)
	else if ( argc == 4 && strcmp ( argv[2],"delete_eth") == 0 )
	    do_pvc ( IF_PROTO_FR_DEL_ETH_PVC, 1, argv[3]);
#endif
#endif
	else if ( argc == 5 && strcmp ( argv[2],"load") == 0 )
	    do_load ( argv[3], argv[4] );
	else if ( argc >= 5 && strcmp ( argv[2],"set") == 0 )
	    do_set ( argc - 3, &argv[3] );
	else if ( argc >= 4 && ( strcmp ( argv[2],"media") == 0
			      || strcmp ( argv[2],"proto") == 0
			      || strcmp ( argv[2],"clock") == 0 
			      || strcmp ( argv[2],"phase") == 0))
	    do_set ( argc - 2, &argv[2] );
	else
	{
	    usage ();
	    exit ( 1 );
	}

	close ( sock );
	exit ( 0 );
}

/* End $Id: farutil.c,v 1.3 2009/12/17 11:31:22 kevinc Exp $ */
