/*
 *      FarSync WAN driver for Linux
 *
 *      Copyright (C) 2001-2004 FarSite Communications Ltd.
 *      www.farsite.co.uk
 *
 *      This program is free software; you can redistribute it and/or
 *      modify it under the terms of the GNU General Public License
 *      as published by the Free Software Foundation; either version
 *      2 of the License, or (at your option) any later version.
 *
 *      Author: K. P. Curtis      <kevin.curtis@farsite.co.uk>
 *
 *      This file contains the data declarations and tables for the
 *      farutil configuration utiliy.  farutl.c was just getting too big.
 */
/*	Revision History
 *	$Log: farutil.h,v $
 *	Revision 1.3  2009/12/17 11:31:22  kevinc
 *	Make the release version 1.09.05 - build 143
 *	
 *	Revision 1.2  2009/07/10 08:23:25  kevinc
 *	Merging 1.09.04 and async branches
 *	
 *	Revision 1.1.1.1  2009/05/06 12:45:30  kevinc
 *	
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <limits.h>
#include <linux/if.h>
#include <linux/types.h>

#ifdef GEN_HDLC
#include <linux/hdlc.h>
#include <linux/sockios.h>
#include <linux/version.h>
#endif

#include "farsync.h"

/* Pet peeve, why something like this never went into a consistent standard
 * header. Or even "enum bool { FALSE, TRUE };"
 */
#ifndef TRUE
# define TRUE   1
# define FALSE  0
  typedef int bool;
#endif

/*
 *	DOS executive header structure
 */
struct exehdr
{
	unsigned short exe_sig_mz;	/* 00 'M','Z' for .exe's	*/
	unsigned short image_length;	/* 02				*/
	unsigned short exe_size;	/* 04 in 512 byte pages		*/
	unsigned short num_relocs;	/* 06				*/
	unsigned short hdr_size;	/* 08 in paragraphs (16 bytes)	*/
	unsigned short min_para;	/* 0A "    "   "      "   "	*/
	unsigned short max_para;	/* 0C "    "   "      "   "	*/
	unsigned short stack_disp;	/* 0E "    "   "      "   "	*/
	unsigned short sp_offset;	/* 10				*/
	unsigned short word_sum;	/* 12 word checksum of code	*/
	unsigned short ip_offset;	/* 14 start address of code	*/
	unsigned short code_disp;	/* 16 starting code segment	*/
	unsigned short reloc_disp;	/* 18				*/
	unsigned short overlay_num;	/* 1A				*/
};


/*	General program variables
 */
static char *progname;
static bool verbose;
static int sock;
static struct ifreq req;
static char *ifname;

/*	Variables and parameters for firmware download
 */
static int bios_start_seg;
static int bios_start_off;


#define CBUFSIZ	(128*1024)	/* Size of the code buffer. Must equal the size
				 * of memory zeroed when the card is started. It
				 * is also assumed that any code to be loaded
				 * will fit in a single buffer.
				 */
static struct fstioc_write *buf;

/*	Parameters for mode setting
 */
#ifdef GEN_HDLC
static int protocol = 0;	/* Selected protocol */
static fr_proto fr;		/* Frame Relay parameters */
static cisco_proto cisco;	/* Cisco HDLC parameters */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,22)
static raw_hdlc_proto raw_hdlc;
#endif
#endif

static char *card_strings[] = {
	"Unknown (0)",
	"FarSync T2P",
	"FarSync T4P",
	"FarSync T1U",
	"FarSync T2U",
	"FarSync T4U",
        "FarSync TE1",
        "FarSync DSL_S1",
        "FarSync T4E",
	"FarSync Flex",
        "FarSync T4UE",
        "FarSync T2UE",
        "FarSync T4E+"
};

#define ALLOW_NONSTD_RATES	1	/* Allow non-standard data rates */

static struct speeds {
	char *text;
	unsigned long val;
} speed_table[] = {
	{"9600", 9600 },	{"9k6", 9600 },
	{"19200", 19200 },	{"19k2", 19200 },
	{"38400", 38400 },	{"38k4", 38400 },
#if ALLOW_NONSTD_RATES
	{"76800", 76800 },	{"76k8", 76800 },
	{"153600", 153600 },	{"153k6", 153600 },
	{"307200", 307200 },	{"307k2", 307200 },
	{"614400", 614400 },	{"614k4", 614400 },
#endif
	{"64000", 64000 },	{"64k", 64000 },
	{"128000", 128000 },	{"128k", 128000 },
	{"256000", 256000 },	{"256k", 256000 },
	{"512000", 512000 },	{"512k", 512000 },
	{"1024000", 1024000 },	{"1024k", 1024000 },	{"1m", 1024000 },
	{"2048000", 2048000 },	{"2048k", 2048000 },	{"2m", 2048000 },
	{"4096000", 4096000 },	{"4096k", 4096000 },	{"4m", 4096000 },
	{"8192000", 8192000 },	{"8192k", 8192000 },	{"8m", 8192000 },
	{ NULL, 0 }
};

static struct speeds async_speeds[] = {
	{"75", 75 },
        {"150", 150 },
        {"300", 300 },
        {"600", 600 },
        {"1200", 1200 },        {"1k2",  1200 },
        {"2400", 2400 },        {"2k4",  2400 },
        {"4800", 4800 },        {"4k8",  4800 },
        {"8000", 8000 },        {"8k",   8000 },
        {"9600", 9600 },        {"9k6",  9600 },
        {"12000", 12000 },      {"12k",  12000 },
	{"16000", 16000 },	{"16k",  16000 },
	{"19200", 19200 },	{"19k2", 19200 },
	{"24000", 24000 },	{"24k",  24000 },
	{"32000", 32000 },	{"32k",  32000 },
	{"38400", 38400 },	{"38k4", 38400 },
	{"48000", 48000 },	{"48k",  48000 },
        {"57600", 57600 },      {"57k6", 57600 },
        {"76800", 76800 },	{"76k8", 76800 },
	{"96000", 96000 },      {"96k",  96000 },
        {"115200", 115200 },    {"115k2",115200},
	{ NULL, 0 }
};

static struct speeds flex_speeds[] = {
	{"15", 15 },
	{"25", 25 },
	{"30", 30 },
	{"50", 50 },
	{"60", 60 },
	{"75", 75 },
	{"100", 100 },
	{"120", 120 },
	{"125", 125 },
        {"150", 150 },
	{"200", 200 },
	{"240", 240 },
        {"300", 300 },
        {"375", 375 },
        {"400", 400 },
        {"480", 480 },
        {"500", 500 },
        {"600", 600 },
        {"750", 750 },
        {"800", 800 },
        {"960", 960 },
        {"1000", 1000 },        {"1k0", 1000 },
        {"1200", 1200 },        {"1k2", 1200 },
        {"1500", 1500 },        {"1k5", 1500 },
        {"1600", 1600 },        {"1k6", 1600 },
        {"1920", 1920 },        {"1k92", 1920 },
        {"2000", 2000 },        {"2k0", 2000 },
        {"2400", 2400 },        {"2k4", 2400 },
        {"3000", 3000 },        {"3k0", 3000 },
        {"3200", 3200 },        {"3k2", 3200 },
        {"3840", 3840 },        {"3k84", 3840 },
        {"4000", 4000 },        {"4k0", 4000 },
        {"4800", 4800 },        {"4k8", 4800 },
        {"6000", 6000 },        {"6k0", 6000 },
        {"6400", 6400 },        {"6k4", 6400 },
        {"7680", 7680 },        {"7k68", 7680 },
        {"8000", 8000 },        {"8k", 8000 },
	{"9600", 9600 },	{"9k6", 9600 },
	{"12000", 12000 },	{"12k", 12000 },
	{"12800", 12800 },	{"12k8", 12800 },
	{"15360", 15360 },	{"15k36", 16360 },
	{"16000", 16000 },	{"16k", 16000 },
	{"19200", 19200 },	{"19k2", 19200 },
	{"24000", 24000 },	{"24k", 24000 },
	{"25600", 25600 },	{"25k6", 25600 },
	{"30720", 30720 },	{"30k72", 30720 },
	{"32000", 32000 },	{"32k", 32000 },
	{"38400", 38400 },	{"38k4", 38400 },
	{"48000", 48000 },	{"48k", 48000 },
	{"51200", 51200 },	{"51k2", 51200 },
	{"61440", 61440 },	{"61k44", 61440 },
	{"64000", 64000 },	{"64k", 64000 },
        {"76800", 76800 },	{"76k8", 76800 },
	{"96000", 96000 },	{"96k", 96000 },
	{"102400", 102400 },	{"102k4", 102400 },
	{"122880", 122880 },	{"122k88", 122880 },
	{"128000", 128000 },	{"128k", 128000 },
	{"153600", 153600 },	{"153k6", 153600 },
	{"192000", 192000 },	{"192k", 192000 },
	{"204800", 204800 },	{"204k8", 240800 },
	{"245760", 245760 },	{"245k76", 245760 },
	{"256000", 256000 },	{"256k", 256000 },
	{"307200", 307200 },	{"307k2", 307200 },
	{"384000", 384000 },	{"384k", 384000 },
	{"409600", 409600 },	{"409k6", 409600 },
	{"491520", 491520 },	{"491k52", 491520 },
	{"512000", 512000 },	{"512k", 512000 },
	{"614400", 614400 },	{"614k4", 614400 },
	{"768000", 768000 },	{"768k", 768000 },
	{"819200", 819200 },	{"819k2", 819200 },
	{"983040", 983040 },	{"983k04", 983040 },
	{"1024000", 1024000 },	{"1024k", 1024000 },	{"1m", 1024000 },
	{"1228800", 1228800 },	{"1228k8", 1228800 },	{"1m2288", 1228800 },
	{"1536000", 1536000 },	{"1536k", 1536000 },    {"1m536", 1536000},
	{"1638400", 1638400 },	{"1638k4", 1638400 },	{"1m6384", 1638400 },
	{"2048000", 2048000 },	{"2048k", 2048000 },	{"2m", 2048000 },
	{"2457600", 2457600 },	{"2457k6", 2457600 },	{"2m4576", 2457600 },
	{"3072000", 3072000 },	{"3072k", 3072000 },    {"3m072", 3072000},
	{"4096000", 4096000 },	{"4096k", 4096000 },	{"4m", 4096000 },
	{"4915200", 4915200 },	{"4915k2", 4915200 },	{"4m9152", 4915200 },
	{"6144000", 6144000 },	{"6144k", 6144000 },    {"6m144", 6144000},
	{"8192000", 8192000 },	{"8192k", 8192000 },	{"8m", 8192000 },
	{ NULL, 0 }
};



/*
 * Another table of additional speeds for T4E
 */
static struct speeds t4e_speeds[] = {
        {"300", 300},
        {"600", 600},
        {"1200", 1200},         {"1k2",  1200},
        {"2400", 2400},         {"2k4",  2400},
        {"4800", 4800},         {"4k8",  4800},
        {"7200", 7200},         {"7k2",  7200},
        {"8000", 8000},         {"8k",   8000},
        {"9600", 9600},         {"9k6",  9600},
        {"12000", 12000},       {"12k",  12000},
        {"14400", 14400},       {"14k4", 14400},
        {"16000", 16000},       {"16k",  16000},
        {"16800", 16800},       {"16k8", 16800},
        {"19200", 19200},       {"19k2", 19200},
        {"21600", 21600},       {"21k6", 21600},
        {"24000", 24000},       {"24k",  24000},
        {"26400", 26400},       {"26k4", 26400},
        {"28800", 28800},       {"28k8", 28000},
        {"31200", 31200},       {"31k2", 31200},
        {"32000", 32000},       {"32k",  32000},
        {"33600", 33600},       {"33k6", 33600},
        {"36000", 36000},       {"36k",  36000},
        {"38400", 38400},       {"38k4", 38400},
        {"40000", 40000},       {"40k",  40000},
        {"40800", 40800},       {"40k8", 40800},
        {"43200", 43200},       {"43k2", 43200},
        {"48000", 48000},       {"48k",  48000},
        {"56000", 56000},       {"56k",  56000},
        {"64000", 64000},       {"64k",  64000},
        {"80000", 80000},       {"80k",  80000},
        {"96000", 96000},       {"96k",  96000},
        {"112000", 112000},     {"112k", 112000},
        {"128000", 128000},     {"128k", 128000},
        {"160000", 160000},     {"160k", 160000},
        {"192000", 192000},     {"192k", 192000},
        {"224000", 224000},     {"224k", 224000},
        {"256000", 256000},     {"256k", 256000},
        {"320000", 320000},     {"320k", 320000},
        {"384000", 384000},     {"384k", 384000},
        {"448000", 448000},     {"448k", 448000},
        {"512000", 512000},     {"512k", 512000},
        {"576000", 576000},     {"576k", 576000},
        {"640000", 640000},     {"640k", 640000},
        {"704000", 704000},     {"704k", 704000},
        {"768000", 768000},     {"768k", 768000},
        {"832000", 832000},     {"832k", 832000},
        {"896000", 896000},     {"896k", 896000},
        {"960000", 960000},     {"960k", 960000},
        {"1000000", 1000000},   {"1000k", 1000000},     {"1m",    1000000},
	{"1024000", 1024000},   {"1024k", 1024000},     {"1m024", 1024000},
        {"1088000", 1088000},   {"1088k", 1088000},     {"1m088", 1088000},
        {"1152000", 1152000},   {"1152k", 1152000},     {"1m152", 1152000},
        {"1216000", 1216000},   {"1216k", 1216000},     {"1m216", 1216000},
        {"1280000", 1280000},   {"1280k", 1280000},     {"1m28",  1280000},
        {"1344000", 1344000},   {"1344k", 1344000},     {"1m344", 1344000},
        {"1408000", 1408000},   {"1408k", 1408000},     {"1m408", 1408000},
        {"1472000", 1472000},   {"1472k", 1472000},     {"1m472", 1472000},
        {"1536000", 1536000},   {"1536k", 1536000},     {"1m536", 1536000},
        {"1600000", 1600000},   {"1600k", 1600000},     {"1m6",   1600000},
        {"1664000", 1664000},   {"1664k", 1664000},     {"1m664", 1664000},
        {"1728000", 1728000},   {"1728k", 1728000},     {"1m728", 1728000},
        {"1792000", 1792000},   {"1792k", 1792000},     {"1m792", 1792000},
        {"1856000", 1856000},   {"1856k", 1856000},     {"1m856", 1856000},
        {"1920000", 1920000},   {"1920k", 1920000},     {"1m920", 1920000},
        {"1984000", 1984000},   {"1984k", 1984000},     {"1m984", 1984000},
        {"2000000", 2000000},   {"2000k", 2000000},     {"2m",    2000000},
        {"2048000", 2048000},   {"2048k", 2048000},     {"2m048", 2048000},
        {"2112000", 2112000},   {"2112k", 2112000},     {"2m112", 2112000},
        {"2176000", 2176000},   {"2176k", 2176000},     {"2m176", 2176000},
        {"2240000", 2240000},   {"2240k", 2240000},     {"2m24",  2240000},
        {"2304000", 2304000},   {"2304k", 2304000},     {"2m304", 2304000},
        {"2368000", 2368000},   {"2368k", 2368000},     {"2m368", 2368000},
        {"2432000", 2432000},   {"2432k", 2432000},     {"2m432", 2432000},
        {"2496000", 2496000},   {"2496k", 2496000},     {"2m496", 2496000},
        {"2560000", 2560000},   {"2560k", 2560000},     {"2m56",  2560000},
        {"2624000", 2624000},   {"2624k", 2624000},     {"2m624", 2624000},
        {"2688000", 2688000},   {"2688k", 2688000},     {"2m688", 2688000},
        {"2752000", 2752000},   {"2752k", 2752000},     {"2m752", 2752000},
        {"2816000", 2816000},   {"2816k", 2816000},     {"2m816", 2816000},
        {"2880000", 2880000},   {"2880k", 2880000},     {"2m88",  2880000},
        {"2944000", 2944000},   {"2944k", 2944000},     {"2m944", 2944000},
        {"3000000", 3000000},   {"3000k", 3008000},     {"3m",    3000000},
        {"3008000", 3008000},   {"3008k", 3008000},     {"3m008", 3008000},
        {"3072000", 3072000},   {"3072k", 3072000},     {"3m072", 3072000},
        {"3136000", 3136000},   {"3136k", 3136000},     {"3m136", 3136000},
        {"3200000", 3200000},   {"3200k", 3200000},     {"3m2",   3200000},
        {"3264000", 3264000},   {"3264k", 3264000},     {"3m264", 3264000},
        {"3328000", 3328000},   {"3328k", 3328000},     {"3m328", 3328000},
        {"3392000", 3392000},   {"3392k", 3392000},     {"3m392", 3392000},
        {"3456000", 3456000},   {"3456k", 3456000},     {"3m456", 3456000},
        {"3520000", 3520000},   {"3520k", 3520000},     {"3m52",  3520000},
        {"3584000", 3584000},   {"3584k", 3584000},     {"3m584", 3584000},
        {"3648000", 3648000},   {"3648k", 3648000},     {"3m648", 3648000},
        {"3712000", 3712000},   {"3712k", 3712000},     {"3m712", 3712000},
        {"3776000", 3776000},   {"3776k", 3776000},     {"3m776", 3776000},
        {"3840000", 3840000},   {"3840k", 3840000},     {"3m84",  3840000},
        {"3904000", 3904000},   {"3904k", 3904000},     {"3m904", 3904000},
        {"3968000", 3968000},   {"3968k", 3968000},     {"3m968", 3968000},
        {"4000000", 4000000},   {"4000k", 4000000},     {"4m",    4000000},
        {"4032000", 4032000},   {"4032k", 4032000},     {"4m032", 4032000},
        {"4096000", 4096000},   {"4096k", 4096000},     {"4m096", 4096000},
        {"4160000", 4160000},   {"4160k", 4160000},     {"4m16",  4160000},
        {"4224000", 4224000},   {"4224k", 4224000},     {"4m224", 4224000},
        {"4288000", 4288000},   {"4288k", 4288000},     {"4m288", 4288000},
        {"4352000", 4352000},   {"4352k", 4352000},     {"4m352", 4352000},
        {"4416000", 4416000},   {"4416k", 4416000},     {"4m416", 4416000},
        {"4480000", 4480000},   {"4480k", 4480000},     {"4m48",  4480000},
        {"4544000", 4544000},   {"4544k", 4544000},     {"4m544", 4544000},
        {"4608000", 4608000},   {"4608k", 4608000},     {"4m608", 4608000},
        {"4672000", 4672000},   {"4672k", 4672000},     {"4m672", 4672000},
        {"4736000", 4736000},   {"4736k", 4736000},     {"4m736", 4736000},
        {"4800000", 4800000},   {"4800k", 4800000},     {"4m8",   4800000},
        {"4864000", 4864000},   {"4864k", 4864000},     {"4m864", 4864000},
        {"4928000", 4928000},   {"4928k", 4928000},     {"4m928", 4928000},
        {"4992000", 4992000},   {"4992k", 4992000},     {"4m992", 4992000},
        {"5000000", 5000000},   {"5000k", 5000000},     {"5m",    5000000},
        {"5056000", 5056000},   {"5056k", 5056000},     {"5m056", 5056000},
        {"5120000", 5120000},   {"5120k", 5120000},     {"5m12",  5120000},
        {"5184000", 5184000},   {"5184k", 5184000},     {"5m184", 5184000},
        {"5248000", 5248000},   {"5248k", 5248000},     {"5m248", 5248000},
        {"5312000", 5312000},   {"5312k", 5312000},     {"5m312", 5312000},
        {"5376000", 5376000},   {"5376k", 5376000},     {"5m376", 5376000},
        {"5440000", 5440000},   {"5440k", 5440000},     {"5m44",  5440000},
        {"5504000", 5504000},   {"5504k", 5504000},     {"5m504", 5504000},
        {"5568000", 5568000},   {"5568k", 5568000},     {"5m568", 5568000},
        {"5632000", 5632000},   {"5632k", 5632000},     {"5m632", 5632000},
        {"5696000", 5696000},   {"5696k", 5696000},     {"5m696", 5696000},
        {"5760000", 5760000},   {"5760k", 5760000},     {"5m76",  5760000},
        {"5824000", 5824000},   {"5824k", 5824000},     {"5m824", 5824000},
        {"5888000", 5888000},   {"5888k", 5888000},     {"5m888", 5888000},
        {"5952000", 5952000},   {"5952k", 5952000},     {"5m952", 5952000},
        {"6000000", 6000000},   {"6000k", 6000000},     {"6m",    6000000},
        {"6016000", 6016000},   {"6016k", 6016000},     {"6m016", 6016000},
        {"6080000", 6080000},   {"6080k", 6080000},     {"6m08",  6080000},
        {"6144000", 6144000},   {"6144k", 6144000},     {"6m144", 6144000},
        {"6208000", 6208000},   {"6208k", 6208000},     {"6m208", 6208000},
        {"6272000", 6272000},   {"6272k", 6272000},     {"6m272", 6272000},
        {"6336000", 6336000},   {"6336k", 6336000},     {"6m336", 6336000},
        {"6400000", 6400000},   {"6400k", 6400000},     {"6m4",   6400000},
        {"6464000", 6464000},   {"6464k", 6464000},     {"6m464", 6464000},
        {"6528000", 6528000},   {"6528k", 6528000},     {"6m528", 6528000},
        {"6553600", 6553600},   {"6553k", 6553600},     {"6m553", 6553600},
        {"6592000", 6592000},   {"6592k", 6592000},     {"6m592", 6592000},
        {"6656000", 6656000},   {"6656k", 6656000},     {"6m656", 6656000},
        {"6720000", 6720000},   {"6720k", 6720000},     {"6m72",  6720000},
        {"6784000", 6784000},   {"6784k", 6784000},     {"6m784", 6784000},
        {"6848000", 6848000},   {"6848k", 6848000},     {"6m848", 6848000},
        {"6912000", 6912000},   {"6912k", 6912000},     {"6m912", 6912000},
        {"6976000", 6976000},   {"6976k", 6976000},     {"6m976", 6976000},
        {"7000000", 7000000},   {"7000k", 7000000},     {"7m",    7040000},
        {"7040000", 7040000},   {"7040k", 7040000},     {"7m04",  7040000},
        {"7104000", 7104000},   {"7104k", 7104000},     {"7m104", 7104000},
        {"7168000", 7168000},   {"7168k", 7168000},     {"7m168", 7168000},
        {"7232000", 7232000},   {"7232k", 7232000},     {"7m232", 7232000},
        {"7296000", 7296000},   {"7296k", 7296000},     {"7m296", 7296000},
        {"7360000", 7360000},   {"7360k", 7360000},     {"7m36",  7360000},
        {"7424000", 7424000},   {"7424k", 7424000},     {"7m424", 7424000},
        {"7488000", 7488000},   {"7488k", 7488000},     {"7m488", 7488000},
        {"7552000", 7552000},   {"7552k", 7552000},     {"7m552", 7552000},
        {"7616000", 7616000},   {"7616k", 7616000},     {"7m616", 7616000},
        {"7680000", 7680000},   {"7680k", 7680000},     {"7m68",  7680000},
        {"7744000", 7744000},   {"7744k", 7744000},     {"7m744", 7744000},
        {"7808000", 7808000},   {"7808k", 7808000},     {"7m808", 7808000},
        {"7872000", 7872000},   {"7872k", 7872000},     {"7m872", 7872000},
        {"7936000", 7936000},   {"7936k", 7936000},     {"7m936", 7936000},
        {"8000000", 8000000},   {"8000k", 8000000},     {"8m",    8000000},
        {"8064000", 8064000},   {"8064k", 8064000},     {"8m064", 8064000},
        {"8128000", 8128000},   {"8128k", 8128000},     {"8m128", 8128000},
        {"8192000", 8192000},   {"8192k", 8192000},     {"8m192", 8192000},
        {"9000000", 9000000},   {"9000k", 9000000},     {"9m",    9000000},
        {"10000000", 10000000},  {"10000k", 10000000},   {"10m",  10000000},
	{ NULL, 0 }
};

static struct speeds shdsl_speed_table[] = {
        {"192000", 192000 },    {"192k", 192000 },
	{"256000", 256000 },	{"256k", 256000 },
        {"320000", 320000 },    {"320k", 320000 },
        {"384000", 384000 },    {"384k", 384000 },
        {"448000", 448000 },    {"448k", 448000 },
	{"512000", 512000 },	{"512k", 512000 },
        {"576000", 576000 },    {"576k", 576000 },
        {"640000", 640000 },    {"640k", 640000 },
        {"704000", 704000 },    {"704k", 704000 },
        {"768000", 768000 },    {"768k", 768000 },
        {"832000", 832000 },    {"832k", 832000 },
        {"896000", 896000 },    {"896k", 896000 },
        {"960000", 960000 },    {"960k", 960000 },
	{"1024000", 1024000 },	{"1024k", 1024000 },	{"1.024m", 1024000 },
	{"1088000", 1088000 },	{"1088k", 1088000 },	{"1.088m", 1088000 },
	{"1152000", 1152000 },	{"1152k", 1152000 },	{"1.152m", 1152000 },
        {"1216000", 1216000 },  {"1216k", 1216000 },	{"1.216m", 1216000 },
	{"1280000", 1280000 },	{"1280k", 1280000 },	{"1.280m", 1280000 },
        {"1344000", 1344000 },  {"1344k", 1344000 },	{"1.344m", 1344000 },
        {"1408000", 1408000 },  {"1408k", 1408000 },	{"1.408m", 1408000 },
        {"1472000", 1472000 },  {"1472k", 1472000 },	{"1.472m", 1472000 },
	{"1536000", 1536000 },	{"1536k", 1536000 },	{"1.536m", 1536000 },
        {"1600000", 1600000 },  {"1600k", 1600000 },	{"1.6m",   1600000 },
        {"1664000", 1664000 },  {"1664k", 1664000 },	{"1.664m", 1664000 },
        {"1728000", 1728000 },  {"1728k", 1728000 },	{"1.728m", 1728000 },
        {"1792000", 1792000 },  {"1792k", 1792000 },	{"1.792m", 1792000 },
        {"1856000", 1856000 },  {"1856k", 1856000 },	{"1.856m", 1856000 },
        {"1920000", 1920000 },  {"1920k", 1920000 },	{"1.92m",  1920000 },
        {"1984000", 1984000 },  {"1984k", 1984000 },	{"1.984m", 1984000 },
	{"2048000", 2048000 },	{"2048k", 2048000 },	{"2m", 2048000 },
        {"2112000", 2112000 },  {"2112k", 2112000 },	{"2.112m", 2112000 },
	{"2176000", 2176000 },	{"2176k", 2176000 },	{"2.176m", 2176000 },
	{"2240000", 2240000 },	{"2240k", 2240000 },	{"2.240m", 2240000 },
	{"2304000", 2304000 },	{"2304k", 2304000 },	{"2.304m", 2304000 },
	{ NULL, 0 }
};

static struct speeds e1_speed_table[] = {
	{"64000", 64000 },	{"64k", 64000 },
	{"128000", 128000 },	{"128k", 128000 },
        {"192000", 192000 },    {"192k", 192000 },
	{"256000", 256000 },	{"256k", 256000 },
        {"320000", 320000 },    {"320k", 320000 },
        {"384000", 384000 },    {"384k", 384000 },
        {"448000", 448000 },    {"448k", 448000 },
	{"512000", 512000 },	{"512k", 512000 },
        {"576000", 576000 },    {"576k", 576000 },
        {"640000", 640000 },    {"640k", 640000 },
        {"704000", 704000 },    {"704k", 704000 },
        {"768000", 768000 },    {"768k", 768000 },
        {"832000", 832000 },    {"832k", 832000 },
        {"896000", 896000 },    {"896k", 896000 },
        {"960000", 960000 },    {"960k", 960000 },
	{"1024000", 1024000 },	{"1024k", 1024000 },	{"1.024m", 1024000 },
	{"1088000", 1088000 },	{"1088k", 1088000 },	{"1.088m", 1088000 },
	{"1152000", 1152000 },	{"1152k", 1152000 },	{"1.152m", 1152000 },
        {"1216000", 1216000 },  {"1216k", 1216000 },	{"1.216m", 1216000 },
	{"1280000", 1280000 },	{"1280k", 1280000 },	{"1.280m", 1280000 },
        {"1344000", 1344000 },  {"1344k", 1344000 },	{"1.344m", 1344000 },
        {"1408000", 1408000 },  {"1408k", 1408000 },	{"1.408m", 1408000 },
        {"1472000", 1472000 },  {"1472k", 1472000 },	{"1.472m", 1472000 },
	{"1536000", 1536000 },	{"1536k", 1536000 },	{"1.536m", 1536000 },
        {"1600000", 1600000 },  {"1600k", 1600000 },	{"1.6m",   1600000 },
        {"1664000", 1664000 },  {"1664k", 1664000 },	{"1.664m", 1664000 },
        {"1728000", 1728000 },  {"1728k", 1728000 },	{"1.728m", 1728000 },
        {"1792000", 1792000 },  {"1792k", 1792000 },	{"1.792m", 1792000 },
        {"1856000", 1856000 },  {"1856k", 1856000 },	{"1.856m", 1856000 },
        {"1920000", 1920000 },  {"1920k", 1920000 },	{"1.92m",  1920000 },
        {"1984000", 1984000 },  {"1984k", 1984000 },	{"1.984m", 1984000 },
	{"2048000", 2048000 },	{"2048k", 2048000 },	{"2m", 2048000 },
	{ NULL, 0 }
};

static struct speeds t1_speed_table[] = {
	{"64000", 64000 },	{"64k", 64000 },
	{"128000", 128000 },	{"128k", 128000 },
        {"192000", 192000 },    {"192k", 192000 },
	{"256000", 256000 },	{"256k", 256000 },
        {"320000", 320000 },    {"320k", 320000 },
        {"384000", 384000 },    {"384k", 384000 },
        {"448000", 448000 },    {"448k", 448000 },
	{"512000", 512000 },	{"512k", 512000 },
        {"576000", 576000 },    {"576k", 576000 },
        {"640000", 640000 },    {"640k", 640000 },
        {"704000", 704000 },    {"704k", 704000 },
        {"768000", 768000 },    {"768k", 768000 },
        {"832000", 832000 },    {"832k", 832000 },
        {"896000", 896000 },    {"896k", 896000 },
        {"960000", 960000 },    {"960k", 960000 },
	{"1024000", 1024000 },	{"1024k", 1024000 },	{"1.024m", 1024000 },
	{"1088000", 1088000 },	{"1088k", 1088000 },	{"1.088m", 1088000 },
	{"1152000", 1152000 },	{"1152k", 1152000 },	{"1.152m", 1152000 },
        {"1216000", 1216000 },  {"1216k", 1216000 },	{"1.216m", 1216000 },
	{"1280000", 1280000 },	{"1280k", 1280000 },	{"1.280m", 1280000 },
        {"1344000", 1344000 },  {"1344k", 1344000 },	{"1.344m", 1344000 },
        {"1408000", 1408000 },  {"1408k", 1408000 },	{"1.408m", 1408000 },
        {"1472000", 1472000 },  {"1472k", 1472000 },	{"1.472m", 1472000 },
	{"1536000", 1536000 },	{"1536k", 1536000 },	{"1.536m", 1536000 },
	{ NULL, 0 }
};

static char *type_strings[] = {
	"Unknown (0)",
	"T2P FarSync WAN T-Series",
	"T4P FarSync WAN T-Series",
	"T1U FarSync WAN T-Series",
	"T2U FarSync WAN T-Series",
	"T4U FarSync WAN T-Series",
        "TE1 FarSync WAN",
        "DSL-S1 FarSync WAN",
        "T4E FarSync WAN T-Series",
        "FarSync Flex WAN",
        "T4UE FarSync WAN T-Series",
        "T2UE FarSync WAN T-Series",
        "T4E+ FarSync WAN T-Series"
};

static char *state_strings[] = {
	"Uninitialised (startup)",
	"Processor reset",
	"Downloading",
	"Firmware initialising",
	"Running normally",
	"Bad shared memory version detected",
	"Processor halted",
	"Firmware initialisation failed"
};

static char *phys_strings[] = {
	"Unknown (0)",
	"V.24 (RS232C)",
	"X.21 (RS422/V.11)",
	"V.35",
	"X.21 (Dual clocking)",
        "No cable",
        "RS530/449",
        "T1",
        "E1",
        "J1",
        "SHDSL",
        "RS485",
        "UX35C"
};

static char *cable_strings[] = {
	"Cable presence detected",
	"No cable has been detected"
};

static char *latency_strings[] = {
        "not enabled",
        "rx enabled",
        "tx enabled",
        "rx and tx enabled"
};

static char *proto_strings[] = {
	"Unknown (0)",
	"Cisco compatible HDLC",
	"Sync PPP",
	"Monitor only",
	"Raw packet interface",
	"Generic HDLC driver"
};

static char *mode_strings[] = {
	"HDLC",
	"Transparent",
        "Bisync",
	"Async"
};

static char *flow_strings[] = {
        "Invalid",
        "None",
        "Hardware",
        "Software"
};

static char *stop_strings[] = {
        "1",
        "1.5",
        "2"
};

static char *parity_strings[] = {
        "No Parity",
        "Odd",
        "Even",
        "Force Parity 1",
        "Force Parity 0"
};

static char *structure_strings[] = {
        "Unframed",
        "Double",
        "CRC4",
        "CRC4M",
        "F4",
        "F12",
        "F24",
        "F72"
};

static char *coding_strings[] = {
        "HDB3",
        "NRZ",
        "CMI",
        "CMI_HDB3",
        "CMI_B8ZS",
        "AMI",
        "AMI_ZCS",
        "B8ZS"
};

static char *lbo_strings[] = {
        "0 dB",
        "-7.5 dB",
        "-15 dB",
        "-22 dB"
};

static char *range_strings[] = {
        "0-133 ft or 0-40 m",
        "133-266 ft or 41-81 m",
        "266-399 ft or 81-122 m",
        "399-533 ft or 122-162 m",
        "533-655 ft or 162-200 m"
};

static char *loop_strings[] = {
        "None",
        "Local",
        "Payload excluding time slot 0",
        "Payload including time slot 0",
        "remote"
};

static char *buff_strings[] = {
        "2 frame",
        "1 frame",
        "96 bit",
        "None"
};

static char *activation_strings[] = {
        "SDSL Power Down",
        "Send Single Pulses",
        "Initialisation",
        "Processing CL",
        "TR Analog Loop Closed",
        "TR Analog Loop Active",
        "Non TR Analog Loop Closed",
        "Non TR Analog Loop Active",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "GHS Startup",
        "GHS Transaction Auto",
        "Line Probing",
        "GHS Finished",
        "GHS Silence",
        "GHS Startup Failed",
        "GHS Transaction Failed",
        "Line Probing Failed",
        "GHS No Common Mode",
        "Time Out",
        "Idle",
        "Wait STU",
        "LP Finished",
        "Time viloation Host",
        "Invalid",
        "Invalid",
        "Invalid",
        "Wait For CR",
        "Transmit SC",
        "Transmit TC",
        "Transmit FC",
        "Invalid",
        "Invalid",
        "Invalid",
        "Transmit CR",
        "CRSR Silence",
        "Transmit SR",
        "Transmit TR",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "SR Not Detected",
        "TR Not Detected",
        "Data Axception",
        "Not Detected",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "TC Not Detected",
        "FC Not Detected",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Data",
        "Data + CRC Anomaly",
        "Data + LOSW Defect",
        "Invalid",
        "Data + Loop Atten Defect",
        "Invalid",
        "Invalid",
        "Invalid",
        "Data + SNR Margin Defect",
        "Data + ???",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Data + LOSW Failure",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Invalid",
        "Data CIE",
        "Invalid"
};
